/*
 * LocalOps.java
 *
 * Created on 20. bezen 2004, 14:03
 */

package SOFA.SOFAnet.Core;

import SOFA.SOFAnet.Repository.*;
import SOFA.SOFAnet.TR.TRInterface;
import SOFA.SOFAnode.TR.Impl.*;
import SOFA.SOFAnode.TR.*;
import SOFA.SOFAnet.Transport.*;
import SOFA.SOFAnet.Search.SearchReplyItem;
import java.io.*;
import java.util.*;

/**
 * Local operations of the SOFAnet.
 *
 * @author  Ladislav Sobr
 */
public class LocalOps
{
  private TRInterface tr;
  private TransportInterface transport;
  private Repository rep;
  private NetOps netOps;
  private ShareOps shareOps;
  private SearchOps searchOps;
  
  /** Creates a new instance of LocalOps */
  public LocalOps()
  {
  }
  
  public void setRepository(Repository repository)
  {
    rep = repository;
  }
  
  public void setTRInterface(TRInterface tr)
  {
    this.tr = tr;
  }
  
  public void setTransportInterface(TransportInterface transport)
  {
    this.transport = transport;
  }
  
  public void setNetOps(NetOps netOps)
  {
    this.netOps = netOps;
  }
  
  public void setShareOps(ShareOps shareOps)
  {
    this.shareOps = shareOps;
  }
  
  public void setSearchOps(SearchOps searchOps)
  {
    this.searchOps = searchOps;
  }
  
  /**
   * Tests whether bundle is installed
   *
   * @param bundleName name of bundle
   * @return true if installed, else false
   */
  public boolean isBundleInstalled(String bundleName)
  {
    LocalInfos localInfos = rep.getLocalInfos();
    synchronized (localInfos)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo == null) return false;
      return localInfo.isInstalled();
    }
  }
  
  /**
   * Installs bundle
   *
   * @param bundleName name of bundle
   * @param reinstall true: installed bundle is uninstalled and installed again; false: bundle must not be already installed
   */
  public void installBundle(String bundleName, boolean reinstall) throws CoreException
  {
    Reporter.startInfo("Installing bundle '" + bundleName + "'");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Installation of bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    BinBundles binBundles = rep.getBinBundles();
    BundleContents installedBundleContents = rep.getInstalledBundleContents();
    LocalInfos localInfos = rep.getLocalInfos();
    Licences licences = rep.getLicences();
    BinBundles shareClientCache = rep.getShareClientCache();
    
    synchronized (binBundles) {
    synchronized (installedBundleContents) {
    synchronized (localInfos) {
    synchronized (licences) {
    synchronized (shareClientCache)
    {
      BundleInfo binBundle = null;
      
      
      //get local info and if present, check STATE_SHARE_CLIENT state
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo != null && localInfo.areShareClientModes())
      {
        if (localInfo.isShareClient())
        {
          //get binary version of bundle from share client cache
          binBundle = shareClientCache.getBinBundle(bundleName);
          if (binBundle == null) throw new CoreException("Installation of bundle '" + bundleName + "' failed: Cannot find binary version of bundle in share client cache");
        }
        else throw new CoreException("Installation of bundle '" + bundleName + "' failed: Bundle has pre-share client active.");
      }
      else
      {
        //get binary bundle
        binBundle = binBundles.getBinBundle(bundleName);
        if (binBundle == null) throw new CoreException("Installation of bundle '" + bundleName + "' failed: Cannot find binary version of bundle");
      }
      
      
      //perform uninstall if needed
      if (isBundleInstalled(bundleName))
      {
        if (reinstall) uninstallBundle(bundleName, true, false, true);
        else throw new CoreException("Installation of bundle '" + bundleName + "' failed: Bundle is already installed");
      }

      //modify local info - if not present, create it using licences
      if (localInfo != null)
      {
        //we already have local info, so just mark it "installed"
        
        Licence licence = localInfo.getLicence();
        if (!licence.isValid()) throw new CoreException("Installation of bundle '" + bundleName + "' failed: Licence is not valid (expired)");
        
        localInfo.setInstalled(true);
        localInfo.saveToStorage();
      }
      else
      {
        //we have to build local info
        
        //try to get licence from Licences
        Licence licence = null;
        Licence lic = licences.getLicence(bundleName);

        if (lic != null)
        {
          //we have found it!
          if (!lic.isValid()) throw new CoreException("Installation of bundle '" + bundleName + "' failed: Licence is not valid (expired)");
          if (lic.getNumberOfCopies() == 0) throw new CoreException("Installation of bundle '" + bundleName + "' failed: No licence copy available in licence file");

          licence = (Licence)lic.clone();

          //take one copy
          if (licence.getNumberOfCopies() == Licence.WITHOUT_COPIES)
          {
            //ok, nothing to do
          }
          else
          {
            licence.setNumberOfCopies(1);
            lic.decreaseNumberOfCopies(1);
            lic.saveToStorage();
          }
        }
        else licence = new Licence();
        
        localInfo = localInfos.addLocalInfo(bundleName, licence, LocalInfo.STATE_INSTALLED, null, null, false, "");
      }
      
    
      //decompress binary bundle to the temp and create BundleImpl structure
      BundleImpl bundle = new BundleImpl();
      try
      {
        bundle._read(new FileInputStream(binBundle.getFile()));
      }
      catch (IOException e)
      {
        throw new CoreException("Installation of bundle '" + bundleName + "' failed while decompressing bundle", e);
      }
      

      //add installed bundle
      if (installedBundleContents.getBundleContent(bundleName) != null)
      {
        //incosistent repository - this shouldn't happen
        installedBundleContents.deleteBundleContent(bundleName);
      }

      installedBundleContents.addBundleContent(bundleName, bundle.getComponents());
        

      //store bundle to TR
      try
      {
        tr.storeBundle(bundle);
      }
      catch (TRException e)
      {
        throw new CoreException("Installation of bundle '" + bundleName + "' failed while storing bundle to TR", e);
      }
      
      bundle = null;
      
    } //shareClientCache
    } //licences
    } //localInfos
    } //installedBundleContents
    } //binBundles

    Reporter.stopInfo("Bundle '" + bundleName + "' successfully installed");
  }
  
  /**
   * Uninstalls bundle.
   *
   * @param bundleName name of bundle
   * @param deleteComponents if true (usually) the components will be deleted in TR
   * @param includingInterfaces if true, also interfaces of bundle will be removed from TR; if false just implementation
   * @param force if true, all possible mentions of installed bundle will be removed even if the repository is in the incosistent state
   */
  public void uninstallBundle(String bundleName, boolean deleteComponents, boolean includingInterfaces, boolean force) throws CoreException
  {
    Reporter.startInfo("Uninstalling bundle '" + bundleName + "'");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Uninstallation of bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    BundleContents installedBundleContents = rep.getInstalledBundleContents();
    LocalInfos localInfos = rep.getLocalInfos();
    Licences licences = rep.getLicences();
    
    synchronized (installedBundleContents) {
    synchronized (localInfos) {
    synchronized (licences)
    {
      //get local info
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo != null)
      {
        if (localInfo.isInMemory()) throw new CoreException("Uninstallation of bundle '" + bundleName + "' failed: Bundle in use (components loaded in memory)");
        
        if (!force && !localInfo.isInstalled()) throw new CoreException("Uninstallation of bundle '" + bundleName + "' failed: Bundle is not installed");
        
        localInfo.setInstalled(false);
        if (localInfo.getState() == LocalInfo.STATE_NONE)
        {
          //we have to return licences and delete local info
          Licence licence = localInfo.getLicence();
          Licence lic = licences.getLicence(bundleName);
          if (lic != null && licence.withCopies() && lic.withCopies())
          {
            //ok, licence file for bundle exists, so return licences
            lic.increaseNumberOfCopies(licence.getNumberOfCopies());
            lic.saveToStorage();
          }
          
          localInfos.deleteLocalInfo(bundleName);
        }
        else
        {
          localInfo.saveToStorage();
        }
      }
      else
      {
        if (force) Reporter.error("Incosistent repository - cannot find LocalInfo for " + bundleName);
        else throw new CoreException("Uninstallation of bundle '" + bundleName + "' failed: Incosistent repository - cannot find LocalInfo");
      }
      
      //get installed bundle
      BundleContent installedBundleContent = installedBundleContents.getBundleContent(bundleName);
      if (installedBundleContent != null)
      {
        if (deleteComponents)
        {
          //delete bundle from TR
          try
          {
            List components = installedBundleContent.getComponents();
            tr.deleteComponents((ComponentInfo[])components.toArray(new ComponentInfoImpl[components.size()]), includingInterfaces, false);
          }
          catch (TRException e)
          {
            if (force) Reporter.error("Incosistent repository - deletion of bundle " + bundleName + " from TR failed");
            else throw new CoreException("Uninstallation of bundle '" + bundleName + "' failed while deleting bundle from TR", e);
          }
        }
        
        //remove installed bundle
        installedBundleContents.deleteBundleContent(bundleName);
      }
      else
      {
        if (force) Reporter.error("Incosistent repository - cannot find BundleContent for " + bundleName);
        else throw new CoreException("Uninstallation of bundle '" + bundleName + "' failed: Incosistent repository - cannot find BundleContent");
      }
      
    } //licences
    } //localInfos
    } //installedBundleContents

    Reporter.stopInfo("Bundle '" + bundleName + "' uninstalled");
  }
  
  /**
   * Creates binary version of bundle from installed bundle.
   *
   * The actions of input triggers are then performed.
   *
   * @param bundleName name of bundle
   * @param overwrite if binary version already exists, then if true, it is overwritten, else the exception is thrown
   */
  public void createBinaryBundle(String bundleName, boolean overwrite) throws CoreException
  {
    Reporter.startInfo("Creating binary bundle '" + bundleName + "'");

    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Creation of binary bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    BinBundles binBundles = rep.getBinBundles();
    BundleContents installedBundleContents = rep.getInstalledBundleContents();
    LocalInfos localInfos = rep.getLocalInfos();
    
    synchronized (binBundles) {
    synchronized (installedBundleContents) {
    synchronized (localInfos)
    {
      //get local info and check STATE_SHARE_CLIENT state
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo == null || !localInfo.isInstalled()) throw new CoreException("Creation of binary bundle '" + bundleName + "' failed: Bundle is not installed");
      if (localInfo != null && localInfo.areShareClientModes()) throw new CoreException("Creation of binary bundle '" + bundleName + "' failed: Creation of binary version of bundle provided by share client is forbidden");
      
      //get installed bundle
      BundleContent installedBundleContent = installedBundleContents.getBundleContent(bundleName);
      if (installedBundleContent == null) throw new CoreException("Creation of binary bundle '" + bundleName + "' failed: Cannot find installed version of bundle");
      
      //get components from TR
      BundleImpl bundle = null;
      try
      {
        List components = installedBundleContent.getComponents();
        bundle = (BundleImpl) tr.getBundle(null, (ComponentInfo[])components.toArray(new ComponentInfoImpl[components.size()]), false);
      }
      catch (TRException e)
      {
        throw new CoreException("Creation of binary bundle '" + bundleName + "' failed while getting bundle from TR", e);
      }
      
      //get binary bundle
      BundleInfo binBundle = binBundles.getBinBundle(bundleName);
      if (binBundle != null)
      {
        if (!overwrite) throw new CoreException("Creation of binary bundle '" + bundleName + "' failed: Binary version of bundle already exists");
      }
      else
      {
        //create binary bundle structure
        binBundle = binBundles.addBinBundle(bundleName);
        if (binBundle == null) throw new CoreException("Creation of binary bundle '" + bundleName + "' failed: Cannot create BinBundle structure");
      }
      
      try
      {
        FileOutputStream os = new FileOutputStream(binBundle.getFile());
        bundle._write(os);
        os.close();
      }
      catch (Exception e)
      {
        binBundles.deleteBinBundle(bundleName);
        throw new CoreException("Creation of binary bundle '" + bundleName + "' failed while saving binary bundle to file", e);
      }
      
      bundle = null;
      
    } //localInfos
    } //installedBundleContents
    } //binBundles
    
    String myNodeName = NodeInfo.getLocalNodeName();
    
    netOps.performInputTriggers(bundleName, myNodeName, InputTrigger.SOURCE_MADE, false);
    
    Reporter.stopInfo("Binary version of bundle '" + bundleName + "' created");
  }

  /**
   * Creates BundleData version of bundle from installed bundle.
   * Only for internal use of SOFAnet!!!!! It should be called only if no problem is expected.
   * @param bundleName name of bundle
   * @return BundleData with installed bundle, null otherwise
   */
  BundleData recreateBundleFromTR(String bundleName)
  {
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      return null;
    }
    
    BundleContents installedBundleContents = rep.getInstalledBundleContents();
    
    synchronized (installedBundleContents)
    {
      //get installed bundle
      BundleContent installedBundleContent = installedBundleContents.getBundleContent(bundleName);
      if (installedBundleContent == null) return null;
      
      //get components from TR
      BundleImpl bundle = null;
      try
      {
        List components = installedBundleContent.getComponents();
        bundle = (BundleImpl) tr.getBundle(null, (ComponentInfo[])components.toArray(new ComponentInfoImpl[components.size()]), false);
      }
      catch (TRException e)
      {
        return null;
      }
      
      File tempFile = new File(BundleData.generateFilename());
      
      try
      {
        FileOutputStream os = new FileOutputStream(tempFile);
        bundle._write(os);
        os.close();
      }
      catch (Exception e)
      {
        bundle = null;
        return null;
      }
      
      bundle = null;

      return new BundleData(tempFile, false);
    } //installedBundleContents
  }
  
  /**
   * Deletes binary version of bundle.
   *
   * @param bundleName name of bundle
   */
  public void deleteBinaryBundle(String bundleName) throws CoreException
  {
    Reporter.startInfo("Deleting binary bundle '" + bundleName + "'");

    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Deletion of binary bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }

    BinBundles binBundles = rep.getBinBundles();
    
    synchronized (binBundles)
    {
      //get binary bundle
      BundleInfo binBundle = binBundles.getBinBundle(bundleName);
      if (binBundle == null)
      {
        throw new CoreException("Deletion of binary bundle '" + bundleName + "' failed: Binary version of bundle is not present");
      }
      
      try
      {
        boolean deleted = binBundle.getFile().delete();
        if (!deleted) throw new CoreException("Deletion of binary bundle '" + bundleName + "' failed: The fail was not deleted");
      }
      catch (Exception e)
      {
        throw new CoreException("Deletion of binary bundle '" + bundleName + "' failed while deleting the file", e);
      }
      
      binBundles.deleteBinBundle(bundleName);
      
      
    } //binBundles

    Reporter.stopInfo("Binary version of bundle '" + bundleName + "' deleted");
  }

  /**
   * Deletes binary bundle and its licence.
   * For successfull call bundle must not be installed or shared.
   * 
   * @param bundleName name of bundle
   */
  public void deleteBundle(String bundleName) throws CoreException
  {
    Reporter.startInfo("Deleting bundle '" + bundleName + "'");
    
    BinBundles binBundles = rep.getBinBundles();
    LocalInfos localInfos = rep.getLocalInfos();
    Licences licences = rep.getLicences();
  
    synchronized (binBundles) {
    synchronized (localInfos) {
    synchronized (licences)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo != null)
      {
        throw new CoreException("Deletion of bundle '" + bundleName + "' failed: Bundle is installed and/or shared");
      }

      binBundles.deleteBinBundle(bundleName);
      licences.deleteLicence(bundleName);
    } //localInfos
    } //licences
    } //binBundles

    Reporter.stopInfo("Bundle '" + bundleName + "' deleted (plus its potential licence file)");
  }
  
  
  /**
   * Creates bundle (in its installed form) from the set of components from TR.
   * The content of parameters vary according to the type of created bundle.
   * <ul>
   * <li>For user bundle: bundleName must be filled according to the user-bundle name specification, components and subsomponents - no restriction
   * <li>For single-component bundle: bundleName must be empty, components must contain one component, subcomponents must be false
   * <li>For complete-component bundle: bundleName must be empty, components must contain one component, subcomponents must be true
   * </ul>
   *
   * @param bundleName name of bundle; empty for single-component of complete-component bundle
   * @param components array of "base" components; must containt one component for single-component of complete-component bundle
   * @param subcomponents if true, also subcomponents of "base" components will be included in the bundle
   * @return name of created bundle
   */
  public String createBundle(String bundleName, ComponentInfo[] components, boolean subcomponents) throws CoreException
  {
    //test parameters
    if (components == null || components.length == 0 ||
        bundleName.length() == 0 && components.length > 1 )
    {
      throw new CoreException("Creation of bundle '" + bundleName + "' failed: Invalid parameters");
    }
    
    boolean componentBundle = false;
    
    if (bundleName.length() == 0)
    {
      componentBundle = true;
      bundleName = BundleInfo.makeComponentBundleName(components[0].getName(), components[0].getImplementationVersion(), subcomponents);
    }
    
    Reporter.startInfo("Creating bundle '" + bundleName + "'");
   
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Creation of bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    if (componentBundle != bundleInfo.isComponent())
    {
      throw new CoreException("Creation of bundle '" + bundleName + "' failed: Invalid parameters");
    }

    //get description of components from repository
    BundleImpl bundle = null;
    try
    {
      bundle = (BundleImpl) tr.getBundle(components, null, subcomponents);
    }
    catch (TRException e)
    {
      new CoreException("Creation of bundle '" + bundleName + "' failed while getting information about components from TR", e);
    }

    BundleContents installedBundleContents = rep.getInstalledBundleContents();
    LocalInfos localInfos = rep.getLocalInfos();

    synchronized (installedBundleContents) {
    synchronized (localInfos)
    {
      if (isBundleInstalled(bundleName))
      {
        throw new CoreException("Creation of bundle '" + bundleName + "' failed: Bundle is already installed");
      }
      
      //get local info and modify it (it shouldn't exist, but we are prepared)
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo != null)
      {
        if (localInfo.areShareClientModes()) new CoreException("Creation of bundle '" + bundleName + "' failed: Share client active for bundle of that name");
        localInfo.setInstalled(true);
        localInfo.saveToStorage();
      }
      else
      {
        localInfos.addLocalInfo(bundleName, new Licence(), LocalInfo.STATE_INSTALLED, null, null, false, "");
      }
      
      //get installed bundle and modify it (if it exists, it is an incosistence of repository, but be can handle it)
      //REMARK: I am not sure, whether objects components returned by TR are "local" (copies only for me
      //or "global" - also someone else can have them - therefore i rather make copies
      installedBundleContents.deleteBundleContent(bundleName); //delete it preventively
      
      ComponentInfo[] bundleComponents = bundle.getComponents();
      ComponentInfoImpl[] myComponents = new ComponentInfoImpl[bundleComponents.length];
      for (int i = 0; i < bundleComponents.length; i++)
      {
        myComponents[i] = new ComponentInfoImpl(bundleComponents[i].getName(), bundleComponents[i].getImplementationVersion());
      }
      
      installedBundleContents.addBundleContent(bundleName, myComponents);
      
    } //localInfos
    } //installedBundleContents
    
    bundle = null;
    
    Reporter.stopInfo("Bundle '" + bundleName + "' created");
    return bundleName;
  }
  
  /**
   * Creates user bundle (in its installed form) from the set of components from TR.
   *
   * @param bundleName name of bundle - according to the user-bundle name specification
   * @param components array of "base" components
   * @param subcomponents if true, also subcomponents of "base" components will be included in the bundle
   */
  public void createUserBundle(String bundleName, ComponentInfo[] components, boolean subcomponents) throws CoreException
  {
    if (bundleName.length() == 0) throw new CoreException("Creation of user bundle failed: Empty name of user bundle");
    createBundle(bundleName, components, subcomponents);
  }
  
  /**
   * Creates single-component bundle (in its installed form) from the component from TR.
   * The bundle will be named by the component and will consist only of this component
   *
   * @param component component to create bundle from
   * @return name of created bundle
   */
  public String createSingleComponentBundle(ComponentInfo component) throws CoreException
  {
    ComponentInfo[] components = new ComponentInfo[1];
    components[0] = component;
    return createBundle("", components, false);
  }

  /**
   * Creates complete-component bundle (in its installed form) from the set of components from TR.
   * The bundle will be named by the component (plus complete-component flag)
   * and will consist of this component and all its subcomponents.
   *
   * @param component component to create bundle from
   * @return name of created bundle
   */
  public String createCompleteComponentBundle(ComponentInfo component) throws CoreException
  {
    ComponentInfo[] components = new ComponentInfo[1];
    components[0] = component;
    return createBundle("", components, true);
  }

  /**
   * Creates BundleData version of bundle consisting of component from TR (and its subcomponents).
   * <p>
   * Only for internal use of SOFAnet!!!!!
   * The exception can be thrown only when an error in TR occurs or BundleData cannot be created.
   * When component is not present in TR, null is returned.
   *
   * @param component description of component
   * @param subcomponents if true, also subcomponents of component will be included in the bundle
   * @return if component is present in TR: BundleData with bundle, null otherwise
   */
  BundleData createBundleFromTR(ComponentInfo component, boolean subcomponents) throws CoreException
  {
    //get description of components from repository
    BundleImpl listBundle = null;
    try
    {
      listBundle = (BundleImpl) tr.list();
    }
    catch (TRException e)
    {
      new CoreException("Creation of bundle of component '" + component.getName() + "[" + component.getImplementationVersion() + "]' failed while listing TR", e);
    }

    if (!listBundle.contains(component)) return null;
    
    
    //get bundle from repository
    BundleImpl bundle = null;
    ComponentInfo[] components = new ComponentInfo[1];
    components[0] = component;
    try
    {
      bundle = (BundleImpl) tr.getBundle(null, components, subcomponents);
    }
    catch (TRException e)
    {
      new CoreException("Creation of bundle of component '" + component.getName() + "[" + component.getImplementationVersion() + "]' failed while getting component from TR", e);
    }

    //create file for bundleData
    File tempFile = new File(BundleData.generateFilename());

    try
    {
      FileOutputStream os = new FileOutputStream(tempFile);
      bundle._write(os);
      os.close();
    }
    catch (Exception e)
    {
      bundle = null;
      new CoreException("Creation of bundle of component '" + component.getName() + "[" + component.getImplementationVersion() + "]' failed while creating BundleData", e);
    }

    bundle = null;

    return new BundleData(tempFile, false);
  }


  /**
   * Test presence of component (and its subcomponents) in TR.
   * <p>
   * Only for internal use of SOFAnet!!!!!
   *
   * @param component description of component
   * @param subcomponents if true, also subcomponents of component will be tested
   * @return true if component is present in TR, false otherwise
   */
  boolean testComponentPresenceInTR(ComponentInfo component, boolean subcomponents)
  {
    return tr.containsComponent(component, subcomponents);
  }
  
  /**
   * List components present in TR
   * <p>
   * Only for internal use of SOFAnet!!!!!
   * @return array of components present in TR
   */
  ComponentInfo[] listComponentsPresentInTR()
  {
    return tr.list().getComponents();
  }
  
  /**
   * Notifies SOFAnet that the component (that exists in TR) is to be used by SOFA runtime.
   * The caller is also informed whether it is allowed to use component (it has valid licence etc.)
   * Every successfull call (use allowed) of this function must by accompanied by call of releaseComponent
   * <p>
   * Return codes:
   * <ul>
   * <li>   0 : use allowed - everything ok, component found
   * <li>   1 : use allowed - component not found (is not contained in the bundle)
   * <li>   2 : use not allowed - component found as part of a bundle, but the bundle is not installed (incosistence of SOFAnet repository)
   * <li>   3 : use not allowed - component found as part of a bundle, but the bundle's licence disallows loading
   * </ul>
   *
   * @param componentFullName full name of component(implementation) in form   name[version]
   * @return see return codes above
   */
  public int acquireComponent(String componentFullName)
  {
    String bundleName;
    boolean shareManager = false;
    NodeNameList shareClients = null;
    
    BundleContents installedBundleContents = rep.getInstalledBundleContents();
    CompBundleMap compBundleMap = installedBundleContents.getCompBundleMap();
    LocalInfos localInfos = rep.getLocalInfos();
    
    boolean secondRound = false;
    for (;;)
    {
      synchronized (installedBundleContents) {
      synchronized (localInfos)
      {
        //find bundle which the comonent belongs to
        bundleName = compBundleMap.findBundle(componentFullName);
        if (bundleName == null)
        {
          Reporter.info("Load of component '" + componentFullName + "' allowed (no bundle info)");
          return 1;
        }

        //get LocalInfo of the bundle
        LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
        if (localInfo == null || !localInfo.isInstalled())
        {
          Reporter.warning("Load of component '" + componentFullName + "' from bundle '" + bundleName + "' PROHIBITED: The bundle int not installed");
          return 3;
        }

        //if the bundle is already in memory, just increase counter
        if (localInfo.isInMemory())
        {
          localInfo.increaseInMemoryCounter();
          Reporter.info("Load of component '" + componentFullName + "' from bundle '" + bundleName + "' allowed (bundle already in memory)");
          return 0;
        }

        //bundle is not in the memory yet, check licence

        Licence licence = localInfo.getLicence();
        //is licence valid?
        if (!licence.isValid())
        {
          Reporter.warning("Load of component '" + componentFullName + "' from bundle '" + bundleName + "' PROHIBITED: Licence is not valid (expired)");
          return 3;
        }

        //can we grant licence immediatelly?
        if (!licence.withCopies() || licence.getNumberOfCopies() > 0)
        {
          if (licence.withCopies()) licence.decreaseNumberOfCopies(1);
          localInfo.setInMemoryCounter(1);
          localInfo.saveToStorage();
          Reporter.info("Load of component '" + componentFullName + "' from bundle '" + bundleName + "' allowed (bundle loaded to memory)");
          return 0;
        }

        //we have licence with copies but don't have free licence copy to use

        if (!secondRound && licence.getType() == Licence.TYPE_ACTIVE && ((localInfo.isShareManager() && localInfo.getSMClients().getList().size() > 0) || localInfo.isShareClient()))
        {
          //this is our last chance how to obtain licence, but we will have to ask other nodes
          if (localInfo.isShareManager())
          {
            shareManager = true;
            shareClients = (NodeNameList)localInfo.getSMClients().clone();
          }
          else
          {
            shareManager = false;
          }
        }
        else
        {
          //no chance to get licence
          Reporter.warning("Load of component '" + componentFullName + "' from bundle '" + bundleName + "' PROHIBITED: Licence copy not available");
          return 3;
        }

      } //localInfos
      } //installedBundleContents


      //now we know we have licence with active copies and there is sharing active
      if (shareManager)
      {
        //we are the share manager of the bundle

        IOParams ioParams = new IOParams();
        ioParams.setBundleName(bundleName);
        ioParams.setLicenceOnly(true);

        shareOps.makeShareClientsReturnLicence(ioParams, bundleName, null, shareClients, null, "Returning of licence to share manager for bundle");

        int errCode = ioParams.getErrCode();

        if (errCode == 2)
        {
          Reporter.warning("Load of component '" + componentFullName + "' from bundle '" + bundleName + "' PROHIBITED: Licence copy not available");
          return 3;
        }

        if (errCode != 0)
        {
          Reporter.error("Load of component '" + componentFullName + "' from bundle '" + bundleName + "' PROHIBITED: Incosistent state of share manager");
          return 3;
        }

      }
      else
      {
        //we are the share client of the bundle
        
        try
        {
          shareOps.acquireSharedBundleLicence(bundleName);
        }
        catch (CoreException e)
        {
          Reporter.error(e);
          Reporter.warning("Load of component '" + componentFullName + "' from bundle '" + bundleName + "' PROHIBITED: Acquisition of licence from share manager failed");
          return 3;
        }
      }
      
      secondRound = true;
    }
  }
  
  /**
   * Notifies SOFAnet that the component is to be released by SOFA runtime.
   * <p>
   * Return codes:
   * <ul>
   * <li>   0 : ok
   * <li>   1 : incosistence of SOFAnet repository
   * </ul>
   *
   * @param componentFullName full name of component(implementation) in form   name[version]
   * @return see return codes above
   */
  public int releaseComponent(String componentFullName)
  {
    BundleContents installedBundleContents = rep.getInstalledBundleContents();
    CompBundleMap compBundleMap = installedBundleContents.getCompBundleMap();
    LocalInfos localInfos = rep.getLocalInfos();
    
    synchronized (installedBundleContents) {
    synchronized (localInfos)
    {
      //find bundle which the comonent belongs to
      String bundleName = compBundleMap.findBundle(componentFullName);
      if (bundleName == null)
      {
        Reporter.info("Component '" + componentFullName + "' released (no bundle info)");
        return 0;
      }

      //get LocalInfo of the bundle
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo == null || !localInfo.isInstalled() || !localInfo.isInMemory())
      {
        Reporter.error("Component '" + componentFullName + "' from bundle '" + bundleName + "' released, but SOFAnet repository in incosistent state");
        return 1;
      }

      localInfo.decreaseInMemoryCounter();
      if (!localInfo.isInMemory())
      {
        //the bundle is release from memory, return licence
        Licence licence = localInfo.getLicence();
        if (licence.withCopies()) licence.increaseNumberOfCopies(1);
        localInfo.saveToStorage();
        Reporter.info("Component '" + componentFullName + "' from bundle '" + bundleName + "' released (bundle released from memory)");
      }
      else
      {
        Reporter.info("Component '" + componentFullName + "' from bundle '" + bundleName + "' released (bundle remains in memory)");
      }
      
      return 0;
      
    } //localInfos
    } //installedBundleContents
  }
  
  /**
   * Tries to provide ("download", "install",...) component that is missng in TR.
   * <p>
   * The fuction tries:
   * <ol>
   * <li>Sharing (share manager/share client)
   * <li>Binary bundles (component binary bundles)
   * <li>Bundle offers
   * <li>Search for the bundle
   * </ol>
   * <p>
   * Return codes (0 = OK, the others = error):
   * <ul>
   * <li>   0 : ok, component ready
   * <li>   1 : component marked as installed
   * <li>   2 : incosistence of SOFAnet repository
   * <li>   3 : installation of bundle failed
   * <li>   4 : cannot return shared bundle back
   * <li>   5 : local I/O error
   * <li>   6 : acquisition of shared bundle failed
   * <li>   7 : pull of bundle failed
   * <li>   8 : searching failed
   * <li>   9 : component not found
   * </ul>
   * 
   * @param componentFullName full name of component(implementation) in form   name[version]
   * @return see return codes above
   */
  public int obtainMissingComponent(String componentFullName) throws CoreException
  {
    Reporter.startInfo("Obtaining missing component '" + componentFullName + "' using SOFAnet");

    BinBundles binBundles = rep.getBinBundles();
    LocalInfos localInfos = rep.getLocalInfos();
    BundleOffers bundleOffers = rep.getBundleOffers();
    CompBundleMap offersCompBundleMap = bundleOffers.getCompBundleMap();
    BinBundles shareClientCache = rep.getShareClientCache();
    BundleContents sharedBundleContents = rep.getSharedBundleContents();
    CompBundleMap sharedCompBundleMap = sharedBundleContents.getCompBundleMap();
    
    String bundleName = "";
    boolean returnLicenceAndBundleFromShareClient = false;
    NodeNameList shareClients = null;
    boolean acquireSharedBundle = false;

    class InstallException extends Exception {}
    
    try
    {
      //first try sharing
      synchronized (binBundles) {
      synchronized (localInfos) {
      synchronized (shareClientCache) {
      synchronized (sharedBundleContents)
      {
        //we have to find name of bundle the component belongs to
        //we look at component-bundle map of sharedBundleContents, because every bundle with
        //active share manager or share client has the records here

        bundleName = sharedCompBundleMap.findBundle(componentFullName);
        if (bundleName != null)
        {
          //get LocalInfo of the bundle
          LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
          if (localInfo == null || !localInfo.isShareManager() && !localInfo.isShareClient())
          {
            Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "' failed: Missing share manager/share client state of the bundle");
            return 2;
          }

          if (localInfo.isInstalled())
          {
            Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "' failed: Bundle already marked as installed");
            return 1;
          }

          if (localInfo.isShareManager())
          {
            //a) binary bundle is present: install  (licence is solved by acquireComponent)
            //b) binary bundle is NOT present: return licence&bundle back from share client and install

            //get binary bundle
            BundleInfo binBundle = binBundles.getBinBundle(bundleName);
            if (binBundle != null)
            {
              //install
              throw new InstallException();
            }

            shareClients = (NodeNameList)localInfo.getSMClients().clone();
            if (shareClients.getList().size() > 0)
            {
              returnLicenceAndBundleFromShareClient = true;
            }
            else
            {
              Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "' failed: No share client to return licence&bundle back to share manager");
              return 4;
            }
          }
          else
          {
            //share client

            Licence licence = localInfo.getLicence();
            if (!licence.withCopies() || licence.getType() != Licence.TYPE_PHYSICAL || licence.getNumberOfCopies() > 0)
            {
              Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "' failed: The share client of bundle active but wrong licence present");
              return 2;
            }
          
            acquireSharedBundle = true;
          }
        }
        else
        {
          //now we can try binary bundles
          Iterator it = binBundles.map.values().iterator();
          while (it.hasNext())
          {
            BundleInfo bundleInfo = (BundleInfo)it.next();
            
            if (bundleInfo.isComponentBundleOf(componentFullName))
            {
              //we found the binary bundle that is the component bundle of our missing component
              bundleName = bundleInfo.getName();
              throw new InstallException();
            }
          }

          //other tries (bundle offers and search is later in the function
        }
      } //sharedBundleContents
      } //shareClientCache
      } //localInfos
      } //binBundles

      if (returnLicenceAndBundleFromShareClient)
      {
        //we are share manager and we have to get back the licence and the bundle from the share client

        IOParams ioParams = new IOParams();
        ioParams.setBundleName(bundleName);
        ioParams.setLicenceOnly(false);

        shareOps.makeShareClientsReturnLicence(ioParams, bundleName, null, shareClients, null, "Returning of licence to share manager for bundle");

        int errCode = ioParams.getErrCode();

        if (errCode == 2)
        {
          Reporter.warning("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "' failed: Cannot return licence&bundle back from share client");
          return 4;
        }

        if (errCode != 0)
        {
          Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "' failed: Incosistent state of share manager");
          return 2;
        }

        ioParams.setBundleName(bundleName);
        ioParams.setNodeName(ioParams.getAddressNodeName());
        ioParams.setPriority(IOParams.PRIORITY_BLOCKING);

        ioParams.setAddress(false);
        ioParams.setAddressNodeName("");

        try
        {
          transport.acquireShared(ioParams);
        }
        catch (TransportException e)
        {
          Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "' failed while calling acquireShared", e);
          return 4;
        }

        if (ioParams.isAddress() || ioParams.getErrCode() != 0)
        {
          Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "' failed: Unexpected result of acquireShared");
          return 4;
        }
        
        BundleData bundleData = ioParams.getBundleData();

        //store bundleData to binary bundles
        synchronized (binBundles)
        {
          BundleInfo binBundle = binBundles.getBinBundle(bundleName);
          if (binBundle == null)
          {
            binBundle = binBundles.addBinBundle(bundleName);
            if (binBundle != null)
            {
              if (!bundleData.copyToFile(binBundle.getFile(), true))
              {
                bundleData.delete();
                binBundles.deleteBinBundle(bundleName);
                Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "' failed while copying bundle to binary bundles");
                return 5;
              }
            }
            else
            {
              Reporter.warning("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "' failed: Cannot add bundle to binary bundles, bundle already present");
            }
          } 
          else
          {
            Reporter.warning("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "' failed: Binary bundle already present");
          }
        } //binBundles

        bundleData.delete();
        bundleData = null;

        throw new InstallException();
      }
      
      
      if (acquireSharedBundle)
      {
        //share client; 0 physical licences
        boolean install = false;
        try
        {
          shareOps.acquireSharedBundle(bundleName, null, false);
          install = true;
        }
        catch (CoreException e)
        {
          if (e.getMessage().endsWith("Not a share manager or access not allowed"))
          {
            install = false;
            //the curent manager is probably not the share manager of the bundle any more;
            //stop the share client and continue as if no share client existed
          }
          else
          {
            Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "': Acquisition of shared bundle failed", e);
            return 6;
          }
        }

        if (install) throw new InstallException();
        else
        {
          //stop the share client
          try
          {
            shareOps.stopShareClientOfBundle(bundleName, true, false);
          }
          catch (CoreException e)
          {
            Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "': Cannot stop obsolete share client", e);
            return 6;
          }
          
          acquireSharedBundle = false;
        }
      }
      
      
      //sharing and binary bundles has been already tried, now we continue by bundle offers
      
      boolean tryPullBundle = false;
      String offeringNodeName = "";
      
      synchronized (binBundles) {
      synchronized (bundleOffers)
      {
        bundleName = offersCompBundleMap.findBundle(componentFullName);
        if (bundleName != null)
        {
          BundleOffer bundleOffer = bundleOffers.getBundleOffer(bundleName);
          
          if (bundleOffer == null)
          {
            Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "' failed: Missing bundle offer (incosistency of component-bundle map of offers and offers themselves)");
            return 2;
          }

          //we found the offer
          
          //test wheter we don't have the bundle already
          BundleInfo binBundle = binBundles.getBinBundle(bundleName);
          if (binBundle != null)
          {
            //oh, we have the bundle
            throw new InstallException();
          }
          
          tryPullBundle = true;
          offeringNodeName = bundleOffer.getNodeName();
        }
      } //bundleOffers
      } //binBundles
      
      if (tryPullBundle)
      {
        //we will try to pull the bundle from offering node
        
        boolean canPull = false;
        try
        {
          canPull = netOps.testPull(bundleName, offeringNodeName, "");
        }
        catch (CoreException e)
        {
          Reporter.warning("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "': Test of pull acceptance failed", e);
        }
        
        if (canPull)
        {
          try
          {
            netOps.synchroPullBundle(bundleName, offeringNodeName, "");
          }
          catch (CoreException e)
          {
            Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "': Pull of bundle failed", e);
            return 6;
          }
          
          //ok, we downloaded bundle, now we can try to install it
          throw new InstallException();
        }
      }
      
      
      //now just searching remains
      
      SearchReplyItem replyItem = null;
      try
      {
        replyItem = searchOps.searchForBundleWithComponent(componentFullName);
      }
      catch (CoreException e)
      {
        Reporter.error("Obtaining of component '" + componentFullName + "failed: Searching failed", e);
        return 8;
      }
      
      if (replyItem != null)
      {
        //we have found some share manager
        
        bundleName = replyItem.getBundleName();
        
        try
        {
          shareOps.acquireSharedBundle(bundleName, replyItem.getNodeName(), true);
        }
        catch (CoreException e)
        {
          Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "': Acquisition of shared bundle failed", e);
          return 6;
        }
        
        throw new InstallException();
      }

      //component not found
      
    }
    catch (InstallException ie)
    {
      //ok, we should install bundle
      try
      {
        installBundle(bundleName, true);
      }
      catch (CoreException e)
      {
        Reporter.error("Obtaining of component '" + componentFullName + "' of bundle '" + bundleName + "' failed", e);
        return 3;
      }
      
      Reporter.stopInfo("Component '" + componentFullName + "' found in bundle '" + bundleName + "'. Bundle obtained and installed.");
      return 0;      //hurrah, we have succeeded
    }

    Reporter.stopInfo("Missing component '" + componentFullName + "' not obtained.");
    return 9; //sorry, component not found
  }
}
