/*
 * ShareOps.java
 *
 * Created on 31. bezen 2004, 17:01
 */

package SOFA.SOFAnet.Core;

import SOFA.SOFAnet.Repository.*;
import SOFA.SOFAnet.Transport.*;
import SOFA.SOFAnode.TR.ComponentInfo;
import SOFA.SOFAnode.TR.Impl.ComponentInfoImpl;
import SOFA.SOFAnode.TR.Impl.BundleImpl;
import SOFA.Util.VMProperties;
import java.util.*;
import java.io.*;

/**
 * Share operations of the SOFAnet.
 *
 * @author  Ladislav Sobr
 */
public class ShareOps
{
  private final boolean shareAllComponents;
  private Repository rep;
  private TransportInterface transport;
  private LocalOps localOps;
  
  private List bundlesForAcquire;
  
  private static class NodeBundleDataPair
  {
    public String nodeName;
    public BundleData bundleData;

    NodeBundleDataPair()
    {
    }
    
    NodeBundleDataPair(String nodeName, BundleData bundleData)
    {
      this.nodeName = nodeName;
      this.bundleData = bundleData;
    }
  }
  
  /** Creates a new instance of ShareOps */
  public ShareOps()
  {
    String shareAllComponentsStr = System.getProperty(VMProperties.NET_SHARE_ALL_COMPONENTS, VMProperties.NET_SHARE_ALL_COMPONENTS_DEF).trim().toLowerCase();
    shareAllComponents = shareAllComponentsStr.compareTo("1") == 0 || shareAllComponentsStr.compareTo("true") == 0;

    bundlesForAcquire = Collections.synchronizedList(new LinkedList());
  }

  public void setRepository(Repository repository)
  {
    rep = repository;
  }

  public void setTransportInterface(TransportInterface transport)
  {
    this.transport = transport;
  }
  
  public void setLocalOps(LocalOps localOps)
  {
    this.localOps = localOps;
  }
  
  /**
   * Returns bundle share status.
   *
   * @param bundleName name of bundle
   * @return LocalInfo.STATE_NONE or LocalInfo.STATE_SHARE_CLIENT or LocalInfo.STATE_SHARE_MANAGER
   */
  public int bundleShareStatus(String bundleName)
  {
    LocalInfos localInfos = rep.getLocalInfos();
    synchronized (localInfos)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo == null) return LocalInfo.STATE_NONE;
      return localInfo.getState() & (LocalInfo.STATE_SHARE_CLIENT | LocalInfo.STATE_SHARE_MANAGER);
    }
  }
  
  /**
   * Starts sharing of bundle.
   *
   * @param numberOfLicences positive integer or Licence.ALL_LICENCES - how many licences should be in local licence (in local info) and therefore available for sharing.
   * @param shareGroups in which share groups should be the bundle visible (can be null == all groups)
   * @param nodeFilter which nodes can access the bundle (can be null == all nodes)
   * @param reshare true: if the sharing is already active, it is updated; false: error if sharing is already active
   */
  public void shareBundle(String bundleName, int numberOfLicences, ShareGroups shareGroups, NodeNameFilter nodeFilter, boolean equality, boolean reshare) throws CoreException
  {
    if (numberOfLicences == 0) throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Invalid parameters - cannost share zero licence copies");

    boolean updating = false;
    Reporter.startInfo("Sharing of bundle '" + bundleName + "'");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    BinBundles binBundles = rep.getBinBundles();
    BundleContents installedBundleContents = rep.getInstalledBundleContents();
    LocalInfos localInfos = rep.getLocalInfos();
    Licences licences = rep.getLicences();
    BundleContents sharedBundleContents = rep.getSharedBundleContents();
    
    synchronized (binBundles) {
    synchronized (installedBundleContents) {
    synchronized (localInfos) {
    synchronized (licences) {
    synchronized (sharedBundleContents)
    {
      //get local info (if present) and check sharing states
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo != null)
      {
        if (localInfo.areShareClientModes())
        {
          throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Cannot share bundle provided by share client");
        }
        
        if (localInfo.isShareManager())
        {
          if (!reshare)
          {
            throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Tha bundle is already shared");
          }
          
          updating = true;
        }
      }
      
      //get licence file (if present)
      Licence lic = licences.getLicence(bundleName);

      //get binary bundle
      BundleInfo binBundle = binBundles.getBinBundle(bundleName);
      
      //-we don't need existence of BinBundle now, it can be created on request of sharing subsystem,
      // bu we at least check, if BinBundle or installed version of bundle exists now
      //-still it is possible to delete BinBundle and also uninstall bundle so that no
      // copy of bundle remains on the system while sharing is active
      
      if (binBundle == null && (localInfo == null || !localInfo.isInstalled()))
      {
        throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Bundle is not present on the system in any form");
      }
      
      BundleImpl bundle = null;
      
      if (binBundle != null && localInfo == null)
      {
        //decompress binary bundle to the temp and create BundleImpl structure - we will need it to create shareBundleContent
        bundle = new BundleImpl();
        try
        {
          bundle._read(new FileInputStream(binBundle.getFile()));
        }
        catch (IOException e)
        {
          throw new CoreException("Sharing of bundle '" + bundleName + "' failed while decompressing bundle", e);
        }
      }

      //get licence and check validity
      Licence licence = null;
      if (localInfo != null)
      {
        licence = localInfo.getLicence();
        if (!licence.isValid()) throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Local licence is not valid (expired)");
      }
      
      if (localInfo != null && localInfo.isInSharing() && licence.withCopies())
      {
        int sharedLicences = localInfo.getSMClients().getList().size();
        
        int localNum = licence.getNumberOfCopies();
        int totalLocalNum = sharedLicences + localNum;
        if (localInfo.isInMemory()) totalLocalNum++;
        
        if (numberOfLicences != Licence.ALL_LICENCES && totalLocalNum > numberOfLicences)
        {
          //we should return licence copies from local info to licence file
          int returnNum = totalLocalNum - numberOfLicences;
          if (returnNum > localNum)
          {
            //but there is not enough copies presently available in local info
            throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Cannot reduce number of local licence copies, too many copies in use");
          }
          else
          {
            if (lic != null && lic.withCopies())
            {
              lic.increaseNumberOfCopies(returnNum);
              lic.saveToStorage();
            }
            
            licence.decreaseNumberOfCopies(returnNum);
          }
        }
        else if (numberOfLicences == Licence.ALL_LICENCES)
        {
          //we should acquire all available licence copies from licence file
          if (lic != null)
          {
            if (!lic.isValid()) throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Licence is not valid (expired)");
            
            if (lic.withCopies())
            {
              int num = lic.getNumberOfCopies();
              lic.decreaseNumberOfCopies(num);
              lic.saveToStorage();
              
              licence.increaseNumberOfCopies(num);
            }
            else
            {
              //licence file has unlimited copies, so transfer this also to local licence
              licence.setNumberOfCopies(Licence.WITHOUT_COPIES);
            }
          }
          else
          {
            //there are no restrictions by licence file (it doesn't exist)
            licence.setNumberOfCopies(Licence.WITHOUT_COPIES);
          }
        }
        else if (totalLocalNum < numberOfLicences)
        {
          //we should acquire licence copies from licence file
          int acquireNum = numberOfLicences - totalLocalNum;
          if (lic != null)
          {
            if (!lic.isValid()) throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Licence is not valid (expired)");
            
            if (lic.withCopies())
            {
              if (lic.getNumberOfCopies() < acquireNum)
              {
                //but there is not enough copies available in licence file
                throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Not enough licence copies available in licence file");
              }

              lic.decreaseNumberOfCopies(acquireNum);
              lic.saveToStorage();
            }
          }
          licence.increaseNumberOfCopies(acquireNum);
        }
        else
        {
          //no licence copies exchange between local info and licence file
        }
        
        localInfo.setSMShareGroups(shareGroups);
        localInfo.setSMNodeFilter(nodeFilter);
        localInfo.setSMEquality(equality);
        localInfo.saveToStorage();
      }
      else if (localInfo != null && localInfo.isInSharing() && !licence.withCopies())
      {
        if (numberOfLicences != Licence.ALL_LICENCES)
        {
          //following behaviour is questionable - we restrict local licence from "without copies"
          //to concrete number of copies
          
          //because sharing is active and share manager is not responsible to maintain list
          //of all clients if licence is "without copies", therefore restricting licence to concrete
          //number of copies can be applied only to future share clients
          
          if (localInfo.isInMemory()) licence.setNumberOfCopies(numberOfLicences - 1);
          else licence.setNumberOfCopies(numberOfLicences);
        }
          
        localInfo.setSMShareGroups(shareGroups);
        localInfo.setSMNodeFilter(nodeFilter);
        localInfo.setSMEquality(equality);
        localInfo.saveToStorage();
      }
      else
      {
        //sharing is surely not active
        
        if (localInfo != null)
        {
          //-local info exists, local licence is also valid, licence != null
          //-we also know:
          //  -there is no share client
          //  -bundle is installed
          //  -if local licence is with copies, there is one copy in it (including "in memory copy")
          //-therefore we will never move licence copies from local licence to licence file
          
          
          if (licence.withCopies())
          {
            if (lic != null)
            {
              if (!lic.isValid()) throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Licence is not valid (expired)");

              if (lic.withCopies())
              {
                //transfer requested number of copies from licence file
                int licNum = lic.getNumberOfCopies();
                int licNumToTransfer;
                if (numberOfLicences == Licence.ALL_LICENCES) licNumToTransfer = licNum;
                else licNumToTransfer = numberOfLicences - 1;  //there is already one copy in local licence

                if (licNumToTransfer > licNum) throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Not enough licence copies available in licence file");

                licence.increaseNumberOfCopies(licNumToTransfer);

                lic.decreaseNumberOfCopies(licNumToTransfer);
                lic.saveToStorage();
              }
              else
              {
                if (numberOfLicences != Licence.ALL_LICENCES)
                {
                  //following behaviour is questionable - we restrict local licence from "without copies"
                  //to concrete number of copies; because sharing is not active yet, it should be ok
                  if (localInfo.isInMemory()) licence.setNumberOfCopies(numberOfLicences - 1);
                  else licence.setNumberOfCopies(numberOfLicences);
                }
                else
                {
                  //licence file is without copies, so we give "without copies" also to local licence
                  licence.setNumberOfCopies(Licence.WITHOUT_COPIES);
                }
              }
            }
            else
            {
              //licence file doesn't exist - there is no limitation on tranfering copies to local licence
              if (numberOfLicences == Licence.ALL_LICENCES) licence.setNumberOfCopies(Licence.WITHOUT_COPIES);
              else licence.increaseNumberOfCopies(numberOfLicences - 1);
            }
          }
          else
          {
            if (numberOfLicences != Licence.ALL_LICENCES)
            {
              //following behaviour is questionable - we restrict local licence from "without copies"
              //to concrete number of copies; because sharing is not active yet, it should be ok
              if (localInfo.isInMemory()) licence.setNumberOfCopies(numberOfLicences - 1);
              else licence.setNumberOfCopies(numberOfLicences);
            }
            else
            {
              //we do nothing because local licence is not limited by number of copies
            }
          }
          
          localInfo.setShareManager(true);
          localInfo.setSMShareGroups(shareGroups);
          localInfo.setSMNodeFilter(nodeFilter);
          localInfo.setSMClients(new NodeNameList());
          localInfo.setSMEquality(equality);
          localInfo.saveToStorage();
        }
        else
        {
          //we have to build local info
          
          if (lic != null)
          {
            if (!lic.isValid()) throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Licence is not valid (expired)");
            
            licence = (Licence)lic.clone();
            
            if (lic.withCopies())
            {
              //transfer requested number of copies from licence file
              int licNum = lic.getNumberOfCopies();
              int wantedNum = numberOfLicences;
              if (wantedNum == Licence.ALL_LICENCES) wantedNum = licNum;
              
              if (wantedNum > licNum) throw new CoreException("Sharing of bundle '" + bundleName + "' failed: Not enough licence copies available in licence file");
              
              licence.setNumberOfCopies(wantedNum);

              lic.decreaseNumberOfCopies(wantedNum);
              lic.saveToStorage();
            }
            else
            {
              //licence file is without copies, so leave "without copies" also in local licence
            }
          }
          else
          {
            //create empty local licence
            licence = new Licence();
          }
          
          //we have prepared local licence (licence != null), now we build local info itself
          localInfo = localInfos.addLocalInfo(bundleName, licence, LocalInfo.STATE_SHARE_MANAGER, shareGroups, nodeFilter, equality, "");
        }              

        //create shareBundleContent
        BundleContent sharedBundleContent = sharedBundleContents.getBundleContent(bundleName);
        if (sharedBundleContent != null) sharedBundleContents.deleteBundleContent(bundleName);
        
        BundleContent installedBundleContent = installedBundleContents.getBundleContent(bundleName);
        if (installedBundleContent != null)
        {
          //use installed bundle
          List components = installedBundleContent.getComponents();
          sharedBundleContents.addBundleContent(bundleName, (ComponentInfo[])components.toArray(new ComponentInfoImpl[components.size()]));
        }
        else
        {
          //use decompressed bundle
          sharedBundleContents.addBundleContent(bundleName, bundle.getComponents());
        }
      }
      
    } //sharedBundleContents
    } //licences
    } //localInfos
    } //installedBundleContents
    } //binBundles

    if (updating) Reporter.stopInfo("Sharing of bundle '" + bundleName + "' updated");
    else Reporter.stopInfo("Sharing of bundle '" + bundleName + "' started");
  }
    
  /**
   * Tests whether node (share client) can return shared bundle (it's licence) back to share manager.
   *
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  sourceNodeName   &nbsp;&nbsp;<i>      name of calling node (share manager)</i>
   * </ul>
   * <p>
   * Output: none
   *
   * @param params Input data for the call
   * @return true if shared bundle (or its licence can be returned to share manager)
   */
  public boolean canReturnSharedBundle(IOParams params) throws CoreException
  {
    String bundleName = params.getBundleName();
    String nodeName = params.getSourceNodeName();
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Return-shared-test of '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    LocalInfos localInfos = rep.getLocalInfos();
    
    synchronized (localInfos)
    {
      //get local info and check sharing state
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (nodeName.compareTo(localInfo.getSCManager()) != 0) return false;
      if (localInfo == null || !localInfo.isShareClient()) return false;
      Licence licence = localInfo.getLicence();
      if (licence.withCopies() && licence.getNumberOfCopies() == 0) return false;
      
      return true;
    } //localInfos
  }
  
  /**
   * Return shared bundle (it's licence) back to share manager (and prepare binary bundle for acquire).
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  sourceNodeName
   * <li>  address
   * <li>  addressNodeName     &nbsp;&nbsp;<i>    (if address == true)</i>
   * </ul>
   * <p>
   * Output (not in params!):
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == Not a share client
   *    <li>    2 == Cannot return bundle(licence)
   *    <li>    3 == Cannot prepare for acquire (but licence is returned)
   *    </ul>
   * </ul>
   *
   * @param params Input/Output data for the call
   * @return see errCode
   */
  public int returnSharedBundle(IOParams params) throws CoreException
  {
    String bundleName = params.getBundleName();
    String nodeName = params.getSourceNodeName();
    
    Reporter.startInfo("Returning (licence of) shared bundle '" + bundleName + "' (requesting node: '" + nodeName + "')");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Returning shared bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    boolean removeBundle = false;
    
    BundleContents installedBundleContents = rep.getInstalledBundleContents();
    LocalInfos localInfos = rep.getLocalInfos();
    Licences licences = rep.getLicences();
    BinBundles shareClientCache = rep.getShareClientCache();
    
    synchronized (installedBundleContents) {       //because of localOps.uninstallBundle
    synchronized (localInfos) {
    synchronized (licences)                 //because of localOps.uninstallBundle
    {
      //get local info and check sharing state
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo == null) return 1;
      if (nodeName.compareTo(localInfo.getSCManager()) != 0) return 1;
      if (localInfo.isSharePreClient()) return 2;
      if (!localInfo.isShareClient()) return 1;
      Licence licence = localInfo.getLicence();
      
      if (licence.withCopies())
      {
        int licNum = licence.getNumberOfCopies();
        boolean physical = licence.getType() == Licence.TYPE_PHYSICAL;
        
        if (licNum == 0) return 2;
      
        if (physical && licNum == 1)
        {
          //there is only one "physical" licence - bundle cannot stay on this node
          if (localInfo.isInstalled())
          {
            try
            {
              localOps.uninstallBundle(bundleName, true, false, false);
            }
            catch (CoreException e)
            {
              throw new CoreException("Returning shared bundle '" + bundleName + "' failed while uninstalling bundle", e);
            }
          }
          
          removeBundle = true;
          
          /* commented by Lada Sobr: we rather let share client, in the future another "acquireSharedBundle" can be done immediately on the share manager node
          //local info should contain just share client state, that we remove
          localInfos.deleteLocalInfo(bundleName);
           */
          
          localInfo = localInfos.getLocalInfo(bundleName);  //just to be sure after uninstall
          if (localInfo != null)
          {
            licence = localInfo.getLicence();
            if (licence.withCopies())
            {
              licence.decreaseNumberOfCopies(1);
              localInfo.saveToStorage();
            }
          }
        }
        else
        {
          //we have "active" licence(s) or more than one "physical" licences, so just remove one
          licence.decreaseNumberOfCopies(1);
          localInfo.saveToStorage();
        }
      }
      else
      {
        //no limitation of copies, we needn't to "change" state of this node
      }
    } //licences
    } //localInfos
    } //installedBundleContents
    
    
    //now we know the licence is removed from this node

    int retCode = 0;

    if (params.isAddress() || removeBundle)
    {
      synchronized (shareClientCache)
      {
        if (params.isAddress())
        {
          //we should prepare bundle for acquire
          BundleInfo cachedBundle = shareClientCache.getBinBundle(bundleName);
          if (cachedBundle != null)
          {
            BundleData bundleData = new BundleData(cachedBundle.getFile(), true);
            if (bundleData.getFile() != null)
            {
              //ok, we have succesfully created bundleData, so remember it in bundlesForAcquire
              bundlesForAcquire.add(new NodeBundleDataPair(params.getAddressNodeName(), bundleData));
            }
            else retCode = 3;
          }
          else retCode = 3;
        }

        if (removeBundle) shareClientCache.deleteBinBundle(bundleName);
        
      } //shareClientCache
    }

    if (retCode == 0)
    {
      if (params.isAddress()) Reporter.stopInfo("Licence of shared bundle '" + bundleName + "' returned successfully to node '" + nodeName + "', bundle is prepared for node '" + params.getAddressNodeName() + "'");
      else Reporter.stopInfo("(licence of) shared bundle '" + bundleName + "' returned successfully to node '" + nodeName + "'");
    }
    else Reporter.stopInfo("Error occured while returning shared bundle '" + bundleName + "'");
    
    return retCode;
  }
  
  /**
   * Call returnShared (on behalf of share manager) on share client that requested it.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  sourceNodeName    &nbsp;&nbsp;<i>   name of requesting node</i>
   * </ul>
   * <p>
   * Output (not in params!):
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == Not a share manager or access not allowed
   *    <li>    2 == returnShared failed
   *    </ul>
   * </ul>
   *
   * @param params Input/Output data for the call
   * @return see errCode
   */
  public int performManualReturnSharedBundle(IOParams params) throws CoreException
  {
    String bundleName = params.getBundleName();
    String nodeName = params.getSourceNodeName();

    Reporter.startInfo("Performing manual returnShared of '" + bundleName + "' on node '" + nodeName +"'");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Performing manual returnShared of '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    boolean wantBundle = false;
    
    BinBundles binBundles = rep.getBinBundles();
    LocalInfos localInfos = rep.getLocalInfos();
    
    synchronized (binBundles) {
    synchronized (localInfos)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo == null || !localInfo.isShareManager()) return 1;
      NodeNameList clients = localInfo.getSMClients();
      if (!clients.contains(nodeName)) return 1;

      //we don't test node filter (it could change)
      
      //do we need physical form of bundle? (in fact we don't need it in all situations, but we rather take it)
      BundleInfo binBundle = binBundles.getBinBundle(bundleName);
      if (binBundle == null) wantBundle = true;
      
    } //localInfos
    } //binBundles
    
    //perform returnShared itself
    final String myNodeName = NodeInfo.getLocalNodeName();

    IOParams ioParams = new IOParams();
    ioParams.setBundleName(bundleName);
    ioParams.setNodeName(nodeName);
    if (wantBundle)
    {
      ioParams.setAddress(true);
      ioParams.setAddressNodeName(myNodeName);
    }
    else
    {
      ioParams.setAddress(false);
      ioParams.setAddressNodeName("");
    }

    int errCode = -1;
    try
    {
      transport.returnShared(ioParams);
      errCode = ioParams.getErrCode();
    }
    catch (TransportException e)
    {
      Reporter.error("Performing manual returnShared of '" + bundleName + "' failed while calling returnShared: ", e);
      return 2;
    }
    
    if (errCode == 1 || errCode == 2 || errCode > 3)
    {
      String errMsg;
      switch (errCode)
      {
      case 1: errMsg = "Not a share client"; break;
      case 2: errMsg = "Cannot return bundle(licence)"; break;
      default: errMsg = "Unknown error"; break;
      }
      
      Reporter.error("Performing manual returnShared of '" + bundleName + "' failed: returnShared's result:" + errMsg + " (error code " + errCode + ")");
      return 2;
    }

    if (errCode == 3)
    {
      Reporter.error("Performing manual returnShared of '" + bundleName + "' partially failed: returnShared's result: Cannot prepare for acquire (but licence is returned) (error code 3)");
    }
    
    //errCode == 0 : OK
    //errCode == 3 : Licence returned, but physical bundle not prepared
    
    BundleData bundleData = null;
    
    if (wantBundle && errCode == 0)
    {
      try
      {
        ioParams.setBundleName(bundleName);
        ioParams.setNodeName(nodeName);
        ioParams.setPriority(IOParams.PRIORITY_BLOCKING);
        
        transport.acquireShared(ioParams);
        
        if (ioParams.isAddress() || ioParams.getErrCode() != 0)
        {
          Reporter.error("Performing manual returnShared of '" + bundleName + "' partially failed: Unexpected result of acquireShared");
        }
        else bundleData = ioParams.getBundleData();
      }
      catch (TransportException e)
      {
        Reporter.error("Performing manual returnShared of '" + bundleName + "' partially failed while calling acquireShared", e);
      }
    }

    //store bundle if needed
    if (bundleData != null)
    {
      synchronized (binBundles)
      {
        BundleInfo binBundle = binBundles.getBinBundle(bundleName);
        if (binBundle == null)
        {
          binBundle = binBundles.addBinBundle(bundleName);
          if (binBundle != null)
          {
            if (!bundleData.copyToFile(binBundle.getFile(), true))
            {
              binBundles.deleteBinBundle(bundleName);
              Reporter.error("Performing manual returnShared of '" + bundleName + "' partiallty failed while copying bundle to binary bundles");
            }
          }
          else
          {
            Reporter.error("Performing manual returnShared of '" + bundleName + "' partially failed: Cannot add bundle to binary bundles");
          }
        } 
      } //binBundles
    
      bundleData.delete();
      bundleData = null;
    }
    
    //change information of share manager
    synchronized (localInfos)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      NodeNameList clients = localInfo.getSMClients();
      if (localInfo != null && localInfo.isShareManager() && clients.contains(nodeName))
      {
        Licence licence = localInfo.getLicence();
        if (licence.withCopies()) licence.increaseNumberOfCopies(1);

        clients.remove(nodeName);
        localInfo.saveToStorage();
      }
      else
      {
        Reporter.error("Performing manual returnShared of '" + bundleName + "' partially failed: Share manager's localInfo is in incosistent state");
      }
    } //localInfos

    Reporter.stopInfo("Performing manual returnShared of '" + bundleName + "' on node '" + nodeName +"' completed");
    
    return 0;
  }
  
  /**
   * Tests whether node accepts acquire (sharing pull).
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  sourceNodeName    &nbsp;&nbsp;<i>   name of acquiring node</i>
   * </ul>
   * <p>
   * Output: none
   *
   * @param params Input data for the call
   * @return true if acquire is acceptable
   */
  public boolean isAcquisitionAcceptable(IOParams params) throws CoreException
  {
    String bundleName = params.getBundleName();
    String nodeName = params.getSourceNodeName();
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Test of acquisition acceptance of '" + bundleName + "' failed: Invalid name of bundle", e);
    }

    NodeNameList nodesForTest = null;
    
    LocalInfos localInfos = rep.getLocalInfos();
    
    synchronized (localInfos)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo == null || !localInfo.isShareManager()) return false;
      NodeNameFilter nodeFilter = localInfo.getSMNodeFilter();
      if (nodeFilter != null && !nodeFilter.pass(nodeName)) return false;
      
      Licence licence = localInfo.getLicence();
      if (!licence.withCopies() || licence.getNumberOfCopies() > 0) return true;
      
      //now we know, that licence is with copy-number limitation and therefore we have to ask current share clients
      //it must not be done in "synchronized" section
      NodeNameList clients = localInfo.getSMClients();
      if (clients == null) return false;
      nodesForTest = (NodeNameList)clients.clone();
    } //localInfos
    
    ReturnSharedAcceptanceTester tester = new ReturnSharedAcceptanceTester(bundleName, nodesForTest, transport);
    int willingNodeNum = tester.go();
    return willingNodeNum != -1;
  }
  
  /**
   * Acquire bundle from sharing manager (or share client).
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  sourceNodeName   &nbsp;&nbsp;<i>    name of acquiring node</i>
   * <li>  licenceOnly
   * </ul>
   * <p>
   * Output (on OK):
   * <ul>
   * <li>  bundleName
   * <li>  address          &nbsp;&nbsp;<i> (only if licenceOnly == false)</i>
   * <li>  bundleData       &nbsp;&nbsp;<i> (only if licenceOnly == false && address == false)</i>
   * <li>  addressNodeName  &nbsp;&nbsp;<i> (only if licenceOnly == false && address == true)</i>
   * <li>  licence          &nbsp;&nbsp;<i> (only if asking share manager)</i>
   * <li>  errCode == 0
   * </ul>
   * <p>
   * Output (on error):
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    1 == Not a share manager or access not allowed
   *    <li>    2 == Licence not available
   *    <li>    3 == Data error (binary bundle not available)  &nbsp;&nbsp;<i>   (only if licenceOnly == false)</i>
   *    <li>    4 == Incosistent state of share manager
   *    </ul>
   * </ul>
   *
   * @param params Input/Output data for the call
   */
  public void prepareBundleForAcquisition(IOParams params) throws CoreException
  {
    String bundleName = params.getBundleName();
    String nodeName = params.getSourceNodeName();
    boolean licenceOnly = params.isLicenceOnly();

    if (licenceOnly) Reporter.startInfo("Preparing licence of shared bundle '" + bundleName + "' for acquisition");
    else Reporter.startInfo("Preparing shared bundle '" + bundleName + "' for acquisition");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Preparation of bundle for acquisition '" + bundleName + "' failed: Invalid name of bundle", e);
    }

    //first test presence of nodeName in bundlesForAcquire
    synchronized (bundlesForAcquire)
    {
      Iterator it = bundlesForAcquire.iterator();
      while (it.hasNext())
      {
        NodeBundleDataPair nbdp = (NodeBundleDataPair)it.next();
        if (nbdp.nodeName.compareTo(nodeName) == 0)
        {
          params.setAddress(false);
          params.setBundleData(nbdp.bundleData);
          params.setErrCode(0);
          
          bundlesForAcquire.remove(nbdp);
          Reporter.stopInfo("Shared Bundle '" + bundleName + "' successfully acquired by node '" + nodeName + "'");
          return;
        }
      }
    }
    
    NodeNameList clients = null;
    BundleData bundleData = null;
    
    //now we are sure, we are asked as share manager, not (former) share client
    
    BinBundles binBundles = rep.getBinBundles();
    BundleContents installedBundleContents = rep.getInstalledBundleContents();
    LocalInfos localInfos = rep.getLocalInfos();

    boolean tryCreateComponentBundle = false;
    
    //test presence of localInfo
    //if not present and requested bundleName is name of component try to create bundle from repository (if it is allowed)
    synchronized (localInfos)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo == null && shareAllComponents && !licenceOnly && bundleInfo.isComponent()) tryCreateComponentBundle = true;
    } //localInfos
    
    if (tryCreateComponentBundle)
    {
      ComponentInfoImpl component = new ComponentInfoImpl(bundleInfo.getShortName(), bundleInfo.getVersion());
      boolean subcomponents = bundleInfo.isCompleteComponent();
      
      if (!localOps.testComponentPresenceInTR(component, subcomponents))
      {
        params.setErrCode(2);
        return;
      }
      
      try
      {
        bundleData = localOps.createBundleFromTR(component, subcomponents);
      }
      catch (CoreException e)
      {
        Reporter.error("Preparation of bundle for acquisition '" + bundleName + "' failed while creating bundle from TR", e);
        params.setErrCode(3);
        return;
      }
      
      if (bundleData == null)
      {
        Reporter.error("Preparation of bundle for acquisition '" + bundleName + "' failed: Cannot create \"component\" bundle from TR");
        params.setErrCode(3);
        return;
      }

      params.setLicence(new Licence());   //WE CREATE NOT LIMITED LICENCE!!!!
      params.setAddress(false);
      params.setBundleData(bundleData);
      params.setErrCode(0);
      return;
    }
    
    
    //consider regular bundles
    synchronized (binBundles) {
    synchronized (installedBundleContents) {   //because of localOps.uninstallBundle
    synchronized (localInfos)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo == null || !localInfo.isShareManager())
      {
        params.setErrCode(1);
        return;
      }

      NodeNameFilter nodeFilter = localInfo.getSMNodeFilter();
      if (nodeFilter != null && !nodeFilter.pass(nodeName))
      {
        params.setErrCode(1);
        return;
      }
      
      boolean equality = localInfo.isSMEquality();
      Licence licence = localInfo.getLicence();
      Licence acqLicence = null;
      boolean decreasedLicNum = false;
      boolean physicalDeletion = false;
      
      if (licence.withCopies())
      {
        int numLic = licence.getNumberOfCopies();
        int numOtherLic = localInfo.getSMClients().getList().size();
        
        if (numLic + numOtherLic == 0)   //no chance for licence
        {
          params.setErrCode(2);
          return;
        }
        
        if (equality)
        {
          //all licence copies can be moved from share manager
        }
        else
        {
          //at least one licence must stay at share manager
          if (!localInfo.isInMemory())      //remove one licence for "local" if it is not already used
          {
            numLic--;
            if (numLic < 0) numLic = 0; //just to be shure
          }
        }

        if (numLic > 0)
        {
          //we can give copy from share manager
          acqLicence = (Licence)licence.clone();
          acqLicence.setNumberOfCopies(1);
          licence.decreaseNumberOfCopies(1);
          localInfo.getSMClients().add(nodeName);
          localInfo.saveToStorage();
          decreasedLicNum = true;
          
          if (licence.getType() == Licence.TYPE_PHYSICAL && equality && !localInfo.isInMemory()) physicalDeletion = true;
        }
        
      }
      else
      {
        acqLicence = (Licence)licence.clone();
      }
      
      
      //if we now have acqLicence, it means, that we can give licence from share manager;
      //otherwise we have to ask share clients for licence (and bundle)
      
      if (acqLicence != null)
      {
        if (licenceOnly)     //if we are asked just for licence, return it
        {
          if (physicalDeletion)   //remove bundle if needed
          {
            try
            {
              if (localInfo.isInstalled()) localOps.uninstallBundle(bundleName, true, false, false);
              localOps.deleteBinaryBundle(bundleName);
            }
            catch (CoreException e)
            {
              Reporter.error("Preparation of bundle for acquisition '" + bundleName + "' failed while removing bundle", e);
            }
          }
          params.setLicence(acqLicence);
          params.setAddress(false);
          params.setBundleData(null);
          params.setErrCode(0);
          return;
        }
        
        //now we must prepare bundleData, physical bundle should be at the node
        BundleInfo binBundle = binBundles.getBinBundle(bundleName);

        if (binBundle != null)
        {
          bundleData = new BundleData(binBundle.getFile(), true);
        }
        else if (localInfo.isInstalled())
        {
          bundleData = localOps.recreateBundleFromTR(bundleName);
        }

        if (bundleData == null) //error
        {
          if (decreasedLicNum)  //return licences back
          {
            licence.increaseNumberOfCopies(1);
            localInfo.saveToStorage();
          }
          params.setErrCode(3);
          return;
        }


        if (physicalDeletion)   //remove bundle if needed
        {
          try
          {
            if (localInfo.isInstalled()) localOps.uninstallBundle(bundleName, true, false, false);
            localOps.deleteBinaryBundle(bundleName);
          }
          catch (CoreException e)
          {
            Reporter.error("Preparation of bundle for acquisition '" + bundleName + "' failed while removing bundle", e);
          }
        }

        params.setLicence(acqLicence);
        params.setAddress(false);
        params.setBundleData(bundleData);
        params.setErrCode(0);
        return;
      }
      
      
      //prepare clients and if possible (and requested) also bundleData for other processing
      
      clients = (NodeNameList)localInfo.getSMClients().clone();
      
      if (!licenceOnly)
      {
        BundleInfo binBundle = binBundles.getBinBundle(bundleName);

        if (binBundle != null)
        {
          bundleData = new BundleData(binBundle.getFile(), true);
        }
        else if (localInfo.isInstalled())
        {
          bundleData = localOps.recreateBundleFromTR(bundleName);
        }
      }
      
    } //localInfos
    } //installedBundleContents
    } //binBundles
    
    //if we get here:
    //1.we know, that nothing was changed (localInfo etc.) on this node and (except clients of share manager) will not change
    //2.we have to ask clients of share manager for licence + bundle
    //3.we have already prepared: clients, bundleData (if it is present on this node and was requested)

    makeShareClientsReturnLicence(params, bundleName, nodeName, clients, bundleData, "Preparation for acquisition of bundle ");
    
    if (params.getErrCode() == 0)
    {
      if (licenceOnly) Reporter.stopInfo("Licence for shared bundle '" + bundleName + "' successfully acquired by node '" + nodeName + "'");
      else Reporter.stopInfo("Shared Bundle '" + bundleName + "' successfully acquired by node '" + nodeName + "'");
    }
    else
    {
      if (licenceOnly) Reporter.stopInfo("Error occured while preparing licence of shared bundle '" + bundleName + "' for acquisition by node '" + nodeName + "'");
      else Reporter.stopInfo("Error occured while preparing shared bundle '" + bundleName + "' for acquisition by node '" + nodeName + "'");
    }
  }

  /**
   * Makes one of the share clients return licence and prepares IOParams structure for sending it to other client.
   * If nodeName == null, the function is called by share manager to return licence to it.
   * <p>
   * Input (in params):
   * <ul>
   * <li>  bundleName
   * <li>  licenceOnly     &nbsp;&nbsp;<i>  whether we request only licence or full bundle (can be provided by bundleData!)</i>
   * </ul>
   * <p>
   * Output (in params, on OK):
   * <ul>
   * <li>  bundleName
   * <li>  address         &nbsp;&nbsp;<i>  (only if licenceOnly == false)</i>
   * <li>  bundleData      &nbsp;&nbsp;<i>  (only if licenceOnly == false && address == false)</i>
   * <li>  addressNodeName &nbsp;&nbsp;<i>  (only if licenceOnly == false && address == true)</i>
   * <li>  licence         &nbsp;&nbsp;<i>  (not when returning licence to share manager)</i>
   * <li>  errCode == 0
   * </ul>
   * <p>
   * Output (in params, on error):
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    2 == Licence not available
   *    <li>    4 == Incosistent state of share manager
   *    </ul>
   * </ul>
   *
   * @param params Input/Output data for the call
   * @param bundleName name of bundle
   * @param nodeName name of asking node (the node that should acquire returned licence); null means this (local) node - licence&bundle is to be returned to share manager
   * @param client list of share clients to ask
   * @param bundleData if not null, that means the share manager has the bundle itself (deleted on error); if null share client should repare bundle if requested
   * @param errorText beggining of the text of error messages
   */
  void makeShareClientsReturnLicence(IOParams params, String bundleName, String nodeName, NodeNameList clients, BundleData bundleData, String errorText)
  {
    boolean shareManagerCalls = false;
    if (nodeName == null)
    {
      //share manager (this local node) is asking to return licence
      nodeName = NodeInfo.getLocalNodeName();
      shareManagerCalls = true;
    }
    
    LocalInfos localInfos = rep.getLocalInfos();
    
    for (int round = 0; round < 3; round++)
    {
      ReturnSharedAcceptanceTester tester = new ReturnSharedAcceptanceTester(bundleName, clients, transport);
      
      int result = tester.go();
      
      if (result == -1)
      {
        //there is no client, that could return licence
        if (bundleData != null) bundleData.delete();
        params.setErrCode(2);
        return;
      }
      
      String clientNodeName = (String)clients.getList().get(result);
      boolean wantBundle = bundleData == null && !params.isLicenceOnly();

      IOParams ioParams = new IOParams();
      ioParams.setBundleName(bundleName);
      ioParams.setNodeName(clientNodeName);
      if (wantBundle)
      {
        ioParams.setAddress(true);
        ioParams.setAddressNodeName(nodeName);
      }
      else
      {
        ioParams.setAddress(false);
        ioParams.setAddressNodeName("");
      }
      
      int errCode = -1;
      try
      {
        transport.returnShared(ioParams);
        errCode = ioParams.getErrCode();
      }
      catch (TransportException e)
      {
        Reporter.error(errorText + " '" + bundleName + "' failed while returning shared bundle (node '" + clientNodeName + "') - ONE LICENCE PROBABLY LOST!!! ", e);
        errCode = -1;
      }
      
      if (errCode < -1 || errCode > 3)
      {
        Reporter.error(errorText + " '" + bundleName + "' failed: Unknown error code of returning shared bundle (node '" + clientNodeName + "'): " + errCode + " - ONE LICENCE PROBABLY LOST!!!");
        errCode = -1;
      }

      //errCodes:
      //    -1 == exception occured
      //     0 == OK
      //     1 == Not a share client
      //     2 == Cannot return bundle(licence)
      //     3 == Cannot prepare for acquire (but licence is returned)
      
      switch (errCode)
      {
      case 1:
        Reporter.error(errorText + " '" + bundleName + "' failed while returning shared bundle (node '" + clientNodeName + "'): The node rejects to be share client, ONE LICENCE PROBABLY LOST!!!!");
        break;
      case 3:
        Reporter.error(errorText + " '" + bundleName + "' failed while returning shared bundle (node '" + clientNodeName + "'): The node cannot prepare physical bundle");
        break;
      }
      
      if (errCode != 2)
      {
        synchronized (localInfos)
        {
          LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
          if (localInfo == null || !localInfo.isShareManager())
          {
            //oh no, someone is working with our localInfo - we have too incosistent state :-(((
            if (errCode == 0 || errCode == 3)
            {
              Reporter.error(errorText + " '" + bundleName + "' failed: Incosistent LocalInfo, ONE LICENCE PROBABLY LOST!!!!");
            }
            else
            {
              Reporter.error(errorText + " '" + bundleName + "' failed: Incosistent LocalInfo");
            }
            if (bundleData != null) bundleData.delete();
            params.setErrCode(4);
            return;
          }

          Licence licence = localInfo.getLicence();
          List clientList = localInfo.getSMClients().getList();
          Licence newLicence = null;
          
          switch (errCode)
          {
          case -1:
          case 1:
            //remove the clientNodeName from clients of share manager
            //remove the clientNodeName from "clients" - not to try it again
            clientList.remove(clientNodeName);
            localInfo.saveToStorage();
            clients.getList().remove(result);
            break;
          case 3:
            //remove the clientNodeName from clients of share manager
            //add return licence to localInfo
            //remove the clientNodeName from "clients" - not to try it again
            clientList.remove(clientNodeName);
            if (licence.withCopies()) licence.increaseNumberOfCopies(1);
            localInfo.saveToStorage();
            clients.getList().remove(result);
            break;
          case 0:
            //hurrah, we found the client node who accepted our request without any troubles
            //remove the clientNodeName from clients of share manager
            //add the nodeName to clients of share manager
            
            clientList.remove(clientNodeName);
            if (shareManagerCalls)
            {
              //return licence back to share manager
              if (licence.withCopies()) licence.increaseNumberOfCopies(1);
              localInfo.saveToStorage();

              params.setLicence(null);
            }
            else
            {
              clientList.add(nodeName);
              localInfo.saveToStorage();
              
              newLicence = (Licence)licence.clone();
              if (licence.withCopies()) newLicence.setNumberOfCopies(1);

              params.setLicence(newLicence);
            }
            

            if (params.isLicenceOnly())
            {
              params.setAddress(false);
              params.setAddressNodeName("");
              params.setBundleData(null);
            }
            else
            {
              if (wantBundle)
              {
                params.setAddress(true);
                params.setAddressNodeName(clientNodeName);
                params.setBundleData(null);
              }
              else
              {
                params.setAddress(false);
                params.setAddressNodeName("");
                params.setBundleData(bundleData);
              }
            }
            params.setErrCode(0);
            return;
          }
        } //localInfos
      }
      else
      {
        //just remove the clientNodeName from "clients" - not to try it again
        clients.getList().remove(result);
      }
    }
    
    //huh, we didn't find any client node that would return it's licence successfully
    if (bundleData != null) bundleData.delete();
    params.setErrCode(2);
  }
  
  /**
   * Stop sharing the bundle.
   *
   * @param bundleName name of bundle
   * @param recoverLicences if true, try to recover licences from share clients
   * @param force true: sharing is stopped even if share manager fails to recover all licence copies from clients (licence copies get lost); false: error if share manager fails to recover all licences
   */
  public void stopSharingBundle(String bundleName, boolean recoverLicences, boolean force) throws CoreException
  {
    Reporter.startInfo("Stopping sharing of bundle '" + bundleName + "'");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Stopping of sharing of bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    LocalInfos localInfos = rep.getLocalInfos();
    Licences licences = rep.getLicences();
    BundleContents sharedBundleContents = rep.getSharedBundleContents();

    //try to recover licences if we are requested to do it
    if (recoverLicences)
    {
      for (;;)
      {
        NodeNameList clients = null;

        synchronized (localInfos)
        {
          //get local info (if present) and check sharing states
          LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
          if (localInfo != null && localInfo.isShareManager())
          {
            clients = (NodeNameList)localInfo.getSMClients().clone();
          }
        } //localInfos

        if (clients == null || clients.getList().size() == 0) break;
        
        IOParams ioParams = new IOParams();
        
        Iterator it = clients.getList().iterator();
        while (it.hasNext())
        {
          String nodeName = (String)it.next();
          
          ioParams.setBundleName(bundleName);
          ioParams.setSourceNodeName(nodeName);
          
          try
          {
            performManualReturnSharedBundle(ioParams);
          }
          catch (CoreException e)
          {
            Reporter.error("Stopping of sharing of bundle '" + bundleName + "' failed while recovering licences from share clients", e);
          }
          
          //we don't check error codes
        }
        
        break;
      }
    }
    
    synchronized (localInfos) {
    synchronized (licences) {
    synchronized (sharedBundleContents)
    {
      //get local info (if present) and check sharing states
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo == null || !localInfo.isShareManager())
      {
        throw new CoreException("Stopping of sharing of bundle '" + bundleName + "' failed: Bundle is not shared");
      }
      
      if (localInfo.getSMClients().getList().size() != 0 && !force)
      {
        throw new CoreException("Stopping of sharing of bundle '" + bundleName + "' failed: Cannot recover all licences");
      }
      
      localInfo.setShareManager(false);
      localInfo.setSMShareGroups(null);
      localInfo.setSMNodeFilter(null);
      localInfo.setSMClients(null);
      
      //delete content information of the shared bundle
      sharedBundleContents.deleteBundleContent(bundleName);
      
      
      //get licence file (if present)
      Licence lic = licences.getLicence(bundleName);

      //get local licence
      Licence licence = localInfo.getLicence();
      
      //find out number of licences to return to licence file (if any)
      int numToReturn = Licence.WITHOUT_COPIES;
      if (licence.withCopies())
      {
        if (localInfo.isInstalled())
        {
          if (localInfo.isInMemory()) numToReturn = licence.getNumberOfCopies();
          else numToReturn = licence.getNumberOfCopies() - 1;
          licence.decreaseNumberOfCopies(numToReturn);
        }
        else
        {
          numToReturn = licence.getNumberOfCopies();
          licence.setNumberOfCopies(0); //just to be sure
        }
      }
      
      //return licences to licence file
      if (numToReturn != Licence.ALL_LICENCES && lic != null && lic.withCopies())
      {
        lic.increaseNumberOfCopies(numToReturn);
        lic.saveToStorage();
      }
      
      if (localInfo.getState() == LocalInfo.STATE_NONE) localInfos.deleteLocalInfo(bundleName);
      else localInfo.saveToStorage();
    
    } //sharedBundleContents
    } //licences
    } //localInfos
    
    Reporter.stopInfo("Sharing of bundle '" + bundleName + "' stopped");
  }

  
  /**
   * Acquires bundle from remote node's share manager (or share-client if we are relocated)
   * and stores it in share client cache.
   *
   * There are two possibilities (situations) how to call this function:
   * <ul>
   * <li>a) Bundle is not present on the node in any form. Requirements:
   *    <ul>
   *    <li>nodeName parameter specifies node with share manager (that should be tried)
   *    <li>localInfo must not exist (bundle must not be installed, shared or in share-client mode).
   *    </ul>
   * <li>b) Bundle has share client active, but budle was returned to share manager and physicaly removed from the node. Requirements:
   *    <ul>
   *    <li>nodeName is null; the share manager node will be taken from share client info
        <li>share client must be active for physical licences and 0 copies
   *    <li>bundle muset not be installed
   *    </ul>
   * </ul>
   * <p>
   * In b) if the acquisition of bundle failes, the share client of the bundle is stopped.
   *
   * @param bundleName name of bundle
   * @param nodeName name of sharing node ; can be null!!! see previous info
   */
  public void acquireSharedBundle(String bundleName, String nodeName, boolean tryNoShareClient) throws CoreException
  {
    if (nodeName != null) Reporter.startInfo("Acquiring shared bundle '" + bundleName + "' from node '" + nodeName + "'");
    else Reporter.startInfo("Acquiring shared bundle '" + bundleName + "'");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    if (nodeName != null)
    {
      NodeInfo nodeInfo = new NodeInfo();
      try
      {
        nodeInfo.setNodeName(nodeName);
      }
      catch (NodeInfo.InvalidNodeNameException e)
      {
        throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Invalid name of SOFA node: " + nodeName, e);
      }
    }
    
    BinBundles binBundles = rep.getBinBundles();
    LocalInfos localInfos = rep.getLocalInfos();
    Licences licences = rep.getLicences();
    BinBundles shareClientCache = rep.getShareClientCache();
    BundleContents sharedBundleContents = rep.getSharedBundleContents();
    
    //test presence of local info
    synchronized (localInfos)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      
      if (nodeName != null)
      {
        // call variant a) - "no share client"
        if (localInfo != null)
        {
          throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Bundle is already present (locally - installed/shared)");
        }

        localInfos.addLocalInfo(bundleName, new Licence(), LocalInfo.STATE_SHARE_PRECLIENT, null, null, false, "");
      }
      else
      {
        // call variant b) - "share client but no physical bundle"
        if (localInfo == null || localInfo.isInstalled() || !localInfo.isShareClient())
        {
          throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Share client not active or bundle is already present (locally - installed/shared)");
        }
        Licence licence = localInfo.getLicence();
        if (!licence.withCopies() || licence.getType() != Licence.TYPE_PHYSICAL || licence.getNumberOfCopies() > 0)
        {
          throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: A physical licence is already present");
        }
        
        nodeName = localInfo.getSCManager();
        localInfo.setState(LocalInfo.STATE_SHARE_PRECLIENT);
        localInfo.saveToStorage();
      }
    } //localInfos

    //if share client cache contains the bundle, delete it
    synchronized (shareClientCache)
    {
      BundleInfo cachedBundle = shareClientCache.getBinBundle(bundleName);
      if (cachedBundle != null) shareClientCache.deleteBinBundle(bundleName);
    } //shareClientCache
    
    //contact remote node's share manager
    IOParams ioParams = new IOParams();
    ioParams.setBundleName(bundleName);
    ioParams.setPriority(IOParams.PRIORITY_BLOCKING);
    ioParams.setNodeName(nodeName);
    ioParams.setLicenceOnly(false);

    try
    {
      transport.acquireShared(ioParams);
    }
    catch (TransportException e)
    {
      localInfos.deleteLocalInfo(bundleName);      //remove share_preclient state
      throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed while acquiring bundle at share manager", e);
    }

    int errCode = ioParams.getErrCode();
    if (errCode != 0)
    {
      localInfos.deleteLocalInfo(bundleName);      //remove share_preclient state
      switch (errCode)
      {
      case 1: throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Share manager returns: Not a share manager or access not allowed"); //DO NOT CHANGE THIS STRING - IT IS USED TO RECOGNIZE THIS TYPE OF ERROR!!!!!
      case 2: throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Share manager returns: Licence not available");
      case 3: throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Share manager returns: Data error (binary bundle not available)");
      case 4: throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Share manager returns: Incosistent state of share manager");
      default: throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Transport layer failed");
      }
    }
    
    if (ioParams.isAddress())
    {
      //binary bundle must be acquired from another share client

      IOParams ioParams2 = new IOParams();
      ioParams2.setBundleName(bundleName);
      ioParams2.setPriority(IOParams.PRIORITY_BLOCKING);
      ioParams2.setNodeName(ioParams.getAddressNodeName());

      //contact share client
      try
      {
        transport.acquireShared(ioParams2);
      }
      catch (TransportException e)
      {
        localInfos.deleteLocalInfo(bundleName);      //remove share_preclient state
        throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed while acquiring binary bundle at share client (" + ioParams.getNodeName() + ")", e);
      }

      int errCode2 = ioParams2.getErrCode();
      if (errCode2 != 0)
      {
        localInfos.deleteLocalInfo(bundleName);      //remove share_preclient state
        switch (errCode2)
        {
        case 1: throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Share client returns: Not a share manager or access not allowed");
        case 2: throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Share client returns: Licence not available");
        case 3: throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Share client returns: Data error (binary bundle not available)");
        case 4: throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Share client returns: Incosistent state of share manager");
        default: throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Transport layer failed");
        }
      }
      
      ioParams.setBundleData(ioParams2.getBundleData());
    }
    else
    {
      if (ioParams.getBundleData() == null) throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: acqureShared returned corrupted output (missing BundleData)");
    }
    
    //now we have in ioParams BundleData as well as licence
    
    Licence licence = ioParams.getLicence();
    BundleData bundleData = ioParams.getBundleData();
    
    if (bundleData == null)
    {
      localInfos.deleteLocalInfo(bundleName);      //remove share_preclient state
      throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Binary version of bundle is not available");
    }
    
    boolean done = false;
    
    //if it is requested and possible, try to store bundle to binary bundles and not to run share client
    if (tryNoShareClient && (licence == null || !licence.withCopies()))
    {
      synchronized (binBundles) {
      synchronized (localInfos) {
      synchronized (licences)
      {
        //test presence of share_preclient state
        LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
        if (localInfo == null || localInfo.getState() != LocalInfo.STATE_SHARE_PRECLIENT)
        {
          bundleData.delete();
          throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Incosistent state of LocalInfo - share preclient not found");
        }
        
        localInfos.deleteLocalInfo(bundleName);      //remove share_preclient state
        localInfo = null;
        
        BundleInfo binBundle = binBundles.getBinBundle(bundleName);
        Licence lic = licences.getLicence(bundleName);
        
        if (binBundle == null && lic == null)
        {
          //copy bundle to binary bundles
          binBundle = binBundles.addBinBundle(bundleName);
          if (binBundle == null)
          {
            bundleData.delete();
            throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Cannot add bundle to binary bundles");
          }

          if (!bundleData.copyToFile(binBundle.getFile(), true))
          {
            bundleData.delete();
            binBundles.deleteBinBundle(bundleName);
            throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Error occured while copying bundle to binary bundles");
          }

          bundleData.delete();
          bundleData = null;

          //add licence file
          if (licence != null)
          {
            lic = licences.addLicence(bundleName, licence.getType(), licence.getNumberOfCopies(),
                                      licence.getValidFrom(), licence.getValidTo(), licence.getText());
            if (lic == null)
            {
              binBundles.deleteBinBundle(bundleName);
              throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Cannot add licence for bundle");
            }
          }
          
          done = true;
        }
      } //licences
      } //localInfos
      } //binBundles
    }
    
    if (!done)
    {
      //run share client
      
      BundleImpl bundle = null;
      
      //decompress binary bundle to the temp and create BundleImpl structure - we will need it to create shareBundleContent
      bundle = new BundleImpl();
      try
      {
        bundle._read(new FileInputStream(bundleData.getFile()));
      }
      catch (IOException e)
      {
        throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed while decompressing bundle", e);
      }

      
      synchronized (localInfos) {
      synchronized (shareClientCache) {
      synchronized (sharedBundleContents)
      {
        //test presence of share_preclient state
        LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
        if (localInfo == null || localInfo.getState() != LocalInfo.STATE_SHARE_PRECLIENT)
        {
          bundleData.delete();
          throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Incosistent state of LocalInfo - share preclient not found");
        }
        
        localInfos.deleteLocalInfo(bundleName);      //remove share_preclient state
        localInfo = null;

        //if share client cache contains the bundle, delete it (we must do it again)
        BundleInfo cachedBundle = shareClientCache.getBinBundle(bundleName);
        if (cachedBundle != null) shareClientCache.deleteBinBundle(bundleName);
        
        //copy bundle to share client cache
        cachedBundle = shareClientCache.addBinBundle(bundleName);
        if (cachedBundle == null)
        {
          bundleData.delete();
          throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Cannot add bundle to share client cache");
        }

        if (!bundleData.copyToFile(cachedBundle.getFile(), true))
        {
          bundleData.delete();
          shareClientCache.deleteBinBundle(bundleName);
          throw new CoreException("Acquiring shared bundle '" + bundleName + "' failed: Error occured while copying bundle to share client cache");
        }
        
        bundleData.delete();
        bundleData = null;
        
        //create local info
        localInfos.addLocalInfo(bundleName, licence, LocalInfo.STATE_SHARE_CLIENT, null, null, false, nodeName);
        
        
        //create shareBundleContent
        BundleContent sharedBundleContent = sharedBundleContents.getBundleContent(bundleName);
        if (sharedBundleContent != null) sharedBundleContents.deleteBundleContent(bundleName);
        sharedBundleContents.addBundleContent(bundleName, bundle.getComponents());
        
      } //sharedBundleContents
      } //shareClientCache
      } //localInfos
    }
    
    if (nodeName != null) Reporter.stopInfo("Shared bundle '" + bundleName + "' acquired from node '" + nodeName + "'");
    else Reporter.stopInfo("Shared bundle '" + bundleName + "' acquired");
  }


  /**
   * Acquires licence for bundle from remote node's share manager
   * <p>
   * For successfull call, share-client must be already running.
   *
   * @param bundleName name of bundle
   */
  public void acquireSharedBundleLicence(String bundleName) throws CoreException
  {
    Reporter.startInfo("Acquiring licence for shared bundle '" + bundleName + "'");
    
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Acquiring licence for shared bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }

    String manager = "";
    
    LocalInfos localInfos = rep.getLocalInfos();
    
    //find out name of node with share manager
    synchronized (localInfos)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo == null || !localInfo.isShareClient())
      {
        throw new CoreException("Acquiring licence for shared bundle '" + bundleName + "' failed: Bundle is not in share-client mode");
      }
      
      manager = localInfo.getSCManager();
    } //localInfos

    if (manager.length() == 0)
    {
      throw new CoreException("Acquiring licence for shared bundle '" + bundleName + "' failed: Empty name of node with share manager");
    }
    

    //contact remote node's share manager
    IOParams ioParams = new IOParams();
    ioParams.setBundleName(bundleName);
    ioParams.setPriority(IOParams.PRIORITY_BLOCKING);
    ioParams.setNodeName(manager);
    ioParams.setLicenceOnly(false);

    try
    {
      transport.acquireShared(ioParams);
    }
    catch (TransportException e)
    {
      throw new CoreException("Acquiring licence for shared bundle '" + bundleName + "' failed while acquiring licence at share manager", e);
    }

    int errCode = ioParams.getErrCode();
    if (errCode != 0)
    {
      switch (errCode)
      {
      case 1: throw new CoreException("Acquiring licence for shared bundle '" + bundleName + "' failed: Share manager returns: Not a share manager or access not allowed");
      case 2: throw new CoreException("Acquiring licence for shared bundle '" + bundleName + "' failed: Share manager returns: Licence not available");
      case 3: throw new CoreException("Acquiring licence for shared bundle '" + bundleName + "' failed: Share manager returns: Data error (binary bundle not available)");
      case 4: throw new CoreException("Acquiring licence for shared bundle '" + bundleName + "' failed: Share manager returns: Incosistent state of share manager");
      default: throw new CoreException("Acquiring licence for shared bundle '" + bundleName + "' failed: Transport layer failed");
      }
    }
    
    //store new licence file to localInfo
    synchronized (localInfos)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo == null || !localInfo.isShareClient())
      {
        throw new CoreException("Acquiring licence for shared bundle '" + bundleName + "' failed: Bundle is not in share-client mode. ACQUIRED LICENCE IS LOST!!!!!!");
      }
      
      localInfo.setLicence(ioParams.getLicence());
    } //localInfos
    
    Reporter.stopInfo("Licence for shared bundle '" + bundleName + "' acquired from node '" + manager + "'");
  }

  /**
   * Returns missing bundle back to this node that is share manager.
   * <p>
   * The function is the equivalent of acquireSharedBundle(bundleName, null, false) for share manager. 
   * For successfull call, bundle must be shared (the node must be the share manager of the bundle) and
   * bundle must be physically missing on the node (must not be installed or binary bundle present) -
   * in fact bundle must have licence of "physical type" and 0 copies present.
   *
   * @param bundleName name of bundle
   */
  public void makeShareClientsReturnBundle(String bundleName) throws CoreException
  {
    NodeNameList shareClients = null;
    
    BinBundles binBundles = rep.getBinBundles();
    LocalInfos localInfos = rep.getLocalInfos();
    
    //find out the state of bundle and clients of the share manager
    synchronized (localInfos)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      
      if (localInfo == null || !localInfo.isShareManager())
      {
        throw new CoreException("Making share clients return bundle '" + bundleName + "' failed: Bundle is not shared (the node is not the Share manager of the bundle)");
      }
      
      shareClients = (NodeNameList)localInfo.getSMClients().clone();
      if (shareClients.getList().size() == 0)
      {
        throw new CoreException("Making share clients return bundle '" + bundleName + "' failed: No share client to return licence&bundle back to share manager");
      }
    } //localInfos

    //get back the licence and the bundle from the share client
    IOParams ioParams = new IOParams();
    ioParams.setBundleName(bundleName);
    ioParams.setLicenceOnly(false);

    makeShareClientsReturnLicence(ioParams, bundleName, null, shareClients, null, "Returning of licence to share manager for bundle");

    int errCode = ioParams.getErrCode();

    if (errCode == 2) throw new CoreException("Making share clients return bundle '" + bundleName + "' failed: Cannot return licence&bundle back from share client");
    if (errCode != 0) throw new CoreException("Making share clients return bundle '" + bundleName + "' failed: Incosistent state of share manager");

    
    ioParams.setBundleName(bundleName);
    ioParams.setNodeName(ioParams.getAddressNodeName());
    ioParams.setPriority(IOParams.PRIORITY_BLOCKING);

    ioParams.setAddress(false);
    ioParams.setAddressNodeName("");

    try
    {
      transport.acquireShared(ioParams);
    }
    catch (TransportException e)
    {
      throw new CoreException("Making share clients return bundle '" + bundleName + "' failed while calling acquireShared", e);
    }

    if (ioParams.isAddress() || ioParams.getErrCode() != 0)
    {
      throw new CoreException("Making share clients return bundle '" + bundleName + "' failed: Unexpected result of acquireShared");
    }

    BundleData bundleData = ioParams.getBundleData();

    //store bundleData to binary bundles
    synchronized (binBundles)
    {
      BundleInfo binBundle = binBundles.getBinBundle(bundleName);
      if (binBundle == null)
      {
        binBundle = binBundles.addBinBundle(bundleName);
        if (binBundle != null)
        {
          if (!bundleData.copyToFile(binBundle.getFile(), true))
          {
            bundleData.delete();
            binBundles.deleteBinBundle(bundleName);
            throw new CoreException("Making share clients return bundle '" + bundleName + "' failed while copying bundle to binary bundles");
          }
        }
        else
        {
          throw new CoreException("Making share clients return bundle '" + bundleName + "' failed: Cannot add bundle to binary bundles, bundle already present");
        }
      } 
      else
      {
        throw new CoreException("Making share clients return bundle '" + bundleName + "' failed: Binary bundle already present");
      }
    } //binBundles

    bundleData.delete();
    bundleData = null;
  }
  
  /**
   * Stop share client of the bundle.
   * <p>
   * For successfull call, bundle must not be installed
   *
   * @param bundleName name of bundle
   * @param returnLicence if true, try to return licence to share manager
   * @param force true: share client is stopped even if licence is not returned to share manager; false: error (and do not stop share client) if licence cannot be returned
   */
  public void stopShareClientOfBundle(String bundleName, boolean returnLicence, boolean force) throws CoreException
  {
    Reporter.startInfo("Stopping share client of bundle '" + bundleName + "'");

    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      throw new CoreException("Stopping share client of bundle '" + bundleName + "' failed: Invalid name of bundle", e);
    }
    
    String manager;
    
    LocalInfos localInfos = rep.getLocalInfos();
    BinBundles shareClientCache = rep.getShareClientCache();
    BundleContents sharedBundleContents = rep.getSharedBundleContents();
    
    synchronized (localInfos)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo == null || !localInfo.isShareClient())
      {
        throw new CoreException("Stopping share client of bundle '" + bundleName + "' failed: Bundle is not in share-client mode");
      }

      if (localInfo.isInstalled())
      {
        throw new CoreException("Stopping share client of bundle '" + bundleName + "' failed: Bundle is installed");
      }
      
      Licence licence = localInfo.getLicence();
      
      if (!licence.withCopies() || licence.getNumberOfCopies() == 0 || !returnLicence)
      {
        //we can do everything now
        
        if (licence.withCopies() && licence.getNumberOfCopies() > 0 && !returnLicence) 
        {
          if (force) Reporter.error("Stopping share client of bundle '" + bundleName + "': Not returning licence to share manager - ONE LICENCE LOST!");
          else throw new CoreException("Stopping share client of bundle '" + bundleName + "' failed: The shared licence copies are still present (not returned to share manager)");
        }
        
        synchronized (shareClientCache) {
        synchronized (sharedBundleContents)
        {
          sharedBundleContents.deleteBundleContent(bundleName);
          shareClientCache.deleteBinBundle(bundleName);
        } //sharedBundleContents
        } //shareClientCache
        
        localInfos.deleteLocalInfo(bundleName);
        
        Reporter.stopInfo("Share client of bundle " + bundleName + " successfully stopped");
        return;
      }
      
      manager = localInfo.getSCManager();
    } //localInfos
    
    
    //now we know that we have to return licences to share manager first

    IOParams ioParams = new IOParams();
    ioParams.setBundleName(bundleName);
    ioParams.setNodeName(manager);

    try
    {
      transport.manualReturnShared(ioParams);
    }
    catch (TransportException e)
    {
      throw new CoreException("Stopping share client of bundle '" + bundleName + "' failed while acquiring licence at share manager", e);
    }
    
    int errCode = ioParams.getErrCode();
    if (errCode != 0)
    {
      switch (errCode)
      {
      case 1: throw new CoreException("Stopping share client of bundle '" + bundleName + "' failed: Share manager returns: Not a share manager or access not allowed");
      case 2: throw new CoreException("Stopping share client of bundle '" + bundleName + "' failed: Share manager returns: returnShared failed");
      default: throw new CoreException("Stopping share client of bundle '" + bundleName + "' failed: Transport layer failed");
      }
    }

    //stop the share client
    synchronized (localInfos)
    {
      LocalInfo localInfo = localInfos.getLocalInfo(bundleName);
      if (localInfo != null && localInfo.isShareClient())
      {
        if (localInfo.isInstalled())
        {
          throw new CoreException("Stopping share client of bundle '" + bundleName + "' failed: Bundle is installed");
        }
        
        synchronized (shareClientCache) {
        synchronized (sharedBundleContents)
        {
          sharedBundleContents.deleteBundleContent(bundleName);
          shareClientCache.deleteBinBundle(bundleName);
        } //sharedBundleContents
        } //shareClientCache

        localInfos.deleteLocalInfo(bundleName);
      }
    } //localInfos
    
    Reporter.stopInfo("Share client of bundle '" + bundleName + "' successfully stopped");
  }
}
