/*
 * NetNode.java
 *
 * Created on 3. kvten 2004, 16:56
 */

package SOFA.SOFAnet;

import SOFA.SOFAnet.Repository.Repository;
import SOFA.SOFAnet.Repository.NodeInfo;
import SOFA.SOFAnet.TR.TRInterface;
import SOFA.SOFAnet.TR.Local.TRWrapper;
import SOFA.SOFAnode.TR.TRException;
import SOFA.SOFAnet.Core.*;
import SOFA.SOFAnet.Transport.TransportInterface;
import SOFA.SOFAnet.Transport.RMI.RMITransportCenter;
import SOFA.SOFAnet.Transport.RMI.RMITransportException;
import SOFA.SOFAnet.Local.RMI.RMILocalCenter;
import SOFA.SOFAnet.Local.RMI.RMILocalException;
import SOFA.SOFAnet.Admin.RMI.RMIAdminCenter;
import SOFA.SOFAnet.Admin.RMI.RMIAdminException;
import SOFA.SOFAnet.Search.SearchInterface;
import SOFA.SOFAnet.Search.RMI.RMISearchCenter;
import SOFA.SOFAnet.Search.RMI.RMISearchException;
import SOFA.Util.VMProperties;
import java.rmi.RMISecurityManager;
import java.io.File;
import java.net.InetAddress;
import java.net.UnknownHostException;

/**
 * Instance of SOFAnet node with:
 * <ul>
 * <li>local Template Repository (this JVM)
 * <li>RMI transport layer
 * <li>RMI search subsystem
 * <li>RMI interface for SOFA runtime ("Local interface")
 * <li>RMI admin interface for Node Browser ("Admin interface")
 * </ul>
 *
 * @author  Ladislav Sobr
 */
public class NetNode
{
  private Repository rep;
  private TRInterface tr;
  private TransportInterface transport;
  private SearchInterface search;
  private LocalOps localOps;
  private NetOps netOps;
  private ShareOps shareOps;
  private SearchOps searchOps;
  
  private RMITransportCenter rmiTransportCenter;
  private RMISearchCenter rmiSearchCenter;
  private RMILocalCenter rmiLocalCenter;
  private RMIAdminCenter rmiAdminCenter;
  
  /**
   * Creates a new instance of SOFA net node.
   *
   * @param infoBaseDir SOFAnet repository: directory for general information (must exist)
   * @param binBundlesDir SOFAnet repository: directory for binary bundles (must exist)
   * @param shareClientCacheDir SOFAnet repository: directory for binary bundles in cache of share client (must exist)
   */
  public NetNode(File trDir, File infoBaseDir, File binBundlesDir, File shareClientCacheDir)
  {
    System.setSecurityManager(new RMISecurityManager());
    
    try
    {
      rep = new Repository(infoBaseDir, binBundlesDir, shareClientCacheDir);
      tr = new TRWrapper(trDir.getPath());
      
      localOps = new LocalOps();
      netOps = new NetOps();
      shareOps = new ShareOps();
      searchOps = new SearchOps();
      
      rmiTransportCenter = new RMITransportCenter();
      rmiSearchCenter = new RMISearchCenter();
      rmiLocalCenter = new RMILocalCenter();
      rmiAdminCenter = new RMIAdminCenter();
      
      transport = rmiTransportCenter.getTransportInterface();
      search = rmiSearchCenter.getSearchInterface();
      
      localOps.setRepository(rep);
      localOps.setTRInterface(tr);
      localOps.setNetOps(netOps);
      localOps.setShareOps(shareOps);
      localOps.setSearchOps(searchOps);
      localOps.setTransportInterface(transport);
      
      netOps.setRepository(rep);
      netOps.setLocalOps(localOps);
      netOps.setShareOps(shareOps);
      netOps.setTransportInterface(transport);
      
      shareOps.setRepository(rep);
      shareOps.setLocalOps(localOps);
      shareOps.setTransportInterface(transport);
      
      searchOps.setRepository(rep);
      searchOps.setLocalOps(localOps);
      searchOps.setSearchInterface(search);
      
      rmiTransportCenter.setNetOps(netOps);
      rmiTransportCenter.setShareOps(shareOps);
      
      rmiSearchCenter.setRepository(rep);
      rmiSearchCenter.setSearchOps(searchOps);
      
      rmiLocalCenter.setLocalOps(localOps);

      rmiAdminCenter.setRepository(rep);
      rmiAdminCenter.setTRInterface(tr);
      rmiAdminCenter.setLocalOps(localOps);
      rmiAdminCenter.setNetOps(netOps);
      rmiAdminCenter.setShareOps(shareOps);
      rmiAdminCenter.setSearchOps(searchOps);
      rmiAdminCenter.setRMISearchCenter(rmiSearchCenter);
    }
    catch (TRException e)
    {
      Reporter.error(e);
    }
  }

  public void start()
  {
    Reporter.startInfo("Starting SOFAnet node " + NodeInfo.getLocalNodeName());
    boolean ok = false;
    
    try
    {
      rmiTransportCenter.start();
      rmiSearchCenter.start();
      rmiLocalCenter.start();
      rmiAdminCenter.start();
      ok = true;
    }
    catch (RMITransportException e)
    {
      Reporter.error(e);
    }
    catch (RMISearchException e)
    {
      Reporter.error(e);
    }
    catch (RMILocalException e)
    {
      Reporter.error(e);
    }
    catch (RMIAdminException e)
    {
      Reporter.error(e);
    }
    
    if (ok) Reporter.stopInfo("SOFAnet node successfully started");
  }
  
  /**
   * Main for starting SOFAnet node.
   *
   * @param args the command line arguments
   */
  public static void main(String[] args)
  {
    String nodeName = System.getProperty(VMProperties.NODE_NAME, null);
    String trDir = System.getProperty(VMProperties.TR_DIR, null);
    String infoBaseDir = System.getProperty(VMProperties.NET_INFO_BASE_DIR, null);
    String binBundlesDir = System.getProperty(VMProperties.NET_BIN_BUNDLES_DIR, null);
    String shareClientCacheDir = System.getProperty(VMProperties.NET_SHARE_CLIENT_CACHE_DIR, null);
    String showDetailedExceptionsString = System.getProperty(VMProperties.NET_SHOW_DETAILED_EXCEPTIONS, "false");

    if (nodeName == null || nodeName.trim().length() == 0)
    {
      try
      {
        nodeName = InetAddress.getLocalHost().getHostName();
      }
      catch (UnknownHostException e)
      {
        System.err.println("Can't get the host name.\nSet the name to the property '" + VMProperties.NODE_NAME + "' when NetNode is being started.");
        System.exit(1);
      }
      
      String rmiPort = System.getProperty(VMProperties.RMI_PORT, null);
      
      if (rmiPort != null) nodeName += ":" + rmiPort;
      
      System.out.println("The property '" + VMProperties.NODE_NAME + "' has not been specified, using '" + nodeName + "'");
      System.setProperty(VMProperties.NODE_NAME, nodeName);
    }

    if (trDir == null)
    {
      System.out.println("The property '" + VMProperties.TR_DIR + "' has not been specified.");
      System.exit(1);
    }

    if (infoBaseDir == null)
    {
      System.out.println("The property '" + VMProperties.NET_INFO_BASE_DIR + "' has not been specified.");
      System.exit(1);
    }

    if (binBundlesDir == null)
    {
      System.out.println("The property '" + VMProperties.NET_BIN_BUNDLES_DIR + "' has not been specified.");
      System.exit(1);
    }

    if (shareClientCacheDir == null)
    {
      System.out.println("The property '" + VMProperties.NET_SHARE_CLIENT_CACHE_DIR + "' has not been specified.");
      System.exit(1);
    }

    if (showDetailedExceptionsString.compareTo("1") == 0 || showDetailedExceptionsString.compareTo("true") == 0)
    {
      Reporter.setShowDetailedExceptions(true);
    }
    
    NetNode netNode = new NetNode(new File(trDir), new File(infoBaseDir), new File(binBundlesDir), new File(shareClientCacheDir));
    
    netNode.start();
  }
}
