/*
 * BundleInfo.java
 *
 * Created on 22. listopad 2003, 20:56
 */

package SOFA.SOFAnet.Repository;

import java.io.*;

/**
 * Representation of bundle name and/or file with binary version of bundle
 *
 * @author  Ladislav Sobr
 */
public class BundleInfo implements StorageItem
{
  private File file;
  private String shortName;        // name without version
  private String version;
  private String fullName;         // == shortName + "[" + version + "]";
  private String fileName;
  private String [] nameParts;
  private boolean component;
  private boolean completeComponent;
  
  public static class InvalidBundleNameException extends SOFA.SOFAnode.SOFAException
  {
    public InvalidBundleNameException(String message)
    {
      super(message);
    }

    public InvalidBundleNameException(String message, Throwable cause)
    {
      super(message, cause);
    }
    
    public InvalidBundleNameException(Throwable cause)
    {
      super(cause);
    }
  }
  
  /** Creates a new instance of BundleInfo */
  public BundleInfo()
  {
  }
  

  public void setFile(File file)
  {
    this.file = file;
  }
  
  public File getFile()
  {
    return file;
  }

  public String getShortName()
  {
    return shortName;
  }
  
  public String getVersion()
  {
    return version;
  }
  
  public String getFullName()
  {
    return fullName;
  }
  
  public String getFileName()
  {
    return fileName;
  }
  
  public String [] getNameParts()
  {
    return nameParts;
  }
  
  public boolean isComponent()
  {
    return component;
  }
  
  public boolean isCompleteComponent()
  {
    return completeComponent;
  }
  
  public boolean isSingleComponent()
  {
    return component && !completeComponent;
  }
  
  /**
   * @param name Name of bundle in filesystem without ".jar" extension.
   */
  public void fromFilename(String name) throws InvalidBundleNameException
  {
    if (name.length() <= 0) throw new InvalidBundleNameException("Empty bundle name");
    
    fileName = name;
    
    if (name.charAt(0) == '-')           //name begins with "-"
    {
      component = true;
      name = name.substring(1);
      if (name.length() > 0 && name.charAt(name.length() - 1) == '$')       //name ends with "-$"
      {
        completeComponent = true;
        name = name.substring(0, name.length() - 2);
      }
      else completeComponent = false;
    }
    else
    {
      component = false;
      completeComponent = false;
    }
    
    int index = name.lastIndexOf('-');
    
    if (index == -1) throw new InvalidBundleNameException("Missing bundle version in " + fileName);
    
    version = name.substring(index + 1);
    name = name.substring(0, index);
    
    if (name.length() <= 0) throw new InvalidBundleNameException("Missing bundle short name in " + fileName);
    if (component) shortName = "::" + name.replaceAll("-", "::");
    else shortName = name;
    
    nameParts = name.split("-");
    
    fullName = shortName + "[" + version + "]";
    if (completeComponent) fullName += "$";
  }
  
  public void fromBundleName(String name) throws InvalidBundleNameException
  {
    if (name.length() <= 0) throw new InvalidBundleNameException("Empty bundle name");
  
    fullName = name;
    
    if (name.startsWith("::"))
    {
      component = true;
      name = name.substring(2);
      if (name.length() <= 0) throw new InvalidBundleNameException("Invalid bundle name: " + fullName);
      if (name.charAt(name.length() - 1) == '$')
      {
        name = name.substring(0, name.length() - 1);
        completeComponent = true;
      }
      else completeComponent = false;
    }
    else
    {
      component = false;
      completeComponent = false;
    } 

    if (name.charAt(name.length() - 1) != ']') throw new InvalidBundleNameException("Invalid version in bundle name: " + fullName);
    name = name.substring(0, name.length() - 1);

    int index = name.lastIndexOf('[');
    if (index == -1 || index == name.length() - 1) throw new InvalidBundleNameException("Missing version in bundle name: " + fullName);
    version = name.substring(index + 1);

    name = name.substring(0, index);
    
    if (component)
    {
      shortName = "::" + name;
      name = name.replaceAll("::", "-");
    }
    else
    {
      shortName = "";
    }
    
    nameParts = name.split("-");
      
    if (component)
    {
      fileName = "-" + name + "-" + version;
      if (completeComponent) fileName += "-$";
    }
    else
    {
      fileName = name + "-" + version;
    }
  }  
  
  /**
   * Returns true when specified pattern match bundle name stored inside this BundleInfo.
   * <p>
   * <ul>
   * <li>Pattern structure:
   * <ul>
   * <li>short_name[version]complete_component
   * <li>* &nbsp;&nbsp;&nbsp; matches any bundle
   * </ul>
   *
   * <li>short_name:
   *   <ul>
   *   <li>::str1::str2::...::strN       for components
   *   <li>str1-str2-...-strN            for other bundles
   *   <li>where:
   *     <ul>
   *     <li>strX can be
   *       <ul>
   *       <li>a) text that doesn't contain "characters" ?, ::, -, *
   *         <ul>
   *         <li>exact match of "pattern text" and "real name text" is required
   *         </ul>
   *       <li>b) character ?
   *         <ul> 
   *         <li>"real name text" can be arbitrary
   *         </ul>
   *       </ul>
   *     <li>short_name can end by character * at any place
   *       <ul>
   *       <li>from this point up to the end od short name "real name text" can be arbitrary
   *       <li>examples:
   *         <ul>
   *         <li>::CUNI::*
   *         <li>::CUNI::SO*       (::CUNI::SOFA::xxx matches this pattern)
   *         </ul>
   *       </ul>
   *     </ul>
   *   </ul>
   *
   *
   * <li>version:
   *   <ul>
   *   <li>a) &gt;pattern_text
   *     <ul>
   *     <li>"real version text" must be lexicographicaly greater than pattern_text
   *     </ul>
   *   <li>b) &gt;=pattern_text
   *     <ul>
   *     <li>"real version text" must be lexicographicaly greater or equal than pattern_text
   *     </ul>
   *   <li>c) &lt;pattern_text
   *     <ul>
   *     <li>"real version text" must be lexicographicaly less than pattern_text
   *     </ul>
   *   <li>d) &lt;=pattern_text
   *     <ul>
   *     <li>"real version text" must be lexicographicaly less or equal than pattern_text
   *     </ul>
   *   <li>e) * or ?
   *     <ul>
   *     <li>"real version text" can be arbitrary
   *     </ul>
   *   <li>f) pattern_text    which doesnt match any of a) - e)
   *     <ul>
   *     <li>exact match of pattern_text and "real version text" is required
   *     </ul>
   *   </ul>
   *
   * <li>complete_component
   *   <ul>
   *   <li>a) "empty"
   *     <ul>
   *     <li>bundle name must not represent complete-component bundle
   *     </ul>
   *   <li>b) $
   *     <ul>
   *     <li>bundle name must represent complete-component bundle
   *     </ul>
   *   <li>c) ? or *
   *     <ul>
   *     <li>bundle name can represent any "kind" of bundle
   *     </ul>
   *   </ul>
   * </ul>
   * <p>
   *
   * Remark: Don't use this function and exception InvalidBundleNameException to
   * test correctness of the pattern format. The function can return result
   * "false" even for badly formatted pattern!
   *
   * @param bundleNamePattern Pattern of bundle name.
   */
  public boolean matchPattern(String bundleNamePattern) throws InvalidBundleNameException
  {
    String name = bundleNamePattern.trim();
    if (name.compareTo("*") == 0) return true;
    
    //find split pattern into short name, version and after-version
    int index1 = name.lastIndexOf('[');
    int index2 = name.lastIndexOf(']');
    if (index1 == -1 || index2 == -1 || index2 < index1) throw new InvalidBundleNameException("Cannot find version in bundle name pattern: " + bundleNamePattern);

    int len;
    char ch, ch2;

    String shortNamePattern = name.substring(0, index1);
    
    //check short name
    if (shortNamePattern.startsWith("::"))
    {
      if (!component) return false;
      shortNamePattern = shortNamePattern.substring(2);
      shortNamePattern = shortNamePattern.replaceAll("::", "-");
    }
    else
    {
      if (component && shortNamePattern.compareTo("*") != 0) return false;
    }
    
    String [] patternParts = shortNamePattern.split("-");
    
    boolean asterisk = false; 
    
    for (int i = 0; i < patternParts.length; i++)
    {
      len = patternParts[i].length();
      if (len == 0) return false;
      if (nameParts.length <= i) return false;
      ch = patternParts[i].charAt(0);
      if (ch == '?')
      {
        if (len == 2 && patternParts[i].charAt(1) == '*')
        {
          asterisk = true;
          break;
        }
        if (len > 1) throw new InvalidBundleNameException("Error in bundle name pattern \"" + bundleNamePattern + "\" at short-name part");
        else continue;
      }
      int index = patternParts[i].indexOf('*');
      if (index != -1 && index != patternParts[i].length() - 1 ||
          patternParts[i].indexOf('?') != -1) throw new InvalidBundleNameException("Error in bundle name pattern \"" + bundleNamePattern + "\" at short-name part");
      if (index == -1)
      {
        if (nameParts[i].compareTo(patternParts[i]) != 0) return false;
      }
      else
      {
        if (nameParts[i].length() < index) return false;
        if (nameParts[i].substring(0, index).compareTo(patternParts[i].substring(0, index)) != 0) return false;
        asterisk = true;
        break;
      }
    }
    
    if (!asterisk && patternParts.length != nameParts.length) return false;
    
    
    //check version
    String versionPattern = name.substring(index1 + 1, index2);
    len = versionPattern.length();
    ch = len > 0 ? versionPattern.charAt(0) : 'x';
    ch2 = len > 1 ? versionPattern.charAt(1) : 'x';
    if (len == 0 ||
        len == 1 && (ch == '<' || ch == '>') ||
        len == 2 && (ch == '<' || ch == '>') && ch2 == '=' ||
        len > 1 && (ch == '?' || ch == '*'))
        throw new InvalidBundleNameException("Error in bundle name pattern \"" + bundleNamePattern + "\" at version part");
    
    if (len == 1)
    {
      if (ch != '?' && ch != '*' && version.compareTo(versionPattern) != 0) return false;
    }
    else //len > 1
    {
      if (ch == '<' && ch2 != '=')
      {
        if (version.compareTo(versionPattern.substring(1)) >= 0) return false;
      }
      else
      if (ch == '<' && ch2 == '=')
      {
        if (version.compareTo(versionPattern.substring(2)) > 0) return false;
      }
      else
      if (ch == '>' && ch2 != '=')
      {
        if (version.compareTo(versionPattern.substring(1)) <= 0) return false;
      }
      else
      if (ch == '>' && ch2 == '=')
      {
        if (version.compareTo(versionPattern.substring(2)) < 0) return false;
      }
      else
      {
        if (version.compareTo(versionPattern) != 0) return false;
      }
    }

    //check after-version
    String afterVersionPattern = name.substring(index2 + 1);
    len = afterVersionPattern.length();
    ch = len > 0 ? afterVersionPattern.charAt(0) : 'x';
    if (len > 1 || len == 1 && ch != '$' && ch != '*' && ch != '?') throw new InvalidBundleNameException("Error in bundle name pattern \"" + bundleNamePattern + "\" at complete-component part");
    
    if (len == 0 && completeComponent) return false;
    if (len == 1 && ch == '$' && (!component || !completeComponent)) return false;
    
    return true;
  }
  
  public boolean isComponentBundleOf(String componentFullName)
  {
    return component && componentFullName.compareTo(shortName + "[" + version + "]") == 0;
  }
  
  public boolean isLoaded()
  {
    return false;
  }
  public void setLoaded(boolean loaded)
  {
  }
  
  public boolean toBeDeleted()
  {
    return true;
  }

  public void loadFromStorage()
  {
  }
  
  public void saveToStorage()
  {
  }

  public String getName()
  {
    return fullName;
  }

  public void deleteFromStorage()
  {
  }
  
  public static String fileNameToBundleName(String fileName)
  {
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromFilename(fileName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      return null;
    }
    return bundleInfo.getFullName();
  }
  
  public static String bundleNameToFileName(String bundleName)
  {
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      return null;
    }
    return bundleInfo.getFileName();
  }
  
  public static String makeComponentBundleName(String nameOfComponent, String version, boolean isCompleteComponent)
  {
    if (isCompleteComponent) return nameOfComponent + "[" + version + "]$";
    else return nameOfComponent + "[" + version + "]";
  }
  
  public static String makeUserBundleName(String name, String version)
  {
    return name + "[" + version + "]";
  }

  public static boolean matchPattern(String bundleName, String bundleNamePattern) throws InvalidBundleNameException
  {
    BundleInfo bundleInfo = new BundleInfo();
    bundleInfo.fromBundleName(bundleName);
    return bundleInfo.matchPattern(bundleNamePattern);
  }
  
}
