/*
 * BundleNameFilter.java
 *
 * Created on 23. listopad 2003, 22:45
 */

package SOFA.SOFAnet.Repository;

import java.util.*;
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import SOFA.Util.XML;
import java.io.Serializable;

/**
 * Filter of bundle names.
 * 
 * Filter is implemnted as list of patterns.
 * If name of bundle passes any of the patterns, it passes whole filter
 *
 * @author  Ladislav Sobr
 */
public class BundleNameFilter implements Cloneable, Serializable
{
  private List patterns;
  
  /** Creates a new instance of BundleNameFilter */
  public BundleNameFilter()
  {
    patterns = Collections.synchronizedList(new LinkedList());
  }

  public Object clone()
  {
    BundleNameFilter clone = null;
    try
    { 
      clone = (BundleNameFilter)super.clone();
    }
    catch (CloneNotSupportedException e)
    { 
      throw new InternalError();
    }
    clone.patterns = Collections.synchronizedList(new LinkedList());
    clone.patterns.addAll(patterns);
    return clone;
  }
  
  public List getPatterns()
  {
    return patterns;
  }
  
  public boolean pass(String bundleName)
  {
    BundleInfo bundleInfo = new BundleInfo();
    try
    {
      bundleInfo.fromBundleName(bundleName);
    }
    catch (BundleInfo.InvalidBundleNameException e)
    {
      return false;
    }
    
    return pass(bundleInfo);
  }
  
  public boolean pass(BundleInfo bundleInfo)
  {
    synchronized (patterns)
    {
      Iterator it = patterns.iterator();
      while (it.hasNext())
      {
        String pattern = (String)it.next();
        try
        {
          if (bundleInfo.matchPattern(pattern)) return true;
        }
        catch (BundleInfo.InvalidBundleNameException e)
        {
        }
      }
      return false;
    }
  }

  public void add(String pattern)
  {
    patterns.add(pattern);
  }

  public void add(BundleNameFilter bundleNameFilter)
  {
    if (bundleNameFilter != null)
    {
      List li = bundleNameFilter.getPatterns();
      synchronized (li)
      {
        Iterator it = li.iterator();
        while (it.hasNext())
        {
          String pattern = (String)it.next();
          patterns.add(pattern);
        }
      }
    }
  }
  
  /**
   * Loads the BundleNameFilter from the XML DOM tree.
   *
   * XML format:
   * <p>
   * <pre>
   *  *&lt;filter pattern="string"/&gt;
   * </pre>
   */
  public void loadFromXML(Element element) throws SAXException
  {
    synchronized (patterns)
    {
      NodeList nl = element.getChildNodes();
      for (int i = 0; i < nl.getLength(); i++)
      {
        Node node = nl.item(i);
        if (node.getNodeType() == Node.ELEMENT_NODE && node.getNodeName().compareTo("filter") == 0)
        {
          Element e = (Element)node;
          String pattern = e.getAttribute("pattern");
          if (pattern.length() == 0) throw new SAXException("Missing 'pattern' attribute in 'filter' tag");
          patterns.add(pattern);
        }
      }
    }
  }
  
  /**
   * Saves the BundleNameFilter to XML DOM tree.
   */
  public void saveToXML(Element element)
  {
    synchronized (patterns)
    {
      Iterator it = patterns.iterator();
      while (it.hasNext())
      {
        String pattern = (String)it.next();
        XML.newOneAttributeElement(element, "filter", "pattern", pattern);
      }
    }
  }
  
}
