/*
 * BundleOffer.java
 *
 * Created on 19. bezen 2004, 20:32
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;
import SOFA.SOFAnode.TR.Impl.ComponentInfoImpl;
import SOFA.SOFAnode.TR.Impl.BundleImpl;
import SOFA.SOFAnet.Transport.BundleData;
import SOFA.Util.XML;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

/**
 * Representation of the bundle offer.
 * One item of persistent storage.
 * <p>
 * The bundle offer contains:
 * <ul>
 *  <li>name of node that can provide the bundle (that is offering the bundle)
 *  <li>list of components contained in the bundle
 * </ul>
 *
 * @author  Ladislav Sobr
 */
public class BundleOffer extends XMLStorageItem implements StorageItem, Serializable
{
  private String nodeName;
  private List components;
  
  /** Creates a new instance of BundleOffer */
  public BundleOffer(String name, File file)
  {
    super(name, file, "bundle_offer");
    components = Collections.synchronizedList(new LinkedList());
  }
  
  public List getComponents()
  {
    return components;
  }

  /**
   * Resets content of the item.
   */
  protected void reset()
  {
    nodeName = "";
    components.clear();
  }
  
  /**
   * Loads the item from the XML DOM tree.
   * <p>
   * XML format of storage item:
   * <p>
   * <pre>
   * &lt;bundle_offer&gt;
   *   BundleOffer XML format
   * &lt;/bundle_offer&gt;
   * </pre>
   * <p>
   * XML format:
   * <p>
   * <pre>
   *   &lt;node name="string"&gt;
   *  *&lt;component name="string" implementation_version="string"/&gt;
   * </pre>
   *
   */
  public void loadFromXML(Element element) throws SAXException
  {
    synchronized (components)
    {
      components.clear();
      NodeList nl = element.getChildNodes();
      for (int i = 0; i < nl.getLength(); i++)
      {
        Node node = nl.item(i);
        if (node.getNodeType() == Node.ELEMENT_NODE)
        {
          Element el = (Element)node;
          String nodeN = node.getNodeName();

          
          if (nodeN.compareTo("node") == 0)
          {
            nodeName = el.getAttribute("name");
          }
          else
          if (nodeN.compareTo("component") == 0)
          {
            String name = el.getAttribute("name");
            String version = el.getAttribute("implementation_version");
            if (name.length() == 0) throw new SAXException("Missing 'name' attribute in 'component' tag");
            if (version.length() == 0) throw new SAXException("Missing 'implementation_version' attribute in 'component' tag");
            components.add(new ComponentInfoImpl(name, version));
          }
        }
      }
    }
  }
  
  /**
   * Saves the item to XML DOM tree.
   */
  public void saveToXML(Element element)
  {
    
    synchronized (components)
    {
      Document doc = element.getOwnerDocument();
      Element el;
      
      el = doc.createElement("node");
      el.setAttribute("name", nodeName);
      element.appendChild(el);
      
      Iterator it = components.iterator();
      while (it.hasNext())
      {
        ComponentInfoImpl ci = (ComponentInfoImpl)it.next();
        el = doc.createElement("component");
        el.setAttribute("name", ci.getName());
        el.setAttribute("implementation_version", ci.getImplementationVersion());
        element.appendChild(el);
      }
    }      
  }
  

  public String getNodeName()
  {
    return nodeName;
  }
  
  public void setNodeName(String nodeName)
  {
    this.nodeName = nodeName;
  }
  
  
  /**
   * Generates BundleData structure (with "binary bundle file" in temp)
   *
   * @return null on error
   */
  
  public BundleData generateBundleData()
  {
    BundleImpl bundleOff = new BundleImpl();
    synchronized (components)
    {
      Iterator it = components.iterator();
      while (it.hasNext())
      {
        ComponentInfoImpl ci = (ComponentInfoImpl)it.next();
        bundleOff.add(new ComponentInfoImpl(ci.getName(), ci.getImplementationVersion()), BundleImpl.NO_FILES);
      }
    }

    File tempFile = new File(BundleData.generateFilename());

    try
    {
      FileOutputStream os = new FileOutputStream(tempFile);
      bundleOff._write(os);
      os.close();
    }
    catch (Exception e)
    {
      bundleOff = null;
      return null;
    }

    bundleOff = null;

    return new BundleData(tempFile, false);
  }
  
}
