/*
 * Contract.java
 *
 * Created on 3. prosinec 2003, 17:19
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;
import java.text.*;
import SOFA.Util.XML;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

/**
 * Representation of the contract.
 * One item of persistent storage.
 * <p>
 * The contract contains:
 * <ul>
 * <li>ID of the contract (randomly generated string)
 * <li>text description of the contract
 * <li>time validity of contract
 * <li>default specifications of thing that needn't be that repeated in every contract rule:
 *   <ul>
 *   <li>filter of "pushing" nodes (contract is taken only if the bundle arrives from such node)
 *   <li>destination nodes (and number of licences) for push operations
 *   <li>filter of nodes that can pull
 *   <li>licence text
 *   </ul>
 * <li>list of contract rules (see ContractRule)
 * </ul>
 * <p>
 * Contracts creates licences (licence files).
 * "Push/Pull" actions of contracts are realized by input and output triggers connected with the contract.
 * These triggers are generated from contract (manually).
 *
 * @author  Ladislav Sobr
 */
public class Contract extends XMLStorageItem implements StorageItem, Serializable
{
  private String contractID;
  private String description;
  private Date validFrom;         //can be null
  private Date validTo;           //can be null
  private boolean autoDelete;
  private NodeNameLicList defaultPushDestinationNodes;
  private NodeNameFilter defaultPushNodeFilter;
  private NodeNameFilter defaultPullNodeFilter;
  private String defaultLicenceText;
  private Map rules;
  
  /** Creates a new instance of Contract */
  public Contract(String name, File file)
  {
    super(name, file, "contract");
    reset();
  }
  
  public Contract()
  {
    super("", null, "contract");
    reset();
  }
  
  /**
   * Resets content of the item.
   */
  protected void reset()
  {
    contractID = "";
    description = "";
    validFrom = null;
    validTo = null;
    autoDelete = false;
    defaultPushDestinationNodes = null;
    defaultPushNodeFilter = null;
    defaultPullNodeFilter = null;
    defaultLicenceText = "";
    rules = Collections.synchronizedMap(new HashMap());
  }
  
  public boolean toBeDeleted()
  {
    Iterator it = rules.values().iterator();
    while (it.hasNext())
    {
      ContractRule rule = (ContractRule)it.next();
      if (rule.toBeDeleted()) it.remove();
    }
    
    return autoDelete && (
       validTo != null && validTo.before(Calendar.getInstance().getTime()) ||
       rules.isEmpty());
  }
  
  /**
   * Loads the item from the XML DOM tree.
   * <p>
   * XML format of storage item:
   * <p>
   * <pre>
   * &lt;contract&gt;
   *   Contract XML format
   * &lt;/contract&gt;
   * </pre>
   * <p>
   * WARNING!!! Name of storage item must be identical to "id" (contractID)!!!
   * <p>
   * XML format:
   * <p>
   * <pre>
   *   &lt;id&gt;string&lt;/id&gt;
   *  ?&lt;descripton&gt;string&lt;/description&gt;
   *  ?&lt;validity ?from="date" ?to="date" ?autodelete="boolean"/&gt;
   *
   *  ?&lt;defaults&gt;
   *    ?&lt;push_node_filter&gt;
   *       NodeNameFilter XML format
   *     &lt;/push_node_filter&gt;
   *    ?&lt;push_destination_nodes&gt;
   *       NodeNameLicList XML format      (presence of "number_of_licences" overrides "number_of_copies" in licence of rule)
   *     &lt;/push_destination_nodes&gt;
   *    ?&lt;pull_node_filter&gt;
   *       NodeNameFilter XML format
   *     &lt;/pull_node_filter&gt;
   *    ?&lt;licence_text&gt;string&lt;/licence_text&gt;
   *   &lt;/defaults&gt;
   *
   *   &lt;rules&gt;
   *    *&lt;rule name="string"&gt;
   *        ContractRule XML format
   *     &lt;/rule&gt;
   *   &lt;/rules&gt;
   * </pre>
   *
   */
  public void loadFromXML(Element element) throws SAXException
  {
    reset();
    NodeList nl = element.getChildNodes();
    for (int i = 0; i < nl.getLength(); i++)
    {
      Node node = nl.item(i);
      if (node.getNodeType() == Node.ELEMENT_NODE)
      {
        Element el = (Element)node;
        String nodeName = node.getNodeName();
        
        if (nodeName.compareTo("id") == 0)
        {
          contractID = XML.getTextContent(el);
        }
        else
        if (nodeName.compareTo("description") == 0)
        {
          description = XML.getTextContent(el);
        }
        else
        if (nodeName.compareTo("validity") == 0)
        {
          String from = el.getAttribute("from");
          String to = el.getAttribute("to");
          String ad = el.getAttribute("autodelete");
          DateFormat dateFormat = DateFormat.getInstance();
          if (from.length() > 0)
          {
            try
            {
              validFrom = dateFormat.parse(from);
            }
            catch (ParseException e)
            {
              throw new SAXException("Cannot parse date/time format of 'from' attribute of 'validity' tag", e);
            }
          }
          
          if (to.length() > 0)
          {
            try
            {
              validTo = dateFormat.parse(to);
            }
            catch (ParseException e)
            {
              throw new SAXException("Cannot parse date/time format of 'to' attribute of 'validity' tag", e);
            }
          }

          if (ad.length() > 0)
          {
            ad = ad.trim().toLowerCase();
            if (ad.compareTo("1") == 0 || ad.compareTo("true") == 0) autoDelete = true;
          }
        }
        else
        if (nodeName.compareTo("defaults") == 0)
        {
          NodeList nl2 = el.getChildNodes();
          for (int i2 = 0; i2 < nl2.getLength(); i2++)
          {
            Node node2 = nl2.item(i2);
            if (node2.getNodeType() == Node.ELEMENT_NODE)
            {
              Element el2 = (Element)node2;
              String nodeName2 = node2.getNodeName();

              if (nodeName2.compareTo("push_node_filter") == 0)
              {
                defaultPushNodeFilter = new NodeNameFilter();
                defaultPushNodeFilter.loadFromXML(el2);
              }
              else
              if (nodeName2.compareTo("push_destination_nodes") == 0)
              {
                defaultPushDestinationNodes = new NodeNameLicList();
                defaultPushDestinationNodes.loadFromXML(el2);
              }
              else
              if (nodeName2.compareTo("pull_node_filter") == 0)
              {
                defaultPullNodeFilter = new NodeNameFilter();
                defaultPullNodeFilter.loadFromXML(el2);
              }
              else
              if (nodeName2.compareTo("licence_text") == 0)
              {
                defaultLicenceText = XML.getTextContent(el2);
              }
            }
          }
        }
        else
        if (nodeName.compareTo("rules") == 0)
        {
          NodeList nl2 = el.getChildNodes();
          for (int i2 = 0; i2 < nl2.getLength(); i2++)
          {
            Node node2 = nl2.item(i2);
            if (node2.getNodeType() == Node.ELEMENT_NODE)
            {
              Element el2 = (Element)node2;
              String nodeName2 = node2.getNodeName();

              if (nodeName2.compareTo("rule") == 0)
              {
                String ruleName = el2.getAttribute("name");
                if (ruleName.length() == 0)
                {
                  throw new SAXException("Missing or empty attribute 'name' in 'rule' tag");
                }
                
                if (rules.containsKey(ruleName))
                {
                  throw new SAXException("Duplicate rule '" + ruleName + "' in contract");
                }
                
                ContractRule rule = new ContractRule(ruleName, this);
                rule.loadFromXML(el2);
                rules.put(ruleName, rule);
              }
            }
          }
        } //rules tag
      }
    }
    
    if (contractID.length() == 0)
    {
      throw new SAXException("Unspecified contractID - tag 'id' is missing or has empty content");
    }
  }
  
  /**
   * Saves the item to XML DOM tree.
   */
  public void saveToXML(Element element)
  {
    Document doc = element.getOwnerDocument();
    Element el;
  
    XML.newTextElement(element, "id", contractID);
    if (description.length() != 0) XML.newTextElement(element, "description", description);
    
    if (validFrom != null || validTo != null || autoDelete)
    {
      DateFormat dateFormat = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.MEDIUM);
      el = doc.createElement("validity");
      if (validFrom != null) el.setAttribute("from", dateFormat.format(validFrom));
      if (validTo != null) el.setAttribute("to", dateFormat.format(validTo));
      if (autoDelete) el.setAttribute("autodelete", "true");
      element.appendChild(el);
    }

    if (defaultPushDestinationNodes != null ||
        defaultPullNodeFilter != null ||
        defaultLicenceText.length() != 0)
    {
      el = doc.createElement("defaults");

      if (defaultPushNodeFilter != null)
      {
        Element el2 = doc.createElement("push_node_filter");
        defaultPushNodeFilter.saveToXML(el2);
        el.appendChild(el2);
      }
      
      if (defaultPushDestinationNodes != null)
      {
        Element el2 = doc.createElement("push_destination_nodes");
        defaultPushDestinationNodes.saveToXML(el2);
        el.appendChild(el2);
      }
      
      if (defaultPullNodeFilter != null)
      {
        Element el2 = doc.createElement("pull_node_filter");
        defaultPullNodeFilter.saveToXML(el2);
        el.appendChild(el2);
      }
      
      if (defaultLicenceText.length() != 0)
      {
        XML.newTextElement(el, "licence_text", defaultLicenceText);
      }
      
      element.appendChild(el);
    }
    
    el = doc.createElement("rules");
    
    Iterator it = rules.values().iterator();
    while (it.hasNext())
    {
      ContractRule rule = (ContractRule)it.next();
      Element el2 = doc.createElement("rule");
      el2.setAttribute("name", rule.getName());
      rule.saveToXML(el2);
      el.appendChild(el2);
    }
      
    element.appendChild(el);
  }
  
  
  public String getContractID()
  {
    return contractID;
  }
  
  public String getDescription()
  {
    return description;
  }

  public boolean getAutoDelete()
  {
    return autoDelete;
  }
  
  public Date getValidFrom()         //can be null
  {
    return validFrom;
  }
  
  public Date getValidTo()           //can be null
  {
    return validTo;
  }
  
  public boolean isValid()
  {
    Date now = Calendar.getInstance().getTime();
    return (validFrom == null || !validFrom.after(now)) && (validTo == null || !validTo.before(now));
  }

  public NodeNameLicList getDefaultPushDestinationNodes()
  {
    return defaultPushDestinationNodes;
  }
  
  public NodeNameFilter getDefaultPushNodeFilter()
  {
    return defaultPushNodeFilter;
  }
  
  public NodeNameFilter getDefaultPullNodeFilter()
  {
    return defaultPullNodeFilter;
  }
  
  public String getDefaultLicenceText()
  {
    return defaultLicenceText;
  }
  
  public Map getRules()
  {
    return rules;
  }

  public void setContractID(String contractID)
  {
    this.contractID = contractID;
  }
  
  public void setDescription(String description)
  {
    this.description = description;
  }

  public void setAutoDelete(boolean autoDelete)
  {
    this.autoDelete = autoDelete;
  }
  
  public void setValidFrom(Date validFrom)         //can be null
  {
    this.validFrom = validFrom;
  }
  
  public void setValidTo(Date validTo)           //can be null
  {
    this.validTo = validTo;
  }
  
  public void setDefaultPushDestinationNodes(NodeNameLicList defaultPushDestinationNodes)   //can be null
  {
    this.defaultPushDestinationNodes = defaultPushDestinationNodes;
  }
  
  public void setDefaultPushNodeFilter(NodeNameFilter defaultPushNodeFilter)
  {
    this.defaultPushNodeFilter = defaultPushNodeFilter;
  }
  
  public void setDefaultPullNodeFilter(NodeNameFilter defaultPullNodeFilter)  // can be null
  {
    this.defaultPullNodeFilter = defaultPullNodeFilter;
  }
  
  public void setDefaultLicenceText(String defaultLicenceText)
  {
    this.defaultLicenceText = defaultLicenceText;
  }

  public void addRule(ContractRule rule)
  {
    rules.put(rule.getName(), rule);
  }

  /**
   * Can contract accept "pull" and if so by which rule?
   *
   * @param bundleInfo "name of bundle"
   * @param nodeInfo node, which requests the bundle
   * @return ContractRule that can accept "pull" or null if no such rule exists
   */
  public ContractRule passPull(BundleInfo bundleInfo, NodeInfo nodeInfo)
  {
    synchronized (rules)
    {
      if (!isValid()) return null;
      
      boolean areDefaultsPresent = false;
      boolean nodePassedInDefaults = false;
      if (defaultPullNodeFilter != null)
      {
        areDefaultsPresent = true;
        nodePassedInDefaults = defaultPullNodeFilter.pass(nodeInfo);
      }
      
      Iterator it = rules.values().iterator();
      while (it.hasNext())
      {
        ContractRule contractRule = (ContractRule)it.next();
        if (contractRule.passPull(bundleInfo, nodeInfo, areDefaultsPresent, nodePassedInDefaults)) return contractRule;
      }
    }
    
    return null;
  }

  /**
   * Can contract accept "pull" by specified rule?
   *
   * @param bundleInfo "name of bundle"
   * @param nodeInfo node, which requests the bundle
   * @param contractRule name of rule in the contract
   * @return ContractRule if it can accept "pull" or null if it can't
   */
  public ContractRule passPull(BundleInfo bundleInfo, NodeInfo nodeInfo, String contractRule)
  {
    synchronized (rules)
    {
      if (!isValid()) return null;
      
      boolean areDefaultsPresent = false;
      boolean nodePassedInDefaults = false;
      if (defaultPullNodeFilter != null)
      {
        areDefaultsPresent = true;
        nodePassedInDefaults = defaultPullNodeFilter.pass(nodeInfo);
      }
      
      ContractRule rule = (ContractRule)rules.get(contractRule);
      if (rule == null) return null;
      if (rule.passPull(bundleInfo, nodeInfo, areDefaultsPresent, nodePassedInDefaults)) return rule;
    }
    
    return null;
  }

  /**
   * Can contract accept "push" and if so, by which rule?
   *
   * @param bundleInfo describes bundle that should be "pushed"
   * @param nodeInfo destination node
   * @return ContractRule that can accept "push" or null if no such rule exists
   */
  public ContractRule passPush(BundleInfo bundleInfo, NodeInfo nodeInfo)
  {
    synchronized (rules)
    {
      if (!isValid()) return null;
      
      boolean nodePassedInDefaults = false;
      if (defaultPullNodeFilter != null) nodePassedInDefaults = defaultPushDestinationNodes.pass(nodeInfo) != null;
      
      Iterator it = rules.values().iterator();
      while (it.hasNext())
      {
        ContractRule contractRule = (ContractRule)it.next();
        if (contractRule.passPush(bundleInfo, nodeInfo, nodePassedInDefaults)) return contractRule;
      }
    }
    
    return null;
  }

  /**
   * Can contract accept "push" by specified rule?
   *
   * @param bundleInfo describes bundle that should be "pushed"
   * @param nodeInfo destination node
   * @param contractRule name of rule in the contract
   * @return ContractRule if it can accept "push" or null if it can't
   */
  public ContractRule passPush(BundleInfo bundleInfo, NodeInfo nodeInfo, String contractRule)
  {
    synchronized (rules)
    {
      if (!isValid()) return null;
      
      boolean nodePassedInDefaults = false;
      if (defaultPullNodeFilter != null) nodePassedInDefaults = defaultPushDestinationNodes.pass(nodeInfo) != null;
      
      ContractRule rule = (ContractRule)rules.get(contractRule);
      if (rule == null) return null;
      if (rule.passPush(bundleInfo, nodeInfo, nodePassedInDefaults)) return rule;
    }
    
    return null;
  }

  
  /**
   * Generates input and output triggers that performs the rules introduced by the contract
   *
   * @param firstDelete if true all input and output triggers that belongs to the contract are deleted first
   * @param inputTriggers storage of Input Triggers
   * @param outputTriggers storage of Output Triggers
   */
  public void generateTriggers(boolean firstDelete, InputTriggers inputTriggers, OutputTriggers outputTriggers)
  {
    if (firstDelete) deleteTriggers(inputTriggers, outputTriggers);

    synchronized (rules)
    {
      Iterator it = rules.values().iterator();
      while (it.hasNext())
      {
        ContractRule rule = (ContractRule)it.next();
        
        Date valFrom = rule.getValidFrom();
        if (valFrom == null) valFrom = validFrom;
        if (valFrom != null) valFrom = (Date)valFrom.clone();

        Date valTo = rule.getValidTo();
        if (valTo == null) valTo = validTo;
        if (valTo != null) valTo = (Date)valTo.clone();
        
        
        if (rule.isPush())
        {
          NodeNameFilter pushNodeFilter = rule.getPushNodeFilter();
          if (pushNodeFilter == null) pushNodeFilter = defaultPushNodeFilter;
          if (pushNodeFilter != null) pushNodeFilter = (NodeNameFilter)pushNodeFilter.clone();
          
          NodeNameLicList pushDestinationNodes = rule.getPushDestinationNodes();
          if (pushDestinationNodes == null) pushDestinationNodes = defaultPushDestinationNodes;
          if (pushDestinationNodes != null) pushDestinationNodes = (NodeNameLicList)pushDestinationNodes.clone();
          
          if (pushDestinationNodes != null)
          {
            inputTriggers.addInputTrigger(
                  contractID,
                  true,
                  InputTrigger.SOURCE_PUSH | InputTrigger.SOURCE_PULL | InputTrigger.SOURCE_MADE,
                  InputTrigger.BUNDLETYPE_FULL,
                  "Generated from Contract (ID:'" + contractID +"'; Rule:'" + rule.getName() + "')",
                  rule.getAutoDelete(),
                  (BundleNameFilter)rule.getSoftware().clone(),
                  pushNodeFilter,
                  valFrom,
                  valTo,
                  contractID,
                  rule.getName(),
                  InputTrigger.ACTION_PUSH,
                  0,
                  null,
                  null,
                  false,
                  null,
                  pushDestinationNodes,
                  null);
          }
        }
        
        if (rule.isPull())
        {
          NodeNameFilter pullNodeFilter = rule.getPullNodeFilter();
          if (pullNodeFilter == null) pullNodeFilter = defaultPullNodeFilter;
          if (pullNodeFilter != null) pullNodeFilter = (NodeNameFilter)pullNodeFilter.clone();
          
          outputTriggers.addOutputTrigger(
                contractID,
                true,
                "Generated from Contract (ID:'" + contractID +"'; Rule:'" + rule.getName() + "')",
                rule.getAutoDelete(),
                (BundleNameFilter)rule.getSoftware().clone(),
                pullNodeFilter,
                valFrom,
                valTo,
                contractID,
                rule.getName());
        }
      }
    }
  }
  
  /**
   * Deletes input and output triggers belongs to the contract
   *
   * @param inputTriggers storage of Input Triggers
   * @param outputTriggers storage of Output Triggers
   */
  public void deleteTriggers(InputTriggers inputTriggers, OutputTriggers outputTriggers)
  {
    synchronized (inputTriggers) {
    synchronized (inputTriggers.map)
    {
      Iterator it = inputTriggers.map.values().iterator();
      while (it.hasNext())
      {
        InputTrigger inputTrigger = (InputTrigger)it.next();
        if (inputTrigger.isContract() && contractID.compareTo(inputTrigger.getContractID()) == 0)
        {
          inputTrigger.deleteFromStorage();
          it.remove();
        }
      }
    }
    } //inputTriggers

    synchronized (outputTriggers) {
    synchronized (outputTriggers.map)
    {
      Iterator it = outputTriggers.map.values().iterator();
      while (it.hasNext())
      {
        OutputTrigger outputTrigger = (OutputTrigger)it.next();
        if (outputTrigger.isContract() && contractID.compareTo(outputTrigger.getContractID()) == 0)
        {
          outputTrigger.deleteFromStorage();
          it.remove();
        }
      }
    }
    } //outputTriggers
  }
}
