/*
 * InputTrigger.java
 *
 * Created on 1. prosinec 2003, 20:47
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;
import java.text.*;
import SOFA.Util.XML;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

/**
 * Representation of the input trigger.
 * One item of persistent storage.
 * <p>
 * The input trigger contains:
 * <ul>
 * <li>at which operations the trigger should be used (push/pull/made)
 * <li>whether is should be used on classical bundles, bundle offers or both
 * <li>text description of the trigger
 * <li>filter that specifies the bundles on which the trigger should be used (for push operations it also says which bundles can be accepted at node)
 * <li>filter of nodes that must be passed to use the trigger (for push operations it also says from which nodes the bundles can be accepted)
 * <li>time validity
 * <li>(optional) ID and Rule of contract that is connected with the input trigger
 * <li>actions that should be done when the trigger is "executed"
 *   <ul>
 *   <li>install
 *   <li>share
 *   <li>allow pull
 *   <li>push
 *   <li>push_offer
 *   <li>delete
 *   </ul>
 * </ul>
 *
 * @author  Ladislav Sobr
 */
public class InputTrigger extends XMLStorageItem implements StorageItem, Serializable
{
  final static public int SOURCE_NONE = 0;
  final static public int SOURCE_PUSH = 1;
  final static public int SOURCE_PULL = 2;
  final static public int SOURCE_MADE = 4;

  final static public int BUNDLETYPE_FULL = 0;
  final static public int BUNDLETYPE_OFFER = 1;
  final static public int BUNDLETYPE_ALL = 2;
  
  final static public int ACTION_INSTALL = 1;
  final static public int ACTION_SHARE = 2;
  final static public int ACTION_ALLOW_PULL = 4;
  final static public int ACTION_PUSH = 8;
  final static public int ACTION_PUSH_OFFER = 16;
  final static public int ACTION_DELETE = 32;
  
  private int source;
  private int bundleType;
  private String description;
  private boolean autoDelete;
  private BundleNameFilter bundleFilter;  //can be null
  private NodeNameFilter nodeFilter;      //can be null
  private Date validFrom;         //can be null
  private Date validTo;           //can be null
  private boolean isContract;
  private String contractID;
  private String contractRule;
  private int actions;
  
  //parameters of INSTALL action
  //parameters of SHARE action
  private int asNoLicences;            // Licence.ONE_IMPLICIT_LICENCE means "one licence", Licence.ALL_LICENCES means "all remaining licences"
  private ShareGroups asShareGroups;   // empty list means "all groups"
  private NodeNameFilter asNodeFilter;
  private boolean asEquality;
  //parameters of ALLOW_PULL action
  private OutputTrigger alpOutputTrigger;
  //parameters of PUSH action
  private NodeNameLicList apNodeNameLicList;
  //parameters of PUSH_OFFER action
  private NodeNameList apoNodeNameList;
  //parameters of PUSH_DELETE action
  
  
  /** Creates a new instance of InputTrigger */
  public InputTrigger(String name, File file)
  {
    super(name, file, "input_trigger");
    reset();
  }
  
  public InputTrigger()
  {
    super("", null, "input_trigger");
    reset();
  }
  
  public boolean toBeDeleted()
  {
    return validTo != null && autoDelete && validTo.before(Calendar.getInstance().getTime());
  }

  /**
   * Resets content of the item.
   */
  protected void reset()
  {
    source = SOURCE_NONE;
    bundleType = BUNDLETYPE_FULL;
    bundleFilter = null;
    nodeFilter = null;
    description = "";
    validFrom = null;
    validTo = null;
    autoDelete = false;
    isContract = false;
    contractID = "";
    contractRule = "";
    actions = 0;
    asNoLicences = Licence.ONE_IMPLICIT_LICENCE;
    asShareGroups = null;
    asNodeFilter = null;
    asEquality = false;
    alpOutputTrigger = null;
    apNodeNameLicList = null;
    apoNodeNameList = null;
  }
  
  /**
   * Loads the item from the XML DOM tree.
   * <p>
   * XML format of storage item:
   * <p>
   * <pre>
   * &lt;input_trigger&gt;
   *   InputTrigger XML format
   * &lt;/input_trigger&gt;
   * </pre>
   * <p>
   * XML format:
   * <p>
   * <pre>
   *   &lt;source push="boolean" pull="boolean" made="boolean"/&gt;
   *   &lt;bundle_type&gt;FULL or OFFER or ALL&lt;/bundle_type&gt;
   *   &lt;description&gt;string&lt;/description&gt;
   *  ?&lt;bundle_filter&gt;
   *     BundleNameFilter XML format
   *   &lt;/bundle_filter&gt;
   *  ?&lt;node_filter&gt;
   *     NodeNameFilter XML format
   *   &lt;/node_filter&gt;
   *  ?&lt;validity ?from="date" ?to="date" ?autodelete="boolean"/&gt;
   *  ?&lt;contract cid="string" rule="string"/&gt;
   *   &lt;actions&gt;
   *    ?&lt;install/&gt;
   *    ?&lt;share ?number_of_licences="positive-integer or *" equality="boolean"&gt;
   *      ?&lt;share_groups&gt;
   *         ShareGroups XML format
   *       &lt;/share_groups&gt;
   *      ?&lt;node_filter&gt;
   *         NodeNameFilter XML format
   *       &lt;/node_filter&gt;
   *     &lt;/share&gt;
   *    ?&lt;allow_pull&gt;
   *       OutputTrigger XML format
   *     &lt;/allow_pull&gt;
   *    ?&lt;push&gt;
   *      NodeNameLicList XML format
   *     &lt;/push&gt;
   *    ?&lt;push_offer&gt;
   *      NodeNameList XML format
   *     &lt;/push_offer&gt;
   *    ?&lt;delete/&gt;
   *   &lt;/actions&gt;
   * </pre>
   */
  public void loadFromXML(Element element) throws SAXException
  {
    reset();
    NodeList nl = element.getChildNodes();
    for (int i = 0; i < nl.getLength(); i++)
    {
      Node node = nl.item(i);
      if (node.getNodeType() == Node.ELEMENT_NODE)
      {
        Element el = (Element)node;
        String nodeName = node.getNodeName();
        
        if (nodeName.compareTo("source") == 0)
        {
          String sourcePush = el.getAttribute("push");
          String sourcePull = el.getAttribute("pull");
          String sourceMade = el.getAttribute("made");
          
          if (sourcePush.length() > 0)
          {
            sourcePush = sourcePush.trim().toLowerCase();
            if (sourcePush.compareTo("1") == 0 || sourcePush.compareTo("true") == 0) source |= SOURCE_PUSH;
          }
          
          if (sourcePull.length() > 0)
          {
            sourcePull = sourcePull.trim().toLowerCase();
            if (sourcePull.compareTo("1") == 0 || sourcePull.compareTo("true") == 0) source |= SOURCE_PULL;
          }
          
          if (sourceMade.length() > 0)
          {
            sourceMade = sourceMade.trim().toLowerCase();
            if (sourceMade.compareTo("1") == 0 || sourceMade.compareTo("true") == 0) source |= SOURCE_MADE;
          }
        }
        else
        if (nodeName.compareTo("bundle_type") == 0)
        {
          String t = XML.getTextContent(el).trim().toLowerCase();
          if (t.compareTo("full") == 0) bundleType = BUNDLETYPE_FULL;
          else if (t.compareTo("offer") == 0) bundleType = BUNDLETYPE_OFFER;
          else if (t.compareTo("all") == 0) bundleType = BUNDLETYPE_ALL;
          else throw new SAXException("Invalid type of input trigger ('bundle_type' tag)");
        }
        else
        if (nodeName.compareTo("description") == 0)
        {
          description = XML.getTextContent(el);
        }
        else
        if (nodeName.compareTo("bundle_filter") == 0)
        {
          bundleFilter = new BundleNameFilter();
          bundleFilter.loadFromXML(el);
        }
        else
        if (nodeName.compareTo("node_filter") == 0)
        {
          nodeFilter = new NodeNameFilter();
          nodeFilter.loadFromXML(el);
        }
        else
        if (nodeName.compareTo("validity") == 0)
        {
          String from = el.getAttribute("from");
          String to = el.getAttribute("to");
          String ad = el.getAttribute("autodelete");
          DateFormat dateFormat = DateFormat.getInstance();
          if (from.length() > 0)
          {
            try
            {
              validFrom = dateFormat.parse(from);
            }
            catch (ParseException e)
            {
              throw new SAXException("Cannot parse date/time format of 'from' attribute of 'validity' tag", e);
            }
          }
          
          if (to.length() > 0)
          {
            try
            {
              validTo = dateFormat.parse(to);
            }
            catch (ParseException e)
            {
              throw new SAXException("Cannot parse date/time format of 'to' attribute of 'validity' tag", e);
            }
          }
          
          if (ad.length() > 0)
          {
            ad = ad.trim().toLowerCase();
            if (ad.compareTo("1") == 0 || ad.compareTo("true") == 0) autoDelete = true;
          }
        }
        else
        if (nodeName.compareTo("contract") == 0)
        {
          isContract = true;
          contractID = el.getAttribute("cid");
          contractRule = el.getAttribute("rule");
        }
        else
        if (nodeName.compareTo("actions") == 0)
        {
          NodeList nl2 = el.getChildNodes();
          for (int i2 = 0; i2 < nl2.getLength(); i2++)
          {
            Node node2 = nl2.item(i2);
            if (node2.getNodeType() == Node.ELEMENT_NODE)
            {
              Element el2 = (Element)node2;
              String nodeName2 = node2.getNodeName();

              if (nodeName2.compareTo("install") == 0)
              {
                actions |= ACTION_INSTALL;
              }
              else
              if (nodeName2.compareTo("share") == 0)
              {
                actions |= ACTION_SHARE;
                
                String s = el2.getAttribute("number_of_licences");
                if (s.compareTo("*") == 0) asNoLicences = Licence.ALL_LICENCES;
                else
                if (s.length() != 0)
                {
                  try
                  {
                    int l = Integer.parseInt(s);
                    if (l <= 0) throw new NumberFormatException();
                    asNoLicences = l;
                  }
                  catch (NumberFormatException e)
                  {
                    throw new SAXException("Invalid 'positive-integer or *' format of 'number_of_licences' attribute of 'share' tag", e);
                  }
                }
                
                String equality = el2.getAttribute("equality").trim().toLowerCase();
                if (equality.compareTo("1") == 0 ||
                    equality.compareTo("true") == 0) asEquality = true;
                
                NodeList nl3 = el2.getChildNodes();
                for (int i3 = 0; i3 < nl3.getLength(); i3++)
                {
                  Node node3 = nl3.item(i3);
                  if (node3.getNodeType() == Node.ELEMENT_NODE)
                  {
                    Element el3 = (Element)node3;
                    String nodeName3 = node3.getNodeName();

                    if (nodeName3.compareTo("share_groups") == 0)
                    {
                      asShareGroups = new ShareGroups();
                      asShareGroups.loadFromXML(el3);
                    }
                    else
                    if (nodeName3.compareTo("node_filter") == 0)
                    {
                      asNodeFilter = new NodeNameFilter();
                      asNodeFilter.loadFromXML(el3);
                    }
                  }
                }
              }
              else
              if (nodeName2.compareTo("allow_pull") == 0)
              {
                actions |= ACTION_ALLOW_PULL;
                
                alpOutputTrigger = new OutputTrigger();
                alpOutputTrigger.loadFromXML(el2);
              }
              else
              if (nodeName2.compareTo("push") == 0)
              {
                actions |= ACTION_PUSH;
                
                apNodeNameLicList = new NodeNameLicList();
                apNodeNameLicList.loadFromXML(el2);
              }
              else
              if (nodeName2.compareTo("push_offer") == 0)
              {
                actions |= ACTION_PUSH_OFFER;
                
                apoNodeNameList = new NodeNameList();
                apoNodeNameList.loadFromXML(el2);
              }
              else
              if (nodeName2.compareTo("delete") == 0)
              {
                actions |= ACTION_DELETE;
              }
            }
          }
        } //actions tag
      }
    }
  }
  
  /**
   * Saves the item to XML DOM tree.
   */
  public void saveToXML(Element element)
  {
    Document doc = element.getOwnerDocument();
    Element el;
    
    el = doc.createElement("source");
    if ((source & SOURCE_PUSH) != 0) el.setAttribute("push", "true");
    else el.setAttribute("push", "false");
    if ((source & SOURCE_PULL) != 0) el.setAttribute("pull", "true");
    else el.setAttribute("pull", "false");
    if ((source & SOURCE_MADE) != 0) el.setAttribute("made", "true");
    else el.setAttribute("made", "false");
    element.appendChild(el);
    
    String bundleTypeText;
    switch (bundleType)
    {
      case BUNDLETYPE_FULL: bundleTypeText = "full"; break;
      case BUNDLETYPE_OFFER: bundleTypeText = "offer"; break;
      case BUNDLETYPE_ALL: bundleTypeText = "all"; break;
      default: bundleTypeText = "full"; break;
    }
    
    XML.newTextElement(element, "bundle_type", bundleTypeText);
    XML.newTextElement(element, "description", description);
    
    if (bundleFilter != null)
    {
      el = doc.createElement("bundle_filter");
      bundleFilter.saveToXML(el);
      element.appendChild(el);
    }
    
    if (nodeFilter != null)
    {
      el = doc.createElement("node_filter");
      nodeFilter.saveToXML(el);
      element.appendChild(el);
    }
    
    if (validFrom != null || validTo != null || autoDelete)
    {
      DateFormat dateFormat = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.MEDIUM);
      el = doc.createElement("validity");
      if (validFrom != null) el.setAttribute("from", dateFormat.format(validFrom));
      if (validTo != null) el.setAttribute("to", dateFormat.format(validTo));
      if (autoDelete) el.setAttribute("autodelete", "true");
      element.appendChild(el);
    }
    
    if (isContract)
    {
      el = doc.createElement("contract");
      el.setAttribute("cid", contractID);
      if (contractRule.length() > 0) el.setAttribute("rule", contractRule);
      element.appendChild(el);
    }
    
    el = doc.createElement("actions");
    
    if ((actions & ACTION_INSTALL) != 0)
    {
      Element el2 = doc.createElement("install");
      el.appendChild(el2);
    }
    
    if ((actions & ACTION_SHARE) != 0)
    {
      Element el2 = doc.createElement("share");
      String s = null;
      if (asNoLicences == Licence.ALL_LICENCES) s = "*";
      else if (asNoLicences > 0) s = Integer.toString(asNoLicences);
      if (s != null) el2.setAttribute("number_of_licences", s);

      if (asEquality) el2.setAttribute("equality", "true");
      else el2.setAttribute("equality", "false");
      
      if (asShareGroups != null)
      {
        Element el3 = doc.createElement("share_groups");
        asShareGroups.saveToXML(el3);
        el2.appendChild(el3);
      }
      
      if (asNodeFilter != null)
      {
        Element el3 = doc.createElement("node_filter");
        asNodeFilter.saveToXML(el3);
        el2.appendChild(el3);
      }

      el.appendChild(el2);
    }
    
    if ((actions & ACTION_ALLOW_PULL) != 0)
    {
      Element el2 = doc.createElement("allow_pull");
      alpOutputTrigger.saveToXML(el2);
      el.appendChild(el2);
    }
    
    if ((actions & ACTION_PUSH) != 0)
    {
      Element el2 = doc.createElement("push");
      apNodeNameLicList.saveToXML(el2);
      el.appendChild(el2);
    }

    if ((actions & ACTION_PUSH_OFFER) != 0)
    {
      Element el2 = doc.createElement("push_offer");
      apoNodeNameList.saveToXML(el2);
      el.appendChild(el2);
    }

    if ((actions & ACTION_DELETE) != 0)
    {
      Element el2 = doc.createElement("delete");
      el.appendChild(el2);
    }
        
    element.appendChild(el);
  }

  public int getSource()
  {
    return source;
  }
  
  public boolean isPushSource()
  {
    return (source & SOURCE_PUSH) != 0;
  }
  
  public boolean isPullSource()
  {
    return (source & SOURCE_PULL) != 0;
  }
  
  public boolean isMadeSource()
  {
    return (source & SOURCE_MADE) != 0;
  }
  
  public int getBundleType()
  {
    return bundleType;
  }

  public String getDescription()
  {
    return description;
  }

  public boolean getAutoDelete()
  {
    return autoDelete;
  }
  
  public BundleNameFilter getBundleFilter()
  {
    return bundleFilter;
  }
  
  public NodeNameFilter getNodeFilter()
  {
    return nodeFilter;
  }
  
  public Date getValidFrom()         //can be null
  {
    return validFrom;
  }
  
  public Date getValidTo()           //can be null
  {
    return validTo;
  }
  
  public boolean isValid()
  {
    Date now = Calendar.getInstance().getTime();
    return (validFrom == null || !validFrom.after(now)) && (validTo == null || !validTo.before(now));
  }  

  public boolean isContract()
  {
    return isContract;
  }
  
  public String getContractID()
  {
    return contractID;
  }
  
  public String getContractRule()
  {
    return contractRule;
  }

  public int getActions()
  {
    return actions;
  }

  public int getAsNoLicences()
  {
    return asNoLicences;
  }
  
  public ShareGroups getAsShareGroups()
  {
    return asShareGroups;
  }
  
  public NodeNameFilter getAsNodeFilter()
  {
    return asNodeFilter;
  }
  
  public boolean getAsEquality()
  {
    return asEquality;
  }
  
  public OutputTrigger getAlpOutputTrigger()
  {
    return alpOutputTrigger;
  }
  
  public NodeNameLicList getApNodeNameLicList()
  {
    return apNodeNameLicList;
  }
  
  public NodeNameList getApoNodeNameList()
  {
    return apoNodeNameList;
  }

  public void setSource(int source)
  {
    this.source = source;
  }
  
  public void setPushSource(boolean isPush)
  {
    if (isPush) source |= SOURCE_PUSH;
    else source &= ~SOURCE_PUSH;
  }
  
  public void setPullSource(boolean isPull)
  {
    if (isPull) source |= SOURCE_PULL;
    else source &= ~SOURCE_PULL;
  }
  
  public void setMadeSource(boolean isMade)
  {
    if (isMade) source |= SOURCE_MADE;
    else source &= ~SOURCE_MADE;
  }
  
  public void setBundleType(int bundleType)
  {
    if (bundleType != BUNDLETYPE_FULL && bundleType != BUNDLETYPE_OFFER && bundleType != BUNDLETYPE_ALL) bundleType = BUNDLETYPE_FULL;
    this.bundleType = bundleType;
  }
  
  public void setDescription(String description)
  {
    this.description = description;
  }
  
  public void setAutoDelete(boolean autoDelete)
  {
    this.autoDelete = autoDelete;
  }
  
  public void setBundleFilter(BundleNameFilter bundleFilter)
  {
    this.bundleFilter = bundleFilter;
  }
  
  public void setNodeFilter(NodeNameFilter nodeFilter)
  {
    this.nodeFilter = nodeFilter;
  }
  
  public void setValidFrom(Date validFrom)         //can be null
  {
    this.validFrom = validFrom;
  }
  
  public void setValidTo(Date validTo)           //can be null
  {
    this.validTo = validTo;
  }
  
  public void setContract(String cid, String rule)
  {
    if (cid == null && rule == null)
    {
      contractID = "";
      contractRule = "";
      isContract = false;
    }
    else
    {
      contractID = cid;
      contractRule = rule;
      isContract = true;
    }
  }
  
  public void setActions(int actions)
  {
    this.actions = actions;
  }

  public void setAsNoLicences(int asNoLicences)
  {
    if (asNoLicences <= 0 && asNoLicences != Licence.ALL_LICENCES) asNoLicences = Licence.ONE_IMPLICIT_LICENCE;
    this.asNoLicences = asNoLicences;
  }

  public void setAsShareGroups(ShareGroups asShareGroups)
  {
    this.asShareGroups = asShareGroups;
  }
  
  public void setAsNodeFilter(NodeNameFilter asNodeFilter)
  {
    this.asNodeFilter = asNodeFilter;
  }
  
  public void setAsEquality(boolean asEquality)
  {
    this.asEquality = asEquality;
  }
  
  public void setAlpOutputTrigger(OutputTrigger alpOutputTrigger)
  {
    this.alpOutputTrigger = alpOutputTrigger;
  }
  
  public void setApNodeNameLicList(NodeNameLicList apNodeNameLicList)
  {
    if (apNodeNameLicList == null) this.apNodeNameLicList = null;
    else this.apNodeNameLicList = apNodeNameLicList;
  }
  
  public void setApoNodeNameList(NodeNameList apoNodeNameList)
  {
    if (apoNodeNameList == null) this.apoNodeNameList = null;
    else this.apoNodeNameList = apoNodeNameList;
  }
  
}
