/*
 * InputTriggers.java
 *
 * Created on 1. prosinec 2003, 23:24
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;

/**
 * Persistent storage of input triggers.
 * Implemented as directory of XML files.
 *
 * @author  Ladislav Sobr
 */
public class InputTriggers extends StorageBase
{
  
  /** Creates a new instance of InputTriggers */
  public InputTriggers(File baseDir)
  {
    super(baseDir, "it", true);
    init();
  }
  
  protected StorageItem newItem(String name, File file)
  {
    return new InputTrigger(name, file);
  }
  
  /**
   * Returns specified input trigger
   * The function assures that the returned InputTrigger is loaded in the memory.
   *
   * @param name name of input trigger
   * @return requested input trigger; null if not present
   */
  public InputTrigger getInputTrigger(String name)
  {
    return (InputTrigger)map.get(name);
  }

  /**
   * Deletes specified input trigger
   *
   * @param name name of input trigger
   * @return true, if input trigger was found (and deleted)
   */
  public boolean deleteInputTrigger(String name)
  {
    synchronized (map)
    {
      InputTrigger inputTrigger = (InputTrigger)map.remove(name);
      if (inputTrigger != null)
      {
        inputTrigger.deleteFromStorage();
        return true;
      }
      else return false;
    }
  }
  
  /**
   * Adds input trigger.
   * <p>
   * If contract is not present, set contractID and contractRule to null
   * <p>
   * <ul>
   * <li>If customizableName == false:
   *   <ul>
   *   <li>name is full name of input trigger and is not changed in any way
   *   <li>function can fail
   *   </ul>
   * <li>If customizableName == true:
   *   <ul>
   *   <li>name is prefix of the name, postfix is generated as sequent numbers
   *   <li>if name is empty, prefix is "auto" + randomly generated string
   *   <li>function never fail
   *   </ul>
   * </ul>
   * <p>
   * For description of all parameters see InputTrigger class
   *  
   * @param name can be null - in that case the name will be generated and function always succeed
   * @param validFrom can be null
   * @param validTo can be null
   * @param asNoLicences Licence.ONE_IMPLICIT_LICENCE means "one licence", Licence.ALL_LICENCES means "all remaining licences"
   * @return added input trigger, if input trigger was added (there was not already the one with the same name); null otherwise
   */
  public InputTrigger addInputTrigger(String name,
                                      boolean customizableName,
                                      int source,
                                      int bundleType,
                                      String description,
                                      boolean autoDelete,
                                      BundleNameFilter bundleFilter,
                                      NodeNameFilter nodeFilter,
                                      Date validFrom,
                                      Date validTo,
                                      String contractID,
                                      String contractRule,
                                      int actions,
                                      int asNoLicences,            // Licence.ONE_IMPLICIT_LICENCE means "one licence", Licence.ALL_LICENCES means "all remaining licences"
                                      ShareGroups asShareGroups,
                                      NodeNameFilter asNodeFilter,
                                      boolean asEquality,
                                      OutputTrigger alpOutputTrigger,
                                      NodeNameLicList apNodeNameLicList,
                                      NodeNameList apoNodeNameList)
  {
    synchronized (map)
    {
      if (customizableName || name == null)
      {
        if (name == null || name.length() == 0)
        {
          name = "auto";
          Random random = new Random();
          for (int i = 0; i < 10; i++) name += Integer.toString(random.nextInt(16), 16);
        }
        
        String postfix = "";
        for (int i = 0; ; i++)
        {
          if (map.get(name + postfix) == null)
          {
            name += postfix;
            break;
          }
          postfix = Integer.toString(i + 1);
        }
      }
      else
      {
        if (map.get(name) != null) return null;
      }

      InputTrigger inputTrigger = new InputTrigger(name, new File(baseDir, name + ".it"));
      inputTrigger.setSource(source);
      inputTrigger.setBundleType(bundleType);
      inputTrigger.setDescription(description);
      inputTrigger.setAutoDelete(autoDelete);
      inputTrigger.setBundleFilter(bundleFilter);
      inputTrigger.setNodeFilter(nodeFilter);
      inputTrigger.setValidFrom(validFrom);
      inputTrigger.setValidTo(validTo);
      inputTrigger.setContract(contractID, contractRule);
      inputTrigger.setActions(actions);
      inputTrigger.setAsNoLicences(asNoLicences);
      inputTrigger.setAsShareGroups(asShareGroups);
      inputTrigger.setAsNodeFilter(asNodeFilter);
      inputTrigger.setAsEquality(asEquality);
      inputTrigger.setAlpOutputTrigger(alpOutputTrigger);
      inputTrigger.setApNodeNameLicList(apNodeNameLicList);
      inputTrigger.setApoNodeNameList(apoNodeNameList);
      
      inputTrigger.saveToStorage();
      map.put(name, inputTrigger);
      return inputTrigger;
    }
  }
}
