/*
 * LocalInfo.java
 *
 * Created on 18. bezen 2004, 23:03
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;
import java.text.*;
import SOFA.Util.XML;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

/**
 * Representation of the local informations about the bundle.
 * One item of persistent storage.
 * <p>
 * The local info contains:
 * <ul>
 * <li>state of the bundle
 *   <ul>
 *   <li>installed
 *   <li>in the memory (+ nonpersistent counter of usage of bundle)
 *   <li>shared (share manager or share client)
 *   </ul>
 * <li>local licence of the bundle (used also for sharing)
 * <li>share manager information:
 *   <ul>
 *   <li>list of clients
 *   <li>share groups
 *   <li>filter of nodes that can become clients
 *   <li>whether share manager is equal to share clients (if so, all copies of bundle can be gone to clients and no copy remains on the share manager)
 *   </ul>
 * <li>share client information:
 *   <ul>
 *   <li>name of node with share manager
 *   </ul>
 * </ul>
 *
 * @author  Ladislav Sobr
 */
public class LocalInfo extends XMLStorageItem implements StorageItem, Serializable
{
  final static public int STATE_NONE = 0;
  final static public int STATE_INSTALLED = 1;
  final static public int STATE_SHARE_MANAGER = 2;
  final static public int STATE_SHARE_CLIENT = 4;
  final static public int STATE_SHARE_PRECLIENT = 8; //temporaty state during instatiating of share client

  private Licence licence;
  private int state;
  private ShareGroups smShareGroups;         // for STATE_SHARE_MANAGER
  private NodeNameFilter smNodeFilter;       // for STATE_SHARE_MANAGER
  private NodeNameList smClients;            // for STATE_SHARE_MANAGER
  private boolean smEquality;                // for STATE_SHARE_MANAGER
  private String scManager;                  // for STATE_SHARE_CLIENT
  private int inMemoryCounter;
  private boolean inMemory;
  
  /** Creates a new instance of LocalInfo */
  public LocalInfo(String name, File file)
  {
    super(name, file, "local_info");
    reset();
  }

  /**
   * Resets content of the item.
   */
  protected void reset()
  {
    licence = new Licence();
    state = STATE_NONE;
    smShareGroups = null;
    smNodeFilter = null;
    smClients = null;
    smEquality = false;
    scManager = "";
    inMemoryCounter = 0;
    inMemory = false;
  }

  /**
   * Loads the item from the XML DOM tree.
   * <p>
   * XML format of storage item:
   * <p>
   * <pre>
   * &lt;local_info&gt;
   *   LocalInfo XML format
   * &lt;/local_info&gt;
   * </pre>
   * <p>
   * XML format:
   * <p>
   * <pre>
   *   &lt;licence&gt;
   *     Licence XML format
   *   &lt;/licence&gt;
   *   &lt;state ?installed="boolean" ?sharing="MANAGER or CLIENT or PRECLIENT"/ ?in_memory="boolean"&gt;
   *  ?&lt;share_manager equality="boolean"&gt;
   *    ?&lt;share_groups&gt;
   *       ShareGroups XML format
   *     &lt;/share_groups&gt;
   *    ?&lt;node_filter&gt;
   *       NodeNameFilter XML format
   *     &lt;/node_filter&gt;
   *     &lt;clients&gt;
   *       NodeNameList XML format
   *     &lt;/clients&gt;
   *   &lt;/share_manager&gt;
   *  ?&lt;share_client manager="string"/&gt;
   * </pre>
   *
   */
  public void loadFromXML(Element element) throws SAXException
  {
    reset();
    NodeList nl = element.getChildNodes();
    for (int i = 0; i < nl.getLength(); i++)
    {
      Node node = nl.item(i);
      if (node.getNodeType() == Node.ELEMENT_NODE)
      {
        Element el = (Element)node;
        String nodeName = node.getNodeName();
        
        if (nodeName.compareTo("licence") == 0)
        {
          licence.loadFromXML(el);
        }
        else
        if (nodeName.compareTo("state") == 0)
        {
          String installed = el.getAttribute("installed").trim().toLowerCase();
          if (installed.compareTo("1") == 0 ||
              installed.compareTo("true") == 0) state |= STATE_INSTALLED;
          
          String sharing = el.getAttribute("sharing").trim().toLowerCase();
          if (sharing.length() > 0)
          {
            if (sharing.compareTo("manager") == 0) state |= STATE_SHARE_MANAGER;
            else if (sharing.compareTo("client") == 0) state |= STATE_SHARE_CLIENT;
            else if (sharing.compareTo("preclient") == 0) state |= STATE_SHARE_PRECLIENT;
            else throw new SAXException("Invalid content of 'sharing' attribute ('state' tag)");
          }
          
          String in_memory = el.getAttribute("in_memory").trim().toLowerCase();
          if (in_memory.compareTo("1") == 0 ||
              in_memory.compareTo("true") == 0)
          {
            inMemory = true;
            inMemoryCounter = 1; //fake value
          }
        }
        else
        if (nodeName.compareTo("share_manager") == 0)
        {
          String equality = el.getAttribute("equality").trim().toLowerCase();
          if (equality.compareTo("1") == 0 ||
              equality.compareTo("true") == 0) smEquality = true;
          
          NodeList nl2 = el.getChildNodes();
          for (int i2 = 0; i2 < nl2.getLength(); i2++)
          {
            Node node2 = nl2.item(i2);
            if (node2.getNodeType() == Node.ELEMENT_NODE)
            {
              Element el2 = (Element)node2;
              String nodeName2 = node2.getNodeName();

              if (nodeName2.compareTo("share_groups") == 0)
              {
                smShareGroups = new ShareGroups();
                smShareGroups.loadFromXML(el2);
              }
              else
              if (nodeName2.compareTo("node_filter") == 0)
              {
                smNodeFilter = new NodeNameFilter();
                smNodeFilter.loadFromXML(el2);
              }
              else
              if (nodeName2.compareTo("clients") == 0)
              {
                smClients = new NodeNameList();
                smClients.loadFromXML(el2);
              }
            }
          }
        }
        else
        if (nodeName.compareTo("share_client") == 0)
        {
          scManager = el.getAttribute("manager");
        }
      }
    }
    
    if (isShareManager() && smClients == null) smClients = new NodeNameList();
  }
  
  /**
   * Saves the item to XML DOM tree.
   */
  public void saveToXML(Element element)
  {
    Document doc = element.getOwnerDocument();
    Element el;
  
    el = doc.createElement("licence");
    licence.saveToXML(el);
    element.appendChild(el);
    
    el = doc.createElement("state");
    
    if ((state & STATE_INSTALLED) != 0) el.setAttribute("installed", "true");
    else el.setAttribute("installed", "false");

    if ((state & STATE_SHARE_MANAGER) != 0) el.setAttribute("sharing", "manager");
    else if ((state & STATE_SHARE_CLIENT) != 0) el.setAttribute("sharing", "client");
    else if ((state & STATE_SHARE_PRECLIENT) != 0) el.setAttribute("sharing", "preclient");
    
    if (inMemory) el.setAttribute("in_memory", "true");
    
    element.appendChild(el);

    if (isShareManager())
    {
      el = doc.createElement("share_manager");
      
      if (smEquality) el.setAttribute("equality", "true");
      else el.setAttribute("equality", "false");
      
      if (smShareGroups != null)
      {
        Element el2 = doc.createElement("share_groups");
        smShareGroups.saveToXML(el2);
        el.appendChild(el2);
      }
      
      if (smNodeFilter != null)
      {
        Element el2 = doc.createElement("node_filter");
        smNodeFilter.saveToXML(el2);
        el.appendChild(el2);
      }
      
      if (smClients != null)
      {
        Element el2 = doc.createElement("clients");
        smClients.saveToXML(el2);
        el.appendChild(el2);
      }
      
      element.appendChild(el);
    }
    
    if (isShareClient())
    {
      el = doc.createElement("share_client");
      el.setAttribute("manager", scManager);
      element.appendChild(el);
    }
  }
  
  public Licence getLicence()
  {
    return licence;
  }
  
  public int getState()
  {
    return state;
  }
  
  public ShareGroups getSMShareGroups()
  {
    return smShareGroups;
  }
  
  public NodeNameFilter getSMNodeFilter()
  {
    return smNodeFilter;
  }
  
  public NodeNameList getSMClients()
  {
    return smClients;
  }
  
  public boolean isSMEquality()
  {
    return smEquality;
  }
  
  public String getSCManager()
  {
    return scManager;
  }
  
  public int getInMemoryCounter()
  {
    return inMemoryCounter;
  }
  
  public void setLicence(Licence licence)
  {
    this.licence = licence;
  }
  
  public void setState(int state)
  {
    this.state = state;
  }
  
  public void setSMShareGroups(ShareGroups shareGroups)
  {
    this.smShareGroups = shareGroups;
  }
  
  public void setSMNodeFilter(NodeNameFilter nodeFilter)
  {
    this.smNodeFilter = nodeFilter;
  }
  
  public void setSMClients(NodeNameList clients)
  {
    this.smClients = clients;
  }
  
  public void setSMEquality(boolean smEquality)
  {
    this.smEquality = smEquality;
  }
  
  public void setSCManager(String manager)
  {
    this.scManager = manager;
  }
  
  /**
   * @return true, if inMemory flag has been changed and local info should be saved to storage
   */
  public boolean setInMemoryCounter(int inMemoryCounter)
  {
    this.inMemoryCounter = inMemoryCounter;
    boolean oldInMemory = inMemory;
    inMemory = inMemoryCounter > 0;
    return oldInMemory != inMemory;
  }
  
  public void increaseInMemoryCounter()
  {
    if (inMemoryCounter == 0) inMemory = true;
    inMemoryCounter++;
  }
  
  public void decreaseInMemoryCounter()
  {
    inMemoryCounter--;
    if (inMemoryCounter < 0) inMemoryCounter = 0;
    if (inMemoryCounter == 0) inMemory = false;
  }
  
  public boolean isInstalled()
  {
    return (state & STATE_INSTALLED) != 0;
  }
  
  public void setInstalled(boolean installed)
  {
    if (installed) state |= STATE_INSTALLED;
    else state &= ~STATE_INSTALLED;
  }
  
  public boolean isShareManager()
  {
    return (state & STATE_SHARE_MANAGER) != 0;
  }
  
  public void setShareManager(boolean shareManager)
  {
    if (shareManager) state |= STATE_SHARE_MANAGER;
    else state &= ~STATE_SHARE_MANAGER;
  }
  
  public boolean isShareClient()
  {
    return (state & STATE_SHARE_CLIENT) != 0;
  }

  public boolean areShareClientModes()
  {
    return (state & (STATE_SHARE_CLIENT | STATE_SHARE_PRECLIENT)) != 0;
  }

  public boolean isSharePreClient()
  {
    return (state & STATE_SHARE_PRECLIENT) != 0;
  }

  public void setShareClient(boolean shareClient)
  {
    if (shareClient) state |= STATE_SHARE_CLIENT;
    else state &= ~STATE_SHARE_CLIENT;
  }
  
  public void setSharePreClient(boolean sharePreClient)
  {
    if (sharePreClient) state |= STATE_SHARE_PRECLIENT;
    else state &= ~STATE_SHARE_PRECLIENT;
  }
  
  public boolean isInSharing()
  {
    return (state & (STATE_SHARE_MANAGER | STATE_SHARE_CLIENT | STATE_SHARE_PRECLIENT)) != 0;
  }
  
  public boolean isInMemory()
  {
    return inMemory;
  }
  
  public void setInMemoryFlag(boolean inMemory)
  {
    this.inMemory = inMemory;
  }

}
