/*
 * NodeInfo.java
 *
 * Created on 25. listopad 2003, 19:05
 */

package SOFA.SOFAnet.Repository;

import java.net.*;
import SOFA.Util.VMProperties;

/**
 * Representation of name of node.
 *
 * @author  Ladislav Sobr
 */
public class NodeInfo
{
  private String name;
  private InetAddress address;
  private int port;
  
  public static class InvalidNodeNameException extends SOFA.SOFAnode.SOFAException
  {
    public InvalidNodeNameException(String message)
    {
      super(message);
    }

    public InvalidNodeNameException(String message, Throwable cause)
    {
      super(message, cause);
    }

    public InvalidNodeNameException(Throwable cause)
    {
      super(cause);
    }
  }
  
  /**
   * Returns representation of node "localhost"
   */
  public static NodeInfo getLocalhost()
  {
    String localhost;
    try
    {
      localhost = InetAddress.getLocalHost().getHostName();
    }
    catch (UnknownHostException e)
    {
      localhost = "localhost";
    }
    
    NodeInfo nodeInfo = new NodeInfo();
    
    try
    {
      nodeInfo.setNodeName(localhost);
    }
    catch (InvalidNodeNameException e)
    {
    }
    
    return nodeInfo;
  }
    
  /** Returns node name of local node */
  public static String getLocalNodeName()
  {
    String nodeName = System.getProperty(VMProperties.NODE_NAME).trim();
    if (!nodeName.toLowerCase().startsWith("sofa://")) nodeName = "sofa://" + nodeName;
    return nodeName;
  }
  
  /** Creates a new instance of NodeInfo */
  public NodeInfo()
  {
    port = -1;
  }
  
  /**
   * Sets name of node.
   * <p>
   * Node name should be in format (text in [] is optional):
   * <p>
   * &nbsp;&nbsp;&nbsp;&nbsp;   [sofa://]ip_address_or_host_name[:port]
   *
   * @param nodeName name of node
   */
  public void setNodeName(String nodeName) throws InvalidNodeNameException
  {
    name = nodeName;
    
    nodeName = nodeName.trim();
    if (nodeName.toLowerCase().startsWith("sofa://"))
        nodeName = nodeName.substring(7).trim();

    if (nodeName.length() <= 0) throw new InvalidNodeNameException("Empty node name");
      
    port = -1;

    int i = nodeName.indexOf(':');

    if (i != -1)
    {
      try
      {
        port = Integer.parseInt(nodeName.substring(i + 1));
      }
      catch (NumberFormatException e)
      {
        throw new NodeInfo.InvalidNodeNameException("Invalid node name format (cannot read port)", e);
      }
      nodeName = nodeName.substring(0, i);
    }

    try
    {
      address = InetAddress.getByName(nodeName);
    }
    catch (UnknownHostException e)
    {
      throw new NodeInfo.InvalidNodeNameException("Unknown host '" + nodeName + "'", e);
    }
  }
  
  /**
   * Tests wheter internal node name can pass specified pattern.
   * <p>
   * Pattern can be in one of following formats (text in [] is optional):
   * <ul>
   * <li>a) concrete host (also port can be present)
   *   <ul>
   *   <li> [sofa://]ip_address_or_host_name[:port]
   *   <li> Example:
   *     <ul>
   *     <li> sofa://192.168.0.1
   *     <li> nenya:1000
   *     </ul>
   *   </ul>
   * <li>b) domain (also port can be present)
   *   <ul>
   *   <li> [sofa://]domain_name[:port]
   *   <li> Example:
   *     <ul>
   *     <li>sofa://cuni.cz
   *     <li>mff.cuni.cz:1000
   *     </ul>
   *   </ul>
   * <li>c) part_of_IPv4_address/number_of_mask_bits[:port]
   *   <ul>
   *   <li> Example:
   *     <ul>
   *     <li>192.168.20/22:2000
   *     </ul>
   *   <li> Mask bits can be ommited:
   *     <ul>
   *     <li>192.168.0.1:2000
   *     <li>in this case mask bits are considered to be 32 (full IPv4 address)
   *     </ul>
   *   </ul>
   * <li>d) [sofa://]* &nbsp;&nbsp; - asterix means any host
   * </ul>
   * <p>
   * If "port" is present (in any formats), the node match the pattern only
   * if the same port is specified in the node.
   *
   * @param nodeNamePattern pattern to test
   * @return true, if the internal node name matches the pattern
   */
  
  public boolean matchPattern(String nodeNamePattern) throws InvalidNodeNameException
  {
    if (address == null) return false;
    
    //remove possible leading string  sofa://
    
    nodeNamePattern = nodeNamePattern.trim();
    if (nodeNamePattern.toLowerCase().startsWith("sofa://"))
        nodeNamePattern = nodeNamePattern.substring(7).trim();

    if (nodeNamePattern.compareTo("*") == 0) return true;

    //try to find port part of pattern and compare it to node port
    
    int i = nodeNamePattern.indexOf(':');

    if (i != -1)
    {
      int patternPort = -1;
      try
      {
        patternPort = Integer.parseInt(nodeNamePattern.substring(i + 1));
      }
      catch (NumberFormatException e)
      {
        throw new NodeInfo.InvalidNodeNameException("Invalid node name pattern (cannot read port)", e);
      }

      //check equality or pattern port and node port
      if (patternPort != port) return false;
      
      //remove port from pattern
      nodeNamePattern = nodeNamePattern.substring(0, i);
    }


    //try to check this format:   part_of_ip_address/number_of_mask_bits:port
    //example:                    192.168.20/22:2000
    
    i = nodeNamePattern.indexOf('/');
    
    if (i != -1)        //format found
    {
      byte [] ipAddr = address.getAddress();
      if (ipAddr.length != 4) return false;     //we support just IPv4 addresses
      
      int maskBits = 32;
      
      try
      {
        maskBits = Integer.parseInt(nodeNamePattern.substring(i + 1));
      }
      catch (NumberFormatException e)
      {
        throw new NodeInfo.InvalidNodeNameException("Invalid node name pattern (cannot read number of mask bits)", e);
      }
      nodeNamePattern = nodeNamePattern.substring(0, i);
    
      String [] ipParts = nodeNamePattern.split("\\.");
      if (ipParts.length > 4)
          throw new NodeInfo.InvalidNodeNameException("Invalid node name pattern (invalid \"IP address\" part)");

      for (i = 0; i < ipParts.length; i++)
      {
        int number;
        int bits = 0;
        try
        {
          number = Integer.parseInt(ipParts[i]);
        }
        catch (NumberFormatException e)
        {
          throw new NodeInfo.InvalidNodeNameException("Invalid node name pattern (cannot read \"IP address\" part)", e);
        }

        if (maskBits >= 8) bits = 8;
        else bits = maskBits;
        maskBits -= bits;
        if (bits == 0) return true;

        int mask = 256 - (1 << (8 - bits));
        if ((number & mask) != (ipAddr[i] & mask)) return false;
      }

      if (maskBits > 0)
          throw new NodeInfo.InvalidNodeNameException("Invalid node name pattern (incomplete \"IP address\" part or to many mask bits)");
      
      return true;
    }

    //try to compare domains of node and pattern
    String nodeHost = address.getCanonicalHostName();
    if (nodeHost.endsWith("." + nodeNamePattern)) return true;
    
    //try to compare IP addresses of pattern and node
    try
    {
      InetAddress addr = InetAddress.getByName(nodeNamePattern);
      if (addr.equals(address)) return true;
    }
    catch (UnknownHostException e)
    {
    }

    return false;
  }
  
  /**
   * Tests whether internal node name has same host as the specified one.
   * @param host host to match
   * @return true if the internal node name has same host as the specified one.
   */
  public boolean isSameHost(String host) throws InvalidNodeNameException
  {
    if (address == null) throw new NodeInfo.InvalidNodeNameException("Empty address");
    
    try
    {
      InetAddress addr = InetAddress.getByName(host);
      if (addr.equals(address)) return true;
      else return false;
    }
    catch (UnknownHostException e)
    {
      throw new NodeInfo.InvalidNodeNameException(e.getMessage(), e);
    }
  }
    
  public String toString()
  {
//    if (address == null) return "";
//    else
//    {
//      String s = "sofa://" + address.getCanonicalHostName();
//      if (port != -1) s += ":" + port;
//      return s;
//    }
    return name;
  }
  
  public String getName()
  {
    return name;
  }
  
  public InetAddress getAddress()
  {
    return address;
  }
  
  public int getPort()
  {
    return port;
  }
  
  public String getAddressAndPort()
  {
    if (address == null) return "";
    if (port == -1) return address.getHostAddress();
    else return address.getHostAddress() + ":" + port;
  }
}
