/*
 * NodeNameFilter.java
 *
 * Created on 27. listopad 2003, 17:41
 */

package SOFA.SOFAnet.Repository;

import java.util.*;
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import SOFA.Util.XML;
import java.io.Serializable;


/**
 * Filter of node names.
 * <p>
 * Filter is implemented as list of patterns.
 * If name of node passes any of the patterns, it passes whole filter
 *
 * @author  Ladislav Sobr
 */
public class NodeNameFilter implements Cloneable, Serializable
{
  private List patterns;
  
  /** Creates a new instance of NodeNameFilter */
  public NodeNameFilter()
  {
    patterns = new LinkedList();
  }
  
  public Object clone()
  {
    NodeNameFilter clone = null;
    try
    { 
      clone = (NodeNameFilter)super.clone();
    }
    catch (CloneNotSupportedException e)
    { 
      throw new InternalError();
    }
    clone.patterns = Collections.synchronizedList(new LinkedList());
    clone.patterns.addAll(patterns);
    return clone;
  }
  
  public List getPatterns()
  {
    return patterns;
  }
  
  public boolean pass(String nodeName)
  {
    NodeInfo nodeInfo = new NodeInfo();
    try
    {
      nodeInfo.setNodeName(nodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      return false;
    }
    
    return pass(nodeInfo);
  }
  
  public boolean pass(NodeInfo nodeInfo)
  {
    Iterator it = patterns.iterator();
    while (it.hasNext())
    {
      String pattern = (String)it.next();
      try
      {
        if (nodeInfo.matchPattern(pattern)) return true;
      }
      catch (NodeInfo.InvalidNodeNameException e)
      {
      }
    }
    return false;
  }
  
  public void add(String pattern)
  {
    patterns.add(pattern);
  }

  public void add(NodeNameFilter nodeFilter)
  {
    if (nodeFilter != null)
    {
      List li = nodeFilter.getPatterns();
      synchronized (li)
      {
        Iterator it = li.iterator();
        while (it.hasNext())
        {
          String pattern = (String)it.next();
          patterns.add(pattern);
        }
      }
    }
  }
  
  /**
   * Loads the NodeNameFiler from the XML DOM tree.
   * <p>
   * XML format:
   * <p>
   * <pre>
   *  *&lt;filter pattern="string"/&gt;
   * </pre>
   */
  public void loadFromXML(Element element) throws SAXException
  {
    NodeList nl = element.getChildNodes();
    for (int i = 0; i < nl.getLength(); i++)
    {
      Node node = nl.item(i);
      if (node.getNodeType() == Node.ELEMENT_NODE && node.getNodeName().compareTo("filter") == 0)
      {
        Element e = (Element)node;
        String pattern = e.getAttribute("pattern");
        if (pattern.length() == 0) throw new SAXException("Missing 'pattern' attribute in 'filter' tag");
        patterns.add(pattern);
      }
    }
  }
  
  /**
   * Saves the NodeNameFilter to XML DOM tree.
   */
  public void saveToXML(Element element)
  {
    Iterator it = patterns.iterator();
    while (it.hasNext())
    {
      String pattern = (String)it.next();
      XML.newOneAttributeElement(element, "filter", "pattern", pattern);
    }
  }
}
