/*
 * NodeLicList.java
 *
 * Created on 3. prosinec 2003, 17:35
 */

package SOFA.SOFAnet.Repository;

import java.util.*;
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import java.io.Serializable;

/**
 * Implementation of list of node names (or node name patterns) and number of licences.
 *
 * @author  Ladislav Sobr
 */
public class NodeNameLicList implements Cloneable, Serializable
{
  private List list;
  
  public static class NodeLicPair implements Cloneable, Serializable
  {
    public String nodeName;
    public int numberOfLicences;    // Licence.ONE_IMPLICIT_LICENCE means "one licence", Licence.ALL_LICENCES means "all remaining licences"
    
    public NodeLicPair()
    {
      numberOfLicences = Licence.ONE_IMPLICIT_LICENCE;
    }
    
    public NodeLicPair(String nodeName, int numberOfLicences)
    {
      this.nodeName = nodeName;
      this.numberOfLicences = numberOfLicences;
    }
    
    public Object clone()
    {
      try
      {
        return super.clone();
      }
      catch (CloneNotSupportedException e)
      {
        throw new InternalError(); 
      }
    }
    
    public String toString()
    {
      if (numberOfLicences == Licence.ONE_IMPLICIT_LICENCE) return nodeName;
      else if (numberOfLicences == Licence.ALL_LICENCES) return nodeName + " [all licences]";
      else return nodeName + " [#licences: " + Integer.toString(numberOfLicences) + "]";
    }
  }

  
  /** Creates a new instance of NodeLicList */
  public NodeNameLicList()
  {
    list = Collections.synchronizedList(new LinkedList());
  }

  public Object clone()
  {
    NodeNameLicList clone = null;
    try
    { 
      clone = (NodeNameLicList)super.clone();
    }
    catch (CloneNotSupportedException e)
    { 
      throw new InternalError();
    }
    clone.list = Collections.synchronizedList(new LinkedList());
    clone.list.addAll(list);
    return clone;
  }
  
  public List getList()
  {
    return list;
  }
  
  /**
   * @param number_of_licences == Licence.ALL_LICENCES means "all of remaining licences"
   */
  public void add(String nodeName, int number_of_licences)
  {
    list.add(new NodeLicPair(nodeName, number_of_licences));
  }

  public void add(NodeNameLicList nodeNameLicList)
  {
    if (nodeNameLicList != null)
    {
      List li = nodeNameLicList.getList();
      synchronized (li)
      {
        Iterator it = li.iterator();
        while (it.hasNext())
        {
          NodeLicPair licPair = (NodeLicPair)it.next();
          list.add(new NodeLicPair(licPair.nodeName, licPair.numberOfLicences));
        }
      }
    }
  }

  public void add(NodeLicPair nodeLicPair)
  {
    list.add(nodeLicPair);
  }
  
  public NodeLicPair find(String nodeName)
  {
    synchronized (list)
    {
      Iterator it = list.iterator();
      while (it.hasNext())
      {
        NodeLicPair nlp = (NodeLicPair)it.next();
        if (nlp.nodeName.compareTo(nodeName) == 0) return nlp;
      }
    }
    
    return null;
  }
  
  /**
   * Test wheter specified name of node matches any of the patterns in the list.
   * @param nodeName name of node to test
   * @return NodeLicPair object representing pair of node name and number of licence copies
   */
  public NodeLicPair pass(String nodeName)
  {
    NodeInfo nodeInfo = new NodeInfo();
    try
    {
      nodeInfo.setNodeName(nodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      return null;
    }
    
    return pass(nodeInfo);
  }
  
  /**
   * Test wheter specified node matches any of the patterns in the list.
   * @param nodeInfo node to test
   * @return NodeLicPair object representing pair of node name and number of licence copies
   */
  public NodeLicPair pass(NodeInfo nodeInfo)
  {
    synchronized (list)
    {
      Iterator it = list.iterator();
      while (it.hasNext())
      {
        NodeLicPair nlp = (NodeLicPair)it.next();
        try
        {
          if (nodeInfo.matchPattern(nlp.nodeName)) return nlp;
        }
        catch (NodeInfo.InvalidNodeNameException e)
        {
        }
      }
      return null;
    }
  }
    
  
  /**
   * Loads the NodeNameLicList from the XML DOM tree.
   * <p>
   * XML format:
   * <p>
   * <pre>
   *  *&lt;node name="string" ?number_of_licences="positive-integer or *"/&gt;
   * </pre
   */
  public void loadFromXML(Element element) throws SAXException
  {
    synchronized (list)
    {
      NodeList nl = element.getChildNodes();
      for (int i = 0; i < nl.getLength(); i++)
      {
        Node node = nl.item(i);
        if (node.getNodeType() == Node.ELEMENT_NODE && node.getNodeName().compareTo("node") == 0)
        {
          Element el = (Element)node;
          NodeLicPair nlp = new NodeLicPair();
          nlp.nodeName = el.getAttribute("name");
          String s = el.getAttribute("number_of_licences");
          if (s.compareTo("*") == 0) nlp.numberOfLicences = Licence.ALL_LICENCES;
          else
          if (s.length() != 0)
          {
            try
            {
              int l = Integer.parseInt(s);
              if (l < 0) throw new NumberFormatException();
              nlp.numberOfLicences = l;
            }
            catch (NumberFormatException e)
            {
              throw new SAXException("Invalid 'positive-integer or *' format of 'number_of_licences' attribute of 'node' tag", e);
            }
          }

          list.add(nlp);
        }
      }
    }
  }
  
  /**
   * Saves the NodeNameLicList to XML DOM tree.
   */
  public void saveToXML(Element element)
  {
    synchronized (list)
    {
      Document doc = element.getOwnerDocument();

      Iterator it = list.iterator();
      while (it.hasNext())
      {
        NodeLicPair nlp = (NodeLicPair)it.next();

        Element el = doc.createElement("node");
        el.setAttribute("name", nlp.nodeName);
        String s = null;
        if (nlp.numberOfLicences == Licence.ALL_LICENCES) s = "*";
        else if (nlp.numberOfLicences > 0) s = Integer.toString(nlp.numberOfLicences);
        if (s != null) el.setAttribute("number_of_licences", s);
        element.appendChild(el);
      }
    }
  }
  
}
