/*
 * OutputTrigger.java
 *
 * Created on 27. listopad 2003, 22:35
 */

package SOFA.SOFAnet.Repository;

import java.io.*;
import java.util.*;
import java.text.*;
import SOFA.Util.XML;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

/**
 * Representation of the output trigger.
 * One item of persistent storage.
 * <p>
 * Output triggers are actually not "triggers", they are more filters. They specifies
 * which nodes can download wich bundles from the node. The pull request must pass at least one Output Trigger.
 * <p>
 * The output trigger contains:
 * <ul>
 * <li>text description of the trigger
 * <li>filter that specifies the bundles on which the trigger should be used
 * <li>filter of nodes that must be passed to use the trigger
 * <li>time validity
 * <li>(optional) ID and Rule of contract that is connected with the input trigger
 * </ul>
 *
 * @author  Ladislav Sobr
 */
public class OutputTrigger extends XMLStorageItem implements StorageItem, Serializable, Cloneable
{
  private String description;
  private boolean autoDelete;
  private BundleNameFilter bundleFilter;   //can be null
  private NodeNameFilter nodeFilter;       //can be null
  private Date validFrom;         //can be null
  private Date validTo;           //can be null
  private boolean isContract;
  private String contractID;
  private String contractRule;
  
  
  
  /** Creates a new instance of OutputTrigger */
  public OutputTrigger(String name, File file)
  {
    super(name, file, "output_trigger");
    reset();
  }
  
  public OutputTrigger()
  {
    super("", null, "output_trigger");
    reset();
  }
  
  public Object clone()
  {
    OutputTrigger clone = null;
    try
    {
      clone = (OutputTrigger)super.clone();
    }
    catch (CloneNotSupportedException e)
    {
      throw new InternalError(); 
    }
    if (bundleFilter != null) clone.bundleFilter = (BundleNameFilter)bundleFilter.clone();
    if (nodeFilter != null) clone.nodeFilter = (NodeNameFilter)nodeFilter.clone();
    return clone;
  }
  
  
  public boolean toBeDeleted()
  {
    return validTo != null && autoDelete && validTo.before(Calendar.getInstance().getTime());
  }
  
  /**
   * Resets content of the item.
   */
  protected void reset()
  {
    bundleFilter = null;
    nodeFilter = null;
    description = "";
    validFrom = null;
    validTo = null;
    autoDelete = false;
    isContract = false;
    contractID = "";
    contractRule = "";
  }

  /**
   * Loads the item from the XML DOM tree.
   * <p>
   * XML format of storage item:
   * <p>
   * <pre>
   * &lt;output_trigger&gt;
   *   OutputTrigger XML format
   * &lt;/output_trigger&gt;
   * </pre>
   * <p>
   * XML format:
   * <p>
   * <pre>
   *   &lt;description&gt;string&lt;/description&gt;
   *  ?&lt;bundle_filter&gt;
   *     BundleNameFilter XML format
   *   &lt;/bundle_filter&gt;
   *  ?&lt;node_filter&gt;
   *     NodeNameFilter XML format
   *   &lt;/node_filter&gt;
   *  ?&lt;validity ?from="date" ?to="date" ?autodelete="boolean"/&gt;
   *  ?&lt;contract cid="string" rule="string"/&gt;
   * </pre>
   *
   */
  public void loadFromXML(Element element) throws SAXException
  {
    reset();
    NodeList nl = element.getChildNodes();
    for (int i = 0; i < nl.getLength(); i++)
    {
      Node node = nl.item(i);
      if (node.getNodeType() == Node.ELEMENT_NODE)
      {
        Element el = (Element)node;
        String nodeName = node.getNodeName();
        
        if (nodeName.compareTo("description") == 0)
        {
          description = XML.getTextContent(el);
        }
        else
        if (nodeName.compareTo("bundle_filter") == 0)
        {
          bundleFilter = new BundleNameFilter();
          bundleFilter.loadFromXML(el);
        }
        else
        if (nodeName.compareTo("node_filter") == 0)
        {
          nodeFilter = new NodeNameFilter();
          nodeFilter.loadFromXML(el);
        }
        else
        if (nodeName.compareTo("validity") == 0)
        {
          String from = el.getAttribute("from");
          String to = el.getAttribute("to");
          String ad = el.getAttribute("autodelete");
          DateFormat dateFormat = DateFormat.getInstance();
          if (from.length() > 0)
          {
            try
            {
              validFrom = dateFormat.parse(from);
            }
            catch (ParseException e)
            {
              throw new SAXException("Cannot parse date/time format of 'from' attribute of 'validity' tag", e);
            }
          }
          
          if (to.length() > 0)
          {
            try
            {
              validTo = dateFormat.parse(to);
            }
            catch (ParseException e)
            {
              throw new SAXException("Cannot parse date/time format of 'to' attribute of 'validity' tag", e);
            }
          }
          
          if (ad.length() > 0)
          {
            ad = ad.trim().toLowerCase();
            if (ad.compareTo("1") == 0 || ad.compareTo("true") == 0) autoDelete = true;
          }
        }
        else
        if (nodeName.compareTo("contract") == 0)
        {
          isContract = true;
          contractID = el.getAttribute("cid");
          contractRule = el.getAttribute("rule");
        }
      }
    }
  }
  
  /**
   * Saves the item to XML DOM tree.
   */
  public void saveToXML(Element element)
  {
    Document doc = element.getOwnerDocument();
    Element el;
    
    XML.newTextElement(element, "description", description);
    
    if (bundleFilter != null)
    {
      el = doc.createElement("bundle_filter");
      bundleFilter.saveToXML(el);
      element.appendChild(el);
    }
    
    if (nodeFilter != null)
    {
      el = doc.createElement("node_filter");
      nodeFilter.saveToXML(el);
      element.appendChild(el);
    }
    
    if (validFrom != null || validTo != null || autoDelete)
    {
      DateFormat dateFormat = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.MEDIUM);
      el = doc.createElement("validity");
      if (validFrom != null) el.setAttribute("from", dateFormat.format(validFrom));
      if (validTo != null) el.setAttribute("to", dateFormat.format(validTo));
      if (autoDelete) el.setAttribute("autodelete", "true");
      element.appendChild(el);
    }
    
    if (isContract)
    {
      el = doc.createElement("contract");
      el.setAttribute("cid", contractID);
      if (contractRule.length() > 0) el.setAttribute("rule", contractRule);
      element.appendChild(el);
    }
  }
  
  public String getDescription()
  {
    return description;
  }

  public boolean getAutoDelete()
  {
    return autoDelete;
  }
  
  public BundleNameFilter getBundleFilter()
  {
    return bundleFilter;
  }
  
  public NodeNameFilter getNodeFilter()
  {
    return nodeFilter;
  }
  
  public Date getValidFrom()         //can be null
  {
    return validFrom;
  }
  
  public Date getValidTo()           //can be null
  {
    return validTo;
  }
  
  public boolean isValid()
  {
    Date now = Calendar.getInstance().getTime();
    return (validFrom == null || !validFrom.after(now)) && (validTo == null || !validTo.before(now));
  }  

  public boolean isContract()
  {
    return isContract;
  }
  
  public String getContractID()
  {
    return contractID;
  }
  
  public String getContractRule()
  {
    return contractRule;
  }

  public void setDescription(String description)
  {
    this.description = description;
  }
  
  public void setAutoDelete(boolean autoDelete)
  {
    this.autoDelete = autoDelete;
  }
  
  public void setBundleFilter(BundleNameFilter bundleFilter)
  {
    this.bundleFilter = bundleFilter;
  }
  
  public void setNodeFilter(NodeNameFilter nodeFilter)
  {
    this.nodeFilter = nodeFilter;
  }
  
  public void setValidFrom(Date validFrom)         //can be null
  {
    this.validFrom = validFrom;
  }
  
  public void setValidTo(Date validTo)           //can be null
  {
    this.validTo = validTo;
  }
  
  public void setContract(String cid, String rule)
  {
    if (cid == null && rule == null)
    {
      contractID = "";
      contractRule = "";
      isContract = false;
    }
    else
    {
      contractID = cid;
      contractRule = rule;
      isContract = true;
    }
  }
  
}
