/*
 * RMISearchCenter.java
 *
 * Created on 29. duben 2004, 16:08
 */

package SOFA.SOFAnet.Search.RMI;

import SOFA.SOFAnet.Search.*;
import SOFA.SOFAnet.Core.SearchOps;
import SOFA.SOFAnet.Repository.Repository;
import SOFA.SOFAnet.Repository.NodeNameFilter;
import SOFA.SOFAnet.Repository.NodeInfo;
import java.util.*;
import java.rmi.RemoteException;
import java.rmi.Naming;
import java.net.MalformedURLException;

/**
 * Managing class of RMI implementation of search subsystem of the SOFAnet.
 * <p>
 * It handles:
 * <ul>
 * <li>loading of search configuration
 * <li>search connections to other nodes
 * <li>creation of RMI server and client
 * <li>"collectors" - objects that collects replies for search requests.
 * </ul>
 *
 * @author  Ladislav Sobr
 */
public class RMISearchCenter
{
  private Repository rep;
  private SearchOps searchOps;
  
  private RMISearchClient client;
  private RMISearchServer server;
  
  private List rmiSearchConnections;
  private RMISearchConfig config;
  private List seenSearchRequestIDs;
  private Map rmiSearchCollectors;
  boolean started;
  private Random randomGenerator;
  private String myNodeName;
  
  /** Creates a new instance of RMISearchCenter */
  public RMISearchCenter()
  {
    started = false;
    
    client = new RMISearchClient(this);
    server = null;
    
    rmiSearchConnections = Collections.synchronizedList(new LinkedList());
    seenSearchRequestIDs = Collections.synchronizedList(new LinkedList());
    rmiSearchCollectors = Collections.synchronizedMap(new HashMap());
    
    randomGenerator = new Random();
    
    myNodeName = NodeInfo.getLocalNodeName();
  }
  
  public void setRepository(Repository repository)
  {
    rep = repository;
  }

  public void setSearchOps(SearchOps searchOps)
  {
    this.searchOps = searchOps;
  }

  public SearchInterface getSearchInterface()
  {
    return client;
  }
  
  /** Starts RMI search subsystem */
  public void start() throws RMISearchException
  {
    //load configuration
    config = new RMISearchConfig(rep, rmiSearchConnections);
    config.loadFromStorage(true);
    
    started = true;
    
    //connect RMI search connections
    synchronized (rmiSearchConnections)
    {
      Iterator it = rmiSearchConnections.iterator();
      while (it.hasNext())
      {
        RMISearchConnection connection = (RMISearchConnection)it.next();
        connection.connect();
      }
    }
    
    //run RMI search server
   
    NodeInfo myNodeInfo = new NodeInfo();
    try
    {
      myNodeInfo.setNodeName(myNodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      throw new RMISearchException("Invalid name of local SOFA node: " + myNodeName, e);
    }
    
    try
    {
      server = new RMISearchServer(this, searchOps, myNodeInfo);
      Naming.rebind("//" + myNodeInfo.getAddressAndPort() + "/SOFAnet/RMISearchServer", server);
    }
    catch (RemoteException e)
    {
      throw new RMISearchException("RMI search server failed to start", e);
    }
    catch (MalformedURLException e)
    {
      throw new RMISearchException("RMI search server failed to start", e);
    }
    
  }
  
  /** Returns configuration of the RMI search subsystem */
  public RMISearchConfiguration getConfiguration()
  {
    return config.getConfiguration();
  }
  
  /** Sets configuration of the RMI search subsystem */
  public void setConfiguration(RMISearchConfiguration configuration)
  {
    config.setConfiguration(configuration);
  }
  
  /**
   * Reloads configuration of the RMI search subsystem and makes it active.
   * <p>
   * Old connections are not destroyed, the new one are created.
   */
  public void applyConfiguration()
  {
    config.loadFromStorage(false);
    
    //connect new RMI search connections
    synchronized (rmiSearchConnections)
    {
      Iterator it = rmiSearchConnections.iterator();
      while (it.hasNext())
      {
        RMISearchConnection connection = (RMISearchConnection)it.next();
        if (!connection.isConnected()) connection.connect();
      }
    }
    
  }
  
  /** Returns list with search connections (RMISearchConnection) to other nodes */
  List getRMISearchConnections()
  {
    return rmiSearchConnections;
  }
  
  boolean isStarted()
  {
    return started;
  }
  
  /** Returns node filter that restricts which nodes can connect to this node */
  NodeNameFilter getNodeFilter()
  {
    if (!started) return null;
    return config.getNodeFilter();
  }
  
  /** Returns list of seen search request IDs (SearchRequestID) - it is used to stop flood routing */
  List getSeenSearchRequestIDs()
  {
    return seenSearchRequestIDs;
  }
  
  /** Returns new collector for new search request */
  RMISearchCollector newCollector()
  {
    int requestMark;
    
    synchronized (rmiSearchCollectors)
    {
      do
      {
        requestMark = randomGenerator.nextInt();
      } while (rmiSearchCollectors.get(new Integer(requestMark)) != null);
      
      RMISearchCollector collector = new RMISearchCollector(this, requestMark, myNodeName);
      rmiSearchCollectors.put(new Integer(requestMark), collector);
      return collector;
    }
  }
  
  /** Returns existing collector that belongs to concrete search request */
  RMISearchCollector getCollector(int requestMark)
  {
    return (RMISearchCollector)rmiSearchCollectors.get(new Integer(requestMark));
  }
  
  /** Deletes colector for concrete search request */
  void removeCollector(int requestMark)
  {
    rmiSearchCollectors.remove(new Integer(requestMark));
  }
}
