/*
 * RMISearchCollector.java
 *
 * Created on 30. duben 2004, 14:38
 */

package SOFA.SOFAnet.Search.RMI;

import SOFA.SOFAnet.Search.*;
import java.util.*;

/**
 * The object sends search request to the search network and gathers the search replies/
 *
 * @author  Ladislav Sobr
 */
public class RMISearchCollector
{
  private RMISearchCenter center;
  private int requestMark;
  private boolean finished;
  private SearchRequestID requestID;
  private List replies;
  
  private SearchPattern searchPattern;
  private boolean firstOnly;
  private String shareGroup;
  private int timeout;
  
  /** Creates a new instance of RMISearchCollector */
  public RMISearchCollector(RMISearchCenter center, int requestMark, String myNodeName)
  {
    this.center = center;
    this.requestMark = requestMark;
    requestID = new SearchRequestID(requestMark, myNodeName);
    replies = new LinkedList();
    finished = false;
  }
  
  /** Returns identification of collector (and its search request) on the local node */
  public int getRequestMark()
  {
    return requestMark;
  }
  
  /** Sets the collector up. See SearchRequest class */
  public void setup(SearchPattern searchPattern, boolean firstOnly, String shareGroup, int timeout)
  {
    this.searchPattern = searchPattern;
    this.firstOnly = firstOnly;
    this.shareGroup = shareGroup;
    this.timeout = timeout;
  }
  
  /**
   * Sends the request to the search network and returns arrived replies
   * @return list of SearchReplyItem objects
   */
  public synchronized List request()
  {
    if (finished) return replies;
    
    //add request to the seen requests
    center.getSeenSearchRequestIDs().add(requestID);
    
    //send request to other nodes
    List rmiSearchConnections = center.getRMISearchConnections();
    synchronized (rmiSearchConnections)
    {
      Iterator it = rmiSearchConnections.iterator();
      while (it.hasNext())
      {
        RMISearchConnection connection = (RMISearchConnection)it.next();
        connection.sendRequest(requestID, searchPattern, firstOnly, shareGroup);
      }
    }
    
    long startTime = System.currentTimeMillis();
    
    for (;;)
    {
      long currentTime = System.currentTimeMillis();
      long t = timeout - (currentTime - startTime);
      if (t < 1) t = 100;
      
      try
      {
        wait(t);
      }
      catch (InterruptedException e)
      {
      }
      
      currentTime = System.currentTimeMillis();
      if (firstOnly && replies.size() > 0 || currentTime - startTime > timeout)
      {
        finished = true;
        return replies;
      }
    }
  }
  
  /**
   * Adds new replies to the collector
   * @param newReplies list of SearchReplyItem objects
   */
  public synchronized void reply(List newReplies)
  {
    if (finished) return;
    if (newReplies == null || newReplies.size() == 0) return;
    
    if (firstOnly)
    {
      if (replies.size() > 0) return;
      
      replies.add(newReplies.get(0));
      notify();
      return;
    }
    else replies.addAll(newReplies);
  }
  
}
