/*
 * RMISearchServer.java
 *
 * Created on 29. duben 2004, 16:09
 */

package SOFA.SOFAnet.Search.RMI;

import java.rmi.server.UnicastRemoteObject;
import java.rmi.server.ServerNotActiveException;
import java.rmi.RemoteException;
import SOFA.SOFAnet.Repository.NodeInfo;
import SOFA.SOFAnet.Repository.NodeNameFilter;
import SOFA.SOFAnet.Repository.ShareGroups;
import SOFA.SOFAnet.Search.*;
import SOFA.SOFAnet.Core.SearchOps;
import java.util.*;

/**
 * RMI implementation of search subsytem - server part.
 *
 * @author  Ladislav Sobr
 */
public class RMISearchServer extends UnicastRemoteObject implements RMISearchInterface
{
  private RMISearchCenter center;
  private SearchOps searchOps;
  private NodeInfo myNodeInfo;

  /** Creates a new instance of RMISearchServer */
  public RMISearchServer(RMISearchCenter center, SearchOps searchOps, NodeInfo myNodeInfo) throws RemoteException
  {
    this.center = center;
    this.searchOps = searchOps;
    this.myNodeInfo = myNodeInfo;
  }
  
  /**
   * Connects remote node to us.
   * <p>
   * The server checks whether the connection in opposite direction exists and if not, the opposite connection is created
   */
  public void connect(String nodeName, ShareGroups shareGroups) throws RMISearchException, RemoteException
  {
    NodeInfo nodeInfo = testNodeName(nodeName);
    
    //test whether the calling node is allowed here
    NodeNameFilter nodeFilter = center.getNodeFilter();
    if (nodeFilter != null && !nodeFilter.pass(nodeName))
    {
      throw new RMISearchException("Access to the search subsystem not allowed");
    }
    
    //do we have this connection?
    boolean connectionFound = false;
    List rmiSearchConnections = center.getRMISearchConnections();
    synchronized (rmiSearchConnections)
    {
      Iterator it = rmiSearchConnections.iterator();
      while (it.hasNext())
      {
        RMISearchConnection connection = (RMISearchConnection)it.next();
        if (connection.isNode(nodeName))
        {
          connectionFound = true;
          break;
        }
      }
    }
    
    //if not, create it
    if (!connectionFound)
    {
      RMISearchConnection connection = new RMISearchConnection(nodeInfo, shareGroups, false);
      connection.connect();
      rmiSearchConnections.add(connection);
    }
  }
  
  /**
   * Processes the search request
   * <p>
   * If the search request already arrived to the node, it is discarded. But if not,
   * the request is sent to all connections (flood routing) from the node (If the request has shareGroup filled, only connections with such shareGroup are used).
   * Then the request is evaluated on this node (this evaluation routine is responsible for replying).
   */
  public void request(String nodeName, SearchRequestID requestID, SearchPattern searchPattern, boolean firstOnly, String shareGroup) throws RMISearchException, RemoteException
  {
    testNodeName(nodeName);
    
    //test whether the calling node is allowed here
    NodeNameFilter nodeFilter = center.getNodeFilter();
    if (nodeFilter != null && !nodeFilter.pass(nodeName))
    {
      throw new RMISearchException("Access to the search subsystem not allowed");
    }
    
    //haven't we seen the request already?
    List seenSearchRequestIDs = center.getSeenSearchRequestIDs();
    
    if (seenSearchRequestIDs.contains(requestID))
    {
      //we have already processed this request
      return;
    }

    //ok we will process the request
    
    //add request to seen requests
    seenSearchRequestIDs.add(requestID);
    if (seenSearchRequestIDs.size() > 100) seenSearchRequestIDs.remove(0); //we remember only last 100 seen requests
    
    //send request to other nodes
    List rmiSearchConnections = center.getRMISearchConnections();
    synchronized (rmiSearchConnections)
    {
      Iterator it = rmiSearchConnections.iterator();
      while (it.hasNext())
      {
        RMISearchConnection connection = (RMISearchConnection)it.next();
        if (!connection.isNode(nodeName))      //do not send to the sending node
        {
          connection.sendRequest(requestID, searchPattern, firstOnly, shareGroup);
        }
      }
    }

    //evaluate request
    searchOps.evaluateRequest(requestID, searchPattern, firstOnly, shareGroup);
  }

  /**
   * Receives reply to the request.
   * The appropriate collector of the request is found and the reply is handed over to it
   */
  public void reply(String nodeName, SearchReply searchReply) throws RMISearchException, RemoteException
  {
    testNodeName(nodeName);
    
    if (myNodeInfo.getName().compareTo(searchReply.getRequestID().getSourceNode()) != 0)
    {
      throw new RMISearchException("Search reply arrived at node '" + myNodeInfo.getName() + "', but request was made by node'" + searchReply.getRequestID().getSourceNode() + "'");
    }
    
    List replies =  searchReply.getReplies();
    if (replies == null || replies.size() == 0)
    {
      throw new RMISearchException("Empty search replies");
    }
    
    RMISearchCollector collector = center.getCollector(searchReply.getRequestID().getRequestMark());
    
    if (collector != null)
    {
      //probablythere is still someone waiting for replies
      collector.reply(replies);
    }
  }
  
  /** Tests wheher specified name of node corresponds to the real host of RMI client */
  private NodeInfo testNodeName(String nodeName) throws RMISearchException
  {
    NodeInfo nodeInfo = new NodeInfo();
    try
    {
      nodeInfo.setNodeName(nodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      throw new RMISearchException("Invalid name of SOFA node: " + nodeName, e);
    }

    String realHost = "";
    
    try
    {
      realHost = getClientHost();
    }
    catch (ServerNotActiveException e)
    {
      throw new RMISearchException(e.getMessage(), e);
    }      
    
    boolean ok = false;
    try
    {
      ok = nodeInfo.isSameHost(realHost);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      throw new RMISearchException("Invalid RMI client host: " + nodeName, e);
    }
    
    if (!ok) throw new RMISearchException("Name of SOFA node (" + nodeName + ") and RMI client host (" + realHost + ") are incosistent");
    
    return nodeInfo;
  }
}
