/*
 * BundleData.java
 *
 * Created on 31. bezen 2004, 16:23
 */

package SOFA.SOFAnet.Transport;

import java.io.*;
import java.util.*;
import SOFA.Util.Tools;
//import cz.cuni.sofa.lib.Impl.ByteArrayOutputStream;

/**
 * BundleData objects represents binary bundles on the interface between transport layer
 * and the rest of the system.
 * <p>
 * The bundle is usually physically located in the file in the temp directory of the system (as jar file).
 * When the BundleData object is destroyed, its internal file is deleted.
 *
 * @author  Ladislav Sobr
 */
public class BundleData
{
  private File tempFile;
  
  /**
   * Creates a new instance of BundleData from file.
   * <p>
   * If error occures, getFile() will return null
   *
   * @param sourceFile the file with bundle
   * @param copy if true, sourceFile is copied to the temp and this copy is managed by the object; otherwise sourceFile becomes internal file of BundleData (no copying)
   */
  public BundleData(File sourceFile, boolean copy)
  {
    if (copy)
    {
      tempFile = new File(generateFilename());
      
      try
      {
        FileInputStream in = new FileInputStream(sourceFile);
        FileOutputStream out = new FileOutputStream(tempFile);
        Tools.copyStream(in, out);
        in.close();
        out.close();
      }
      catch (FileNotFoundException e)
      {
        delete();
      }
      catch (IOException e)
      {
        delete();
      }
    }
    else
    {
      this.tempFile = sourceFile;
    } 
  }
  
  /**
   * Creates a new instance of BundleData from byte array.
   * <p>
   * The byte array is saved to file in temp,
   * If error occures, getFile() will return null
   */
  public BundleData(byte[] data)
  {
    if (data == null)
    {
      delete();
      return;
    }
    
    tempFile = new File(generateFilename());

    try
    {
      ByteArrayInputStream in = new ByteArrayInputStream(data);
      FileOutputStream out = new FileOutputStream(tempFile);
      Tools.copyStream(in, out);
      in.close();
      out.close();
    }
    catch (FileNotFoundException e)
    {
      delete();
    }
    catch (IOException e)
    {
      delete();
    }
  }
  
  /** Deletes internal file */
  public void delete()
  {
    if (tempFile != null)
    {
      tempFile.delete();
      tempFile = null;
    }
  }
  
  /** Releases reference to internal file without deleting it */
  public void release()
  {
    tempFile = null;
  }
  
  /** Returns reference to the internal file */
  public File getFile()
  {
    return tempFile;
  }
  
  /**
   * Returns bundle in the form of byte array
   * @return can be null if internal file doesn't exist or an error occures during loading to memory
   */
  public byte[] getData()
  {
    if (tempFile == null) return null;
    
    try
    {
      FileInputStream is = new FileInputStream(tempFile);
      ByteArrayOutputStream os = new ByteArrayOutputStream();
      Tools.copyStream(is, os);
      is.close();
      byte[] result = os.toByteArray();
      os.close();
      return result;
    }
    catch (FileNotFoundException e)
    {
      return null;
    }
    catch (IOException e)
    {
      return null;
    }
  }
  
  /**
   * Copies content of BundleData (it's inetnal file) to other file.
   *
   * @param destFile destination file
   * @param deleteOnError if true, destination file is deleted if an error occures
   */
  public boolean copyToFile(File destFile, boolean deleteOnError)
  {
    if (destFile == null) return false;
    if (tempFile == null)
    {
      if (deleteOnError) destFile.delete();
      return false;
    }
    
    try
    {
      FileInputStream in = new FileInputStream(tempFile);
      FileOutputStream out = new FileOutputStream(destFile);
      Tools.copyStream(in, out);
      in.close();
      out.close();
    }
    catch (FileNotFoundException e)
    {
      if (deleteOnError) destFile.delete();
      return false;
    }
    catch (IOException e)
    {
      if (deleteOnError) destFile.delete();
      return false;
    }
    
    return true;
  }
  
  /** Deletes internal file when the object is destroyed */
  protected void finalize () throws Throwable
  {
    delete();
    super.finalize();
  }

  /** Generates filename for the internal file of BundleData - it is located in temp directory of the system */
  public static String generateFilename()
  {
    return System.getProperty("java.io.tmpdir") + File.separator + "sofabundle" + Long.toString(new Date().getTime()) + ".jar";
  }
}
