/*
 * RMITransportClient.java
 *
 * Created on 8. duben 2004, 14:59
 */

package SOFA.SOFAnet.Transport.RMI;

import java.rmi.*;
import java.util.*;
import SOFA.SOFAnet.Core.NetOps;
import SOFA.SOFAnet.Core.ShareOps;
import SOFA.SOFAnet.Transport.*;
import SOFA.SOFAnet.Repository.NodeInfo;

/**
 * RMI Implementation of TransportInterface - client part of transport subsystem.
 * <p>
 * It creates connections to the other nodes, when needed.
 * These conncetions are kept alive certain time after last operation on them.
 * <p>
 * The class is also responsible for splitting multi-direction "push" operation into single-direction connections.
 *
 * @author  Ladislav Sobr
 */
public class RMITransportClient implements TransportInterface
{
  private Map nodeMap;
  private NetOps netOps;
  private ShareOps shareOps;
  
  /** Creates a new instance of RMITransportClient */
  public RMITransportClient()
  {
    System.setSecurityManager(new RMISecurityManager());
    nodeMap = Collections.synchronizedMap(new HashMap());
    netOps = null;
    shareOps = null;
  }
  
  public void testPush(IOParams params) throws TransportException
  {
    RMITransportNodeConnection nodeConnection;
    
    nodeConnection = getNodeConnection(params.getNodeName());
    nodeConnection.testPush(params);
    
    List otherNodes = params.getOtherNodes();
    if (otherNodes != null && otherNodes.size() != 0)
    {
      IOParams ioParams = new IOParams();
      ioParams.setBundleName(params.getBundleName());
      ioParams.setOffer(params.isOffer());
      
      Iterator it = otherNodes.iterator();
      while (it.hasNext())
      {
        IOParams.OtherNode otherNode = (IOParams.OtherNode)it.next();
        
        nodeConnection = getNodeConnection(otherNode.nodeName);
        nodeConnection.testPush(ioParams);
        otherNode.errCode = ioParams.getErrCode();
      }
      
      ioParams = null;
    }
  }
  
  public void push(IOParams params) throws TransportException
  {
    RMITransportNodeConnection nodeConnection;
    
    for (;;)
    {
      nodeConnection = getNodeConnection(params.getNodeName());
      if (nodeConnection.push(params)) break;
    }
    
    List otherNodes = params.getOtherNodes();
    if (otherNodes != null && otherNodes.size() != 0)
    {
      Iterator it = otherNodes.iterator();
      while (it.hasNext())
      {
        IOParams.OtherNode otherNode = (IOParams.OtherNode)it.next();

        IOParams ioParams = new IOParams();
        ioParams.setBundleName(params.getBundleName());
        ioParams.setBundleData(params.getBundleData());
        ioParams.setOffer(params.isOffer());
        ioParams.setPriority(params.getPriority());
        ioParams.setLicence(otherNode.licence);
        
        for (;;)
        {
          nodeConnection = getNodeConnection(otherNode.nodeName);
          if (nodeConnection.push(params)) break;
        }
        
        ioParams = null;
      }
    }
  }
  
  public void testPull(IOParams params) throws TransportException
  {
    RMITransportNodeConnection nodeConnection = getNodeConnection(params.getNodeName());
    nodeConnection.testPull(params);
  }

  public void pull(IOParams params) throws TransportException
  {
    for (;;)
    {
      RMITransportNodeConnection nodeConnection = getNodeConnection(params.getNodeName());
      if (nodeConnection.pull(params)) break;
    }
  }
  
  public void synchroPull(IOParams params) throws TransportException
  {
    RMITransportNodeConnection nodeConnection = getNodeConnection(params.getNodeName());
    nodeConnection.synchroPull(params);
  }
  
  public void testAcquireShared(IOParams params) throws TransportException
  {
    RMITransportNodeConnection nodeConnection = getNodeConnection(params.getNodeName());
    nodeConnection.testAcquireShared(params);
  }

  public void acquireShared(IOParams params) throws TransportException
  {
    RMITransportNodeConnection nodeConnection = getNodeConnection(params.getNodeName());
    nodeConnection.acquireShared(params);
  }  
  
  public void canReturnShared(IOParams params) throws TransportException
  {
    RMITransportNodeConnection nodeConnection = getNodeConnection(params.getNodeName());
    nodeConnection.canReturnShared(params);
  }
  
  public void returnShared(IOParams params) throws TransportException
  {
    RMITransportNodeConnection nodeConnection = getNodeConnection(params.getNodeName());
    nodeConnection.returnShared(params);
  }
  
  public void manualReturnShared(IOParams params) throws TransportException
  {
    RMITransportNodeConnection nodeConnection = getNodeConnection(params.getNodeName());
    nodeConnection.manualReturnShared(params);
  }
  
  public void setNetOps(NetOps netOps)
  {
    this.netOps = netOps;
  }
  
  public void setShareOps(ShareOps shareOps)
  {
    this.shareOps = shareOps;
  }
  
  NetOps getNetOps()
  {
    return netOps;
  }

  ShareOps getShareOps()
  {
    return shareOps;
  }

  private synchronized RMITransportNodeConnection getNodeConnection(String nodeName) throws TransportException
  {
    NodeInfo nodeInfo = new NodeInfo();
    try
    {
      nodeInfo.setNodeName(nodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      throw new TransportException("Invalid name of SOFA node: " + nodeName, e);
    }
    
    String addrPortString = nodeInfo.getAddressAndPort();
    RMITransportNodeConnection nodeConnection = (RMITransportNodeConnection)nodeMap.get(addrPortString);
    
    if (nodeConnection == null)
    {
      nodeConnection = new RMITransportNodeConnection(nodeInfo, this);
      
      nodeMap.put(addrPortString, nodeConnection);
      return nodeConnection;
    }
    else return nodeConnection;
  }

  synchronized void removeFromMap(String addrPortString, RMITransportNodeConnection nodeConnection)
  {
    RMITransportNodeConnection myNodeConnection = (RMITransportNodeConnection)nodeMap.get(addrPortString);
    if (myNodeConnection == nodeConnection) nodeMap.remove(addrPortString);
  }
}
