/*
 * RMITransportNodeConnection.java
 *
 * Created on 8. duben 2004, 15:19
 */

package SOFA.SOFAnet.Transport.RMI;

import SOFA.SOFAnet.Repository.NodeInfo;
import SOFA.SOFAnet.Core.NetOps;
import SOFA.SOFAnet.Core.CoreException;
import SOFA.SOFAnet.Core.Reporter;
import SOFA.SOFAnet.Transport.*;
import java.rmi.RemoteException;
import java.util.*;

/**
 * The object of this class implements connection to other nodes (in transport layer).
 * The connection is kept alive certain time (one minute) after last operation on the connection.
 * It prevents the system from creating many connections if there are more transport requests.
 * <p>
 * The "push" and "pull" operations are performed by extra thread, therefore these "distribution" operations
 * creates just one RMI connection between two nodes.
 *
 * @author  Ladislav Sobr
 */
public class RMITransportNodeConnection
{
  private RMITransportClient parent;
  private NetOps netOps;
  private RMITransportServerHolder serverHolder;
  private NodeInfo nodeInfo;
  private Thread asyncThread;
  private Thread stopThread;
  private Object stopEvent;
  private boolean stopped;
  private long lastUseTime;
  private List asyncOps;
  private String myNodeName;
  
  final static private int OP_STOP = 0;
  final static private int OP_PUSH = 1;
  final static private int OP_PULL = 2;
  
  private static class AsyncOperation
  {
    public int op;
    public IOParams params;
    
    AsyncOperation(int op, IOParams params)
    {
      this.op = op;
      this.params = params;
    }
  }
  
  /** Creates a new instance of RMITransportNodeConnection */
  public RMITransportNodeConnection(NodeInfo nodeInfo, RMITransportClient parent)
  {
    this.parent = parent;
    
    serverHolder = new RMITransportServerHolder();
    serverHolder.setNode(nodeInfo);
    this.nodeInfo = nodeInfo;
    
    stopEvent = new Object();
    
    stopped = false;
    lastUseTime = System.currentTimeMillis();
    asyncOps = new LinkedList();
    
    myNodeName = NodeInfo.getLocalNodeName();

    
    asyncThread = new Thread()
                    {
                      public void run()
                      {
                        asyncThreadProcedure();
                      }
                    };
                        
    stopThread = new Thread()
                    {
                      public void run()
                      {
                        stopThreadProcedure();
                      }
                    };
    
    asyncThread.start();
    stopThread.start();

  }
  
  private void asyncThreadProcedure()
  {
    for (;;)
    {
      AsyncOperation operation;
      
      synchronized (asyncOps)
      {
        while (asyncOps.isEmpty())
        {
          try
          {
            asyncOps.wait();
          }
          catch (InterruptedException e)
          {
          }
        }

        touch();
        
        //go through operation list and choose operation with highest priority
        
        int index = -1;
        int priority = -1;      //2 == interactive ; 1 == distribution ; 0 == stop

        Iterator it = asyncOps.iterator();
        int i = 0;
        while (it.hasNext())
        {
          AsyncOperation asyncOp = (AsyncOperation)it.next();

          int p;
          if (asyncOp.op == OP_STOP) p = 0;
          else if (asyncOp.params.getPriority() == IOParams.PRIORITY_INTERACTIVE ||
                   asyncOp.params.getPriority() == IOParams.PRIORITY_BLOCKING) p = 2;
          else p = 1;

          if (index == -1)
          {
            priority = p;
            index = i;
          }
          else if (p > priority)
          {
            priority = p;
            index = i;
          }

          i++;
        }

        operation = (AsyncOperation)asyncOps.remove(index);
      }
      
      //perform operation
      
      switch (operation.op)
      {
      case OP_STOP:
        return;
        
      case OP_PUSH:
        {
          try
          {
            RMITransportInterface server = serverHolder.getServer();
            LicenceRMI licenceRMI = null;
            if (operation.params.getLicence() != null) licenceRMI = new LicenceRMI(operation.params.getLicence());
            byte[] bundle = operation.params.getBundleData().getData();
            if (bundle != null)
            {
              server.push(myNodeName, operation.params.getBundleName(), bundle, licenceRMI, operation.params.isOffer());
              bundle = null;
            }
            else
            {
              Reporter.error("Asynchronous push of '" + operation.params.getBundleName() + "' failed: Cannot load bundle from BundleData");
            }
          }
          catch (RemoteException e)
          {
            serverHolder.releaseServer();
            Reporter.error("Asynchronous push of '" + operation.params.getBundleName() + "' failed", e);
          }
          catch (RMITransportException e)
          {
            Reporter.error("Asynchronous push of '" + operation.params.getBundleName() + "' failed", e);
          }
          
          operation.params.getBundleData().delete();
        }
        break;
        
      
      case OP_PULL:
        {
          RMIPullOutputHolder outputHolder = null;
          try
          {
            RMITransportInterface server = serverHolder.getServer();
            outputHolder = server.pull(myNodeName, operation.params.getBundleName(), false, false, operation.params.getContractID());
          }
          catch (RemoteException e)
          {
            serverHolder.releaseServer();
            Reporter.error("Asynchronous pull of '" + operation.params.getBundleName() + "' failed", e);
          }
          catch (RMITransportException e)
          {
            Reporter.error("Asynchronous pull of '" + operation.params.getBundleName() + "' failed", e);
          }
          
          touch();

          if (outputHolder != null)
          {
            if (outputHolder.errCode == 0)
            {
              if (outputHolder.licenceRMI != null) operation.params.setLicence(outputHolder.licenceRMI.getLicence());
              else operation.params.setLicence(null);
              operation.params.setBundleData(new BundleData(outputHolder.bundle));
              operation.params.setOffer(false);
              operation.params.setSourceNodeName(nodeInfo.getName());
              
              try
              {
                parent.getNetOps().deliverBundlePull(operation.params);
              }
              catch (CoreException e)
              {
                Reporter.error("Asynchronous pull of '" + operation.params.getBundleName() + "' failed", e);
              }
              
              operation.params.getBundleData().delete();  //just to be sure
            }
            else
            {
              switch (outputHolder.errCode)
              {
              case 1: Reporter.error("Asynchronous pull of '" + operation.params.getBundleName() + "' failed: Cannot find binary bundle on remote node"); break;
              case 2: Reporter.error("Asynchronous pull of '" + operation.params.getBundleName() + "' failed: Cannot create BundleData for bundle"); break;
              case 3: Reporter.error("Asynchronous pull of '" + operation.params.getBundleName() + "' failed: Pull not allowed by remote node"); break;
              default: Reporter.error("Asynchronous pull of '" + operation.params.getBundleName() + "' failed: Transport layer failed"); break;
              }
            }
            
            outputHolder = null;
          }

        }
        break;
      }
      
      touch();
    }
  }
  
  private void stopThreadProcedure()
  {
    synchronized (stopEvent)
    {
      do
      {
        try
        {
          stopEvent.wait(1000 * 60); 
        }
        catch (InterruptedException e)
        {
        }
      }
      while (System.currentTimeMillis() - lastUseTime < 1000 * 60);
    }
    
    parent.removeFromMap(nodeInfo.getAddressAndPort(), this);
    
    synchronized (asyncOps)
    {
      asyncOps.add(new AsyncOperation(OP_STOP, null));
      asyncOps.notify();
      stopped = true;
    }
    
  }
  
  private void touch()
  {
    lastUseTime = System.currentTimeMillis();
    synchronized (stopEvent)
    {
      stopEvent.notify();
    }
  }
  
  /**
   * Test whether 'push' of bundle or bundle offer is accepted by addressee.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  offer
   * </ul>
   * <p>
   * Output:
   * <ul>
   * <li>  errCode
   * </ul>
   *
   * @param params Input/Output data for the call
   */
  public void testPush(IOParams params) throws TransportException
  {
    touch();
    
    boolean result = false;
    try
    {
      RMITransportInterface server = serverHolder.getServer();
      result = server.testPush(myNodeName, params.getBundleName(), params.isOffer());
    }
    catch (RemoteException e)
    {
      touch();
      serverHolder.releaseServer();
      throw new TransportException(e.getMessage(), e);
    }
    catch (RMITransportException e)
    {
      touch();
      throw new TransportException(e.getMessage(), e);
    }
    
    if (result) params.setErrCode(0);
    else params.setErrCode(1);

    touch();
  }
  
  /**
   * Push bundle or bundle offer to node.
   * <p>
   * Asynchronous (doesn't wait for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  bundleData
   * <li>  offer
   * <li>  priority
   * <li>  licence              &nbsp;&nbsp;<i>     (only if offer == false)</i>
   * </ul>
   *
   * Output: none
   *
   * @param params Input data for the call
   * @return false if push cannot be processed any more by this object
   */
  public boolean push(IOParams params)
  {
    touch();
    synchronized (asyncOps)
    {
      if (stopped) return false;
      
      asyncOps.add(new AsyncOperation(OP_PUSH, params));
      asyncOps.notify();
    }
    
    return true;
  }
      
  /**
   * Test whether 'pull' of bundle is accepted by source node.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  contractID
   * </ul>
   *
   * Output:
   * <ul>
   * <li>  errCode
   * </ul>
   *
   * @param params Input/Output data for the call
   */
  public void testPull(IOParams params) throws TransportException
  {
    touch();
    
    boolean result = false;
    try
    {
      RMITransportInterface server = serverHolder.getServer();
      result = server.testPull(myNodeName, params.getBundleName(), false, params.getContractID());
    }
    catch (RemoteException e)
    {
      touch();
      serverHolder.releaseServer();
      throw new TransportException(e.getMessage(), e);
    }
    catch (RMITransportException e)
    {
      touch();
      throw new TransportException(e.getMessage(), e);
    }

    if (result) params.setErrCode(0);
    else params.setErrCode(1);
    
    touch();
  }
  
  /**
   * Pulls bundle from node.
   * <p>
   * Asynchronous (doesn't wait for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  priority
   * <li>  contractID
   * </ul>
   * <p>
   * Output: none  (calls NetOps.deliverBundle)
   *   
   *
   * @param params Input data for the call
   * @return false if pull cannot be processed any more by this object
   */
  public boolean pull(IOParams params)
  {
    touch();
    synchronized (asyncOps)
    {
      if (stopped) return false;
      
      asyncOps.add(new AsyncOperation(OP_PULL, params));
      asyncOps.notify();
    }
    
    return true;
  }

  /**
   * Pulls bundle from node.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  priority
   * <li>  contractID
   * </ul>
   *
   * Output:
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == Cannot find binary bundle
   *    <li>    2 == Cannot create BundleData for bundle
   *    <li>    3 == Pull not allowed
   *    </ul>
   * <li>  bundleData   &nbsp;&nbsp;<i>  (only if errCode == 0)</i>
   * <li>  licence      &nbsp;&nbsp;<i>  (only if errCode == 0)</i>
   * </ul>
   *   
   *
   * @param params Input/Output data for the call
   */
  public void synchroPull(IOParams params) throws TransportException
  {
    touch();
    
    RMIPullOutputHolder outputHolder = null;
    try
    {
      RMITransportInterface server = serverHolder.getServer();
      outputHolder = server.pull(myNodeName, params.getBundleName(), false, false, params.getContractID());
    }
    catch (RemoteException e)
    {
      touch();
      serverHolder.releaseServer();
      throw new TransportException(e.getMessage(), e);
    }
    catch (RMITransportException e)
    {
      touch();
      throw new TransportException(e.getMessage(), e);
    }

    params.setErrCode(outputHolder.errCode);
    if (outputHolder.licenceRMI != null) params.setLicence(outputHolder.licenceRMI.getLicence());
    else params.setLicence(null);
    params.setBundleData(new BundleData(outputHolder.bundle));
    
    touch();
  }
  
  /**
   * Test whether source node's share manager can accept acquire.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * </ul>
   * <p>
   * Output:
   * <ul>
   * <li>  errCode
   * </ul>
   *
   *
   * @param params Input/Output data for the call
   */
  public void testAcquireShared(IOParams params) throws TransportException
  {
    touch();
    
    boolean result = false;
    try
    {
      RMITransportInterface server = serverHolder.getServer();
      result = server.testPull(myNodeName, params.getBundleName(), true, "");
    }
    catch (RemoteException e)
    {
      touch();
      serverHolder.releaseServer();
      throw new TransportException(e.getMessage(), e);
    }
    catch (RMITransportException e)
    {
      touch();
      throw new TransportException(e.getMessage(), e);
    }

    if (result) params.setErrCode(0);
    else params.setErrCode(1);
    
    touch();
  }
  
  /**
   * Acquire bundle from node's share manager.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  licenceOnly
   * <li>  priority
   * </ul>
   *
   * Output:
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == Not a share manager or access not allowed
   *    <li>    2 == Licence not available
   *    <li>    3 == Data error (binary bundle not available)
   *    <li>    4 == Incosistent state of share manager
   *    </ul>
   * <li>  address          &nbsp;&nbsp;<i> (only if errCode == 0 && licenceOnly == false)</i>
   * <li>  bundleData       &nbsp;&nbsp;<i> (only if errCode == 0 && licenceOnly == false && address == false)</i>
   * <li>  addressNodeName  &nbsp;&nbsp;<i> (only if errCode == 0 && licenceOnly == false && address == true)</i>
   * <li>  licence          &nbsp;&nbsp;<i> (only if errCode == 0)</i>
   * </ul>
   *   
   *
   * @param params Input/Output data for the call
   */
  public void acquireShared(IOParams params) throws TransportException
  {
    touch();
    
    RMIPullOutputHolder outputHolder = null;
    try
    {
      RMITransportInterface server = serverHolder.getServer();
      outputHolder = server.pull(myNodeName, params.getBundleName(), true, params.isLicenceOnly(), "");
    }
    catch (RemoteException e)
    {
      touch();
      serverHolder.releaseServer();
      throw new TransportException(e.getMessage(), e);
    }
    catch (RMITransportException e)
    {
      touch();
      throw new TransportException(e.getMessage(), e);
    }

    params.setErrCode(outputHolder.errCode);
    if (outputHolder.licenceRMI != null) params.setLicence(outputHolder.licenceRMI.getLicence());
    else params.setLicence(null);
    if (params.isLicenceOnly())
    {
      //just to be sure, but nobody should read it
      params.setAddress(false);
      params.setAddressNodeName("");
      params.setBundleData(null);
    }
    else
    {
      if (outputHolder.address == null || outputHolder.address.length() == 0)
      {
        params.setAddress(false);
        params.setAddressNodeName("");
        params.setBundleData(new BundleData(outputHolder.bundle));
      }
      else
      {
        params.setAddress(true);
        params.setAddressNodeName(outputHolder.address);
        params.setBundleData(null);
      }
    }
    
    touch();
  }
  
  /**
   * Tests whether node (share client) can return shared bundle (it's licence) back to share manager.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * </ul>
   *
   * Output:
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == Cannot return bundle(licence)
   *    </ul>
   * </ul>
   *
   * @param params Input/Output data for the call
   */
  public void canReturnShared(IOParams params) throws TransportException
  {
    touch();
    
    boolean result = false;
    try
    {
      RMITransportInterface server = serverHolder.getServer();
      result = server.canReturnShared(myNodeName, params.getBundleName());
    }
    catch (RemoteException e)
    {
      touch();
      serverHolder.releaseServer();
      throw new TransportException(e.getMessage(), e);
    }
    catch (RMITransportException e)
    {
      touch();
      throw new TransportException(e.getMessage(), e);
    }

    if (result) params.setErrCode(0);
    else params.setErrCode(1);
    touch();
  }
  
  /**
   * Return shared bundle (it's licence) back to share manager (and prepare binary bundle for acquire).
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  address
   * <li>  addressNodeName    &nbsp;&nbsp;<i>     (if address == true)</i>
   * </ul>
   *
   * Output:
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == Not a share client
   *    <li>    2 == Cannot return bundle(licence)
   *    <li>    3 == Cannot prepare for acquire
   *    </ul>
   * </ul>
   *
   * @param params Input/Output data for the call
   */
  public void returnShared(IOParams params) throws TransportException
  {
    touch();
    
    int errCode = 0;
    try
    {
      RMITransportInterface server = serverHolder.getServer();
      errCode = server.returnShared(myNodeName, params.getBundleName(), params.isAddress(), params.getAddressNodeName());
    }
    catch (RemoteException e)
    {
      touch();
      serverHolder.releaseServer();
      throw new TransportException(e.getMessage(), e);
    }
    catch (RMITransportException e)
    {
      touch();
      throw new TransportException(e.getMessage(), e);
    }

    params.setErrCode(errCode);
    touch();
  }

  /**
   * Ask share manager to call returnShared on us.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * </ul>
   *
   * Output:
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == Not a share manager or access not allowed
   *    <li>    2 == returnShared failed
   *    </ul>
   * </ul>
   *
   * @param params Input/Output data for the call
   */
  public void manualReturnShared(IOParams params) throws TransportException
  {
    touch();
    
    int errCode = 0;
    try
    {
      RMITransportInterface server = serverHolder.getServer();
      errCode = server.manualReturnShared(myNodeName, params.getBundleName());
    }
    catch (RemoteException e)
    {
      touch();
      serverHolder.releaseServer();
      throw new TransportException(e.getMessage(), e);
    }
    catch (RMITransportException e)
    {
      touch();
      throw new TransportException(e.getMessage(), e);
    }

    params.setErrCode(errCode);
    touch();
  }
  
}
