/*
 * RMITransportServer.java
 *
 * Created on 7. duben 2004, 18:13
 */

package SOFA.SOFAnet.Transport.RMI;

import java.rmi.server.UnicastRemoteObject;
import java.rmi.server.ServerNotActiveException;
import java.rmi.RemoteException;
import SOFA.SOFAnet.Core.NetOps;
import SOFA.SOFAnet.Core.ShareOps;
import SOFA.SOFAnet.Core.CoreException;
import SOFA.SOFAnet.Repository.NodeInfo;
import SOFA.SOFAnet.Transport.*;

/**
 * RMI implementation of server part of transport layer.
 * <p>
 * The server just transforms calls on RMITransportInterface to the transpor layer interface (provided by NetOps and ShareOps).
 *
 * @author  Ladislav Sobr
 */
public class RMITransportServer extends UnicastRemoteObject implements RMITransportInterface
{
  private NetOps netOps;
  private ShareOps shareOps;

  /** Creates a new instance of RMITransportServer */
  public RMITransportServer(NetOps netOps, ShareOps shareOps) throws RemoteException
  {
    this.netOps = netOps;
    this.shareOps = shareOps;
  }
  
  public boolean testPush(String nodeName, String bundleName, boolean offer) throws RMITransportException, RemoteException
  {
    testNodeName(nodeName);
    
    boolean result;
    try
    {
      IOParams ioParams = new IOParams();
      
      ioParams.setBundleName(bundleName);
      ioParams.setOffer(offer);
      ioParams.setSourceNodeName(nodeName);

      result = netOps.isPushAcceptable(ioParams);
    }
    catch (CoreException e)
    {
      throw new RMITransportException(e.getMessage(), e);
    }

    return result;
  }

  public void push(String nodeName, String bundleName, byte[] bundle, LicenceRMI licenceRMI, boolean offer) throws RMITransportException, RemoteException
  {
    testNodeName(nodeName);
    
    try
    {
      IOParams ioParams = new IOParams();
      
      ioParams.setBundleName(bundleName);
      ioParams.setBundleData(new BundleData(bundle));
      ioParams.setOffer(offer);
      ioParams.setSourceNodeName(nodeName);
      if (licenceRMI == null) ioParams.setLicence(null);
      else ioParams.setLicence(licenceRMI.getLicence());
      
      netOps.deliverBundlePush(ioParams);
    }
    catch (CoreException e)
    {
      throw new RMITransportException(e.getMessage(), e);
    }
  }
  
  public boolean testPull(String nodeName, String bundleName, boolean sharing, String contractID) throws RMITransportException, RemoteException
  {
    testNodeName(nodeName);
    
    boolean result;
    try
    {
      IOParams ioParams = new IOParams();
      
      ioParams.setBundleName(bundleName);
      ioParams.setSourceNodeName(nodeName);

      if (sharing) result = shareOps.isAcquisitionAcceptable(ioParams);
      else
      {
        ioParams.setContractID(contractID);
        result = netOps.isPullAcceptable(ioParams);
      }
    }
    catch (CoreException e)
    {
      throw new RMITransportException(e.getMessage(), e);
    }

    return result;
  }
  
  public RMIPullOutputHolder pull(String nodeName, String bundleName, boolean sharing, boolean licenceOnly, String contractID) throws RMITransportException, RemoteException
  {
    testNodeName(nodeName);

    try
    {
      IOParams ioParams = new IOParams();
      
      ioParams.setBundleName(bundleName);
      ioParams.setSourceNodeName(nodeName);

      if (sharing)
      {
        ioParams.setLicenceOnly(licenceOnly);
        shareOps.prepareBundleForAcquisition(ioParams);
      }
      else
      {
        ioParams.setContractID(contractID);
        netOps.preparePulledBundle(ioParams);
      }
      
      RMIPullOutputHolder outputHolder = new RMIPullOutputHolder();
      outputHolder.errCode = ioParams.getErrCode();
      if (outputHolder.errCode == 0)
      {
        if (ioParams.isAddress() && ioParams.getAddressNodeName() != null && ioParams.getAddressNodeName().length() != 0)
        {
          outputHolder.address = ioParams.getAddressNodeName();
        }
        else
        {
          BundleData bundleData = ioParams.getBundleData();
          if (bundleData != null)
          {
            outputHolder.bundle = bundleData.getData();
            bundleData.delete();
            if (outputHolder.bundle == null) throw new RMITransportException("I/O error when copying bundle data");
          }
          else outputHolder.bundle = null;
        }
        
        if (ioParams.getLicence() == null) outputHolder.licenceRMI = null;
        else outputHolder.licenceRMI = new LicenceRMI(ioParams.getLicence());
      }
      
      return outputHolder;
    }
    catch (CoreException e)
    {
      throw new RMITransportException(e.getMessage(), e);
    }
  }

  public boolean canReturnShared(String nodeName, String bundleName) throws RMITransportException, RemoteException
  {
    testNodeName(nodeName);
    
    boolean result;
    try
    {
      IOParams ioParams = new IOParams();
      
      ioParams.setBundleName(bundleName);
      ioParams.setSourceNodeName(nodeName);

      result = shareOps.canReturnSharedBundle(ioParams);
    }
    catch (CoreException e)
    {
      throw new RMITransportException(e.getMessage(), e);
    }

    return result;
  }
  
  public int returnShared(String nodeName, String bundleName, boolean address, String addressNodeName) throws RMITransportException, RemoteException
  {
    testNodeName(nodeName);
    
    int result;
    try
    {
      IOParams ioParams = new IOParams();
      
      ioParams.setBundleName(bundleName);
      ioParams.setSourceNodeName(nodeName);
      ioParams.setAddress(address);
      ioParams.setAddressNodeName(addressNodeName);
      
      result = shareOps.returnSharedBundle(ioParams);
    }
    catch (CoreException e)
    {
      throw new RMITransportException(e.getMessage(), e);
    }

    return result;
  }

  public int manualReturnShared(String nodeName, String bundleName) throws RMITransportException, RemoteException
  {
    testNodeName(nodeName);
    
    int result;
    try
    {
      IOParams ioParams = new IOParams();
      
      ioParams.setBundleName(bundleName);
      ioParams.setSourceNodeName(nodeName);

      result = shareOps.performManualReturnSharedBundle(ioParams);
    }
    catch (CoreException e)
    {
      throw new RMITransportException(e.getMessage(), e);
    }
    
    return result;
  }
  
  /** Tests wheher specified name of node corresponds to the real host of RMI client */
  private NodeInfo testNodeName(String nodeName) throws RMITransportException
  {
    NodeInfo nodeInfo = new NodeInfo();
    try
    {
      nodeInfo.setNodeName(nodeName);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      throw new RMITransportException("Invalid name of SOFA node: " + nodeName, e);
    }

    String realHost = "";
    
    try
    {
      realHost = getClientHost();
    }
    catch (ServerNotActiveException e)
    {
      throw new RMITransportException(e.getMessage(), e);
    }      
    
    boolean ok = false;
    try
    {
      ok = nodeInfo.isSameHost(realHost);
    }
    catch (NodeInfo.InvalidNodeNameException e)
    {
      throw new RMITransportException("Invalid RMI client host: " + nodeName, e);
    }
    
    if (!ok) throw new RMITransportException("Name of SOFA node (" + nodeName + ") and RMI client host (" + realHost + ") are incosistent");
    
    return nodeInfo;
  }
}
