/*
 * TransportInterface.java
 *
 * Created on 31. bezen 2004, 17:20
 */

package SOFA.SOFAnet.Transport;

import SOFA.SOFAnet.Core.NetOps;
import SOFA.SOFAnet.Core.ShareOps;

/**
 * Interface of transport layer of the SOFAnet.
 *
 * @author  Ladislav Sobr
 */
public interface TransportInterface
{
  /**
   * Test whether 'push' of bundle or bundle offer is accepted by addressees.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  offer
   * <li>  nodeName
   * <li>  *OtherNodes.nodeName
   * </ul>
   * <p>
   * Output:
   * <ul>
   * <li>  nodeName
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == not allowed
   *    </ul>
   * <li>  *OtherNodes.nodeName
   * <li>  *OtherNodes.errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == not allowed
   *    </ul>
   * </ul>
   *
   * @param params Input/Output data for the call
   */
  void testPush(IOParams params) throws TransportException;

  /**
   * Push bundle or bundle offer to nodes.
   * <p>
   * Asynchronous (doesn't wait for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  bundleData
   * <li>  offer
   * <li>  priority
   * <li>  nodeName
   * <li>  licence                &nbsp;&nbsp;<i>   (only if offer == false)</i>
   * <li>  *OtherNodes.nodeName
   * <li>  *OtherNodes.licence    &nbsp;&nbsp;<i>   (only if offer == false)</i>
   * </ul>
   * <p>
   * Output: none
   *
   * @param params Input data for the call
   */
  void push(IOParams params) throws TransportException;
  
  /**
   * Test whether 'pull' of bundle is accepted by source node.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  nodeName
   * <li>  contractID
   * </ul>
   * <p>
   * Output:
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == not allowed
   *    </ul>
   * </ul>
   *
   * @param params Input/Output data for the call
   */
  void testPull(IOParams params) throws TransportException;
  
  /**
   * Pulls bundle from node.
   * <p>
   * Asynchronous (doesn't wait for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  priority
   * <li>  nodeName
   * <li>  contractID
   * </ul>
   * <p>
   * Output: none  (calls NetOps.deliverBundlePull)
   *   
   *
   * @param params Input data for the call
   */
  void pull(IOParams params) throws TransportException;

  /**
   * Pulls bundle from node.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  priority
   * <li>  nodeName
   * <li>  contractID
   * </ul>
   * <p>
   * Output:
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == Cannot find binary bundle
   *    <li>    2 == Cannot create BundleData for bundle
   *    <li>    3 == Pull not allowed
   *    </ul>
   * <li>  bundleData   &nbsp;&nbsp;<i>  (only if errCode == 0)</i>
   * <li>  licence      &nbsp;&nbsp;<i>  (only if errCode == 0)</i>
   * </ul>
   *   
   *
   * @param params Input/Output data for the call
   */
  void synchroPull(IOParams params) throws TransportException;

  /**
   * Test whether source node's share manager can accept acquire.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  nodeName
   * </ul>
   * <p>
   * Output:
   * <ul>
   * <li>  errCode
   * </ul>
   *
   *
   * @param params Input/Output data for the call
   */
  void testAcquireShared(IOParams params) throws TransportException;
  
  /**
   * Acquire bundle from node's share manager (or share client).
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  priority
   * <li>  nodeName
   * </ul>
   *
   * Output:
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == Not a share manager or access not allowed
   *    <li>    2 == Licence not available
   *    <li>    3 == Data error (binary bundle not available)
   *    <li>    4 == Incosistent state of share manager
   *    </ul>
   * <li>  address         &nbsp;&nbsp;<i> (only if errCode == 0 ; can be true only when calling share manager)</i>
   * <li>  bundleData      &nbsp;&nbsp;<i> (only if errCode == 0 && address == false)</i>
   * <li>  addressNodeName &nbsp;&nbsp;<i> (only if errCode == 0 && address == true)</i>
   * <li>  licence         &nbsp;&nbsp;<i> (only if errCode == 0 and when calling share manager)</i>
   * </ul>
   *   
   *
   * @param params Input/Output data for the call
   */
  void acquireShared(IOParams params) throws TransportException;
  
  /**
   * Tests whether node (share client) can return shared bundle (it's licence) back to share manager.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  nodeName
   * </ul>
   *
   * Output:
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == Cannot return bundle(licence)
   *    </ul>
   * </ul>
   *
   * @param params Input/Output data for the call
   */
  void canReturnShared(IOParams params) throws TransportException;
  
  /**
   * Return shared bundle (it's licence) back to share manager (and prepare binary bundle for acquire).
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  nodeName
   * <li>  address
   * <li>  addressNodeName      &nbsp;&nbsp;<i>   (only if address == true)</i>
   * </ul>
   * <p>
   * Output:
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == Not a share client
   *    <li>    2 == Cannot return bundle(licence)
   *    <li>    3 == Cannot prepare for acquire (but licence is returned)
   *    </ul>
   * </ul>
   *
   * @param params Input/Output data for the call
   */
  void returnShared(IOParams params) throws TransportException;
  
  /**
   * Ask share manager to call returnShared on us.
   * <p>
   * Synchronous (waits for completion).
   * <p>
   * Input:
   * <ul>
   * <li>  bundleName
   * <li>  nodeName
   * </ul>
   * <p>
   * Output:
   * <ul>
   * <li>  errCode
   *    <ul>
   *    <li>    0 == OK
   *    <li>    1 == Not a share manager or access not allowed
   *    <li>    2 == returnShared failed
   *    </ul>
   * </ul>
   *
   * @param params Input/Output data for the call
   */
  void manualReturnShared(IOParams params) throws TransportException;

}
