/**
 * SOFANode.java is a part of the SOFA project.
 * This file was created by pepan on 18.4.2003.
 */
package SOFA.SOFAnode.InOut.Impl;

import javax.swing.JTextArea;
import javax.swing.tree.DefaultMutableTreeNode;

import SOFA.Connector.ConnectorException;
import SOFA.Connector.ConnectorTransportException;
import SOFA.Connector.Reference;
import SOFA.SOFAnode.InOut.Bundle;
import SOFA.SOFAnode.InOut.InOut2Client;
import SOFA.SOFAnode.InOut.Connector.InOut2ClientConnector;
import SOFA.SOFAnode.TR.ComponentInfo;

/**
 * Contains necessary information about one node in the SOFA network. Each SOFA node
 * creates a set of such objects for every other SOFA node (known to the local node) in the network.
 * @author Petr Panuska
 */
class SOFANode {
  /**
   * Name of the SOFA node as it appears in a client.
   */
  private String name = null;

  /**
   * The SOFA reference to the SOFA node.
   */
  private Reference reference = null;

  /**
   * The SOFA connector to the SOFA node. If <code>null</code>, there is no connection made
   * from the local SOFA node to this SOFA node.
   */
  private InOut2Client inOut = null;

  /**
   * An array of components stored on this SOFA node.
   */
  private ComponentInfo[] components = null;

  /**
   * A text area where messages about connecting, etc. are written to.
   */
  private JTextArea messages = null;

  /**
   * Creates an instance of this SOFA node.
   * @param name The name of this SOFA node.
   * @param ref The string representation of the {@link #reference} to this SOFA node.
   * @param messages The text area where messages are written to.
   */
  SOFANode (String name, String ref, JTextArea messages) {
    this.name = name;
    reference = Reference.fromString(ref);
    this.messages = messages;
  }

  /**
   * Tries to connect the local SOFA node to this SOFA node. The {@link #inOut} field is
   * set according to the result. If the connecting fails, the field is set to <code>null</code>.
   * It also writes a message about the result to the {@link #messages} field.
   */
  private void _connect () {
    try {
      inOut = (InOut2Client) InOut2ClientConnector.createClt(reference);
      messages.append("A connection to '" + name + "' has been made successfully.\n");
    } catch (ConnectorException e) {
      messages.append("When connecting to '" + name + "', the following exception was thrown:\n" + e + "\n");
      inOut = null;
      components = null;
    }
  }

  /**
   * Connects the local SOFA node to this one. It also adds an instance of  a
   * <code>TreeNode</code> to the tree.
   * @param root The parent node the new tree node is added to.
   * @return The newly created tree node.
   */
  DefaultMutableTreeNode connect (DefaultMutableTreeNode root) {
    _connect();
    DefaultMutableTreeNode tNode = new DefaultMutableTreeNode(this, inOut != null); // allows children if connected
    root.add(tNode);
    return tNode;
  }

  /**
   * Reconnects the local SOFA node to this one. It also changes an instance of  a
   * <code>TreeNode</code> according to the result of connecting.
   * @param tNode The tree node which is being changed.
   * @return The changed tree node.
   */
  DefaultMutableTreeNode reconnect (DefaultMutableTreeNode tNode) {
    _connect();
    tNode = new DefaultMutableTreeNode(this, inOut != null); // allows children if connected
//    tNode = new DefaultMutableTreeNode(this, true); // allows always children (even if not connected now)
    return tNode;
  }

  /**
   * Lists components stored on this SOFA node. It also removes all previous child nodes
   * (= installed components) and adds the new ones to the specified parent tree node.
   * <br>If the local SOFA node is not connected to this one, it just removes the child tree nodes
   * (but it doesn't list anything).
   * @param tNode The parent tree node where the listed components are stored to.
   */
  void list (DefaultMutableTreeNode tNode) {
    tNode.removeAllChildren();
    if (isConnected()) {
      Bundle bundle;
      try {
        bundle = inOut.list();
      } catch (ConnectorTransportException e) {
        messages.append("When listing components on the SOFA node '" + name + "', the following exception was thrown:\n" + e + "\n");
        inOut = null;
        return;
      }
      components = bundle.getComponents();
      if (components.length > 0) {
        tNode.setAllowsChildren(true);
        for (int i = 0; i < components.length; i++) {
          ComponentInfo info = components[i];
          tNode.add(new DefaultMutableTreeNode(info));
        }
      }
    }
  }

  /**
   * Returns <code>true</code> when the local SOFA node is connected to this one.
   * @return <code>false</code> when the local node is not connected to this one.
   */
  boolean isConnected () {
    return inOut != null;
  }

  /**
   * Returns the name of this SOFA node. In case that the local node is not connected to
   * this one, a string ' - not accessible' is added to the name.
   * @return the name of this SOFA node as it was stated in the configuration file.
   */
  public String toString () {
    return name + ((isConnected()) ? "" : " - not accessible");
  }

  /**
   * Gets the name of this SOFA node.
   * @return the SOFA node name.
   */
  String getName () {
    return name;
  }

  /**
   * Gets the reference to this SOFA node.
   * @return the SOFA node reference.
   */
  Reference getReference () {
    return reference;
  }

  /**
   * Gets the connector from the local SOFA node to this one.
   * @return the SOFA connector.
   */
  InOut2Client getInOut () {
    return inOut;
  }

  /**
   * Gets an array of components stored on this SOFA node.
   * @return an array of stored components.
   */
  ComponentInfo[] getComponents () {
    return components;
  }

}
