/**
 * ComponentInfoImpl.java is a part of the SOFA project.
 * This file was created by pepan on 19.3.2003.
 */
package SOFA.SOFAnode.TR.Impl;

import java.io.Serializable;

import SOFA.SOFAnode.TR.ComponentInfo;

/**
 * Describes one component. It contains all attributes a component can have.
 * However, the current implementation does not use them all.
 * @author Petr Panuska
 */
public class ComponentInfoImpl implements ComponentInfo, Serializable {

  /**
   * The component name. Usually, it consists of a few strings delimited by '::'.
   * The first ones define the vendor, the other specify the name hierarchy.
   */
  private String name;

  /**
   * The specification version of the component interface. It is not used in the current
   * implementation.
   */
  private String specificationVersion;

  /**
   * The implementation version of the component interface provided by a single vendor.
   */
  private String implementationVersion;

  /**
   * The location where this component can be obtained from. It could be either
   * an absolute URL beginning with 'sofa://' prefix and pointing to a SOFA node, or
   * a relative path to a directory within a bundle. The latter case is used when
   * the component is being sent packed in a bundle only.
   */
  private String location;

  /**
   * List of sub-components, this component consists of. It is not used in the current
   * implementation.
   */
  private ComponentInfo[] subComponents;

  /**
   * Creates a component info where just 2 fields are known.
   * @param name the component name.
   * @param implementationVersion the component implementation version.
   */
  public ComponentInfoImpl (String name, String implementationVersion) {
    this.name = name;
    this.implementationVersion = implementationVersion;
  }

  /**
   * Creates a component info where 3 fields are known.
   * @param name the component name.
   * @param implementationVersion the component implementation version.
   * @param location the location where this component can be obtained from.
   */
  public ComponentInfoImpl (String name, String implementationVersion, String location) {
    this(name, implementationVersion);
    this.location = location;
  }

  /**
   * Gets the component name.
   * @return the component name.
   */
  public String getName () {
    return name;
  }

  /**
   * Gets the component specification version.
   * @return the specification version.
   */
  public String getSpecificationVersion () {
    return specificationVersion;
  }

  /**
   * Gets the component implementation version.
   * @return the implementation version.
   */
  public String getImplementationVersion () {
    return implementationVersion;
  }

  /**
   * Gets the location where this componenta can be obtained from.
   * @return the location of this component.
   */
  public String getLocation () {
    return location;
  }

  /**
   * Gets components this component consists of.
   * @return an array of components.
   */
  public ComponentInfo[] getSubComponents () {
    return subComponents;
  }

  /**
   * Sets the location where this component can be obtained from.
   * @param location an absolute URL beginning with the 'sofa://' prefix, a relative
   * path to a directory within a bundle otherwise.
   */
  void setLocation (String location) {
    this.location = location;
  }

  void setSubComponents (ComponentInfoImpl[] subComponents) {
    this.subComponents = subComponents;
  }

  /**
   * Compares this component info with the specified one. Two instances of this class
   * equal if their names and implementation versions are the same.
   * @param o the second component info.
   * @return <code>true</code> if both instances equal.
   */
  public boolean equals (Object o) {
    if (this == o) return true;
    if (!(o instanceof ComponentInfoImpl)) return false;

    final ComponentInfoImpl componentInfo = (ComponentInfoImpl) o;

    if (!implementationVersion.equals(componentInfo.implementationVersion)) return false;
    if (!name.equals(componentInfo.name)) return false;

    return true;
  }

  public int hashCode () {
    int result;
    result = name.hashCode();
    result = 29 * result + implementationVersion.hashCode();
    return result;
  }

  /**
   * Returns a string of following format XXX[YYY]. XXX denotes the component name;
   * YYY denotes the component implementation version.
   * @return the describing string.
   */
  public String toString () {
    return name + "[" + implementationVersion + "]";
  }

  /**
   * Creates the component info from a string returned by the {@link #toString} method.
   * Only the name and the implementation version are filled in the created instance.
   * @param fullName a string consisting of XXX[YYY], where XXX denotes the name and
   * YYY denotes the implementation version.
   * @return an instance of this class.
   */
  public static ComponentInfoImpl fromString (String fullName) {
    int delimiter = fullName.lastIndexOf('[');
    String name = fullName.substring(0, delimiter);
    String version = fullName.substring(delimiter + 1, fullName.length() - 1);
    return new ComponentInfoImpl(name, version);
  }
}
