/* 
 * $Id: AlternativeNode.java,v 1.4 2005/07/08 12:04:12 kofron Exp $
 *
 * Copyright 2004
 *    Distributed Systems Research Group
 *    Department of Software Engineering
 *    Faculty of Mathematics and Physics
 *    Charles University, Prague
 *
 * Copyright 2005
 *    Formal Methods In Software Engineering Group
 *    Institute of Computer Science
 *    Academy of Sciences of the Czech Republic
 * 
 * This code was developed by Jan Kofron <kofron@nenya.ms.mff.cuni.cz>
 */


package SOFA.SOFAnode.Util.DFSRChecker.node;

import java.util.ArrayList;


import SOFA.SOFAnode.Util.DFSRChecker.DFSR.CheckingException;
import SOFA.SOFAnode.Util.DFSRChecker.state.DenotedState;
import SOFA.SOFAnode.Util.DFSRChecker.state.SimpleState;
import SOFA.SOFAnode.Util.DFSRChecker.state.State;
import SOFA.SOFAnode.Util.DFSRChecker.state.TransitionPair;
import SOFA.SOFAnode.Util.DFSRChecker.state.TransitionPairs;
import SOFA.SOFAnode.Util.DFSRChecker.utils.AnotatedProtocol;


/**
 * Alternative node represent the alternative operator.
 */
public class AlternativeNode extends TreeNode {

    /** Creates a new instance of AlternativeNode */
    public AlternativeNode(TreeNode[] nodes) {
        super(getProtocol(nodes));
        this.nodes = nodes;

        isorparallel = false;
        initState = null;
    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getInitial()
     */
    public State getInitial() {
        if (initState == null) {
            /*
             * State[] states = new State[nodes.length]; for (int i = 0; i <
             * states.length; ++i) states[i] = nodes[i].getInitial();
             */
            initState = new DenotedState(-1, new SimpleState(0), logchildrencnt);
        }

        return initState;
    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#isAccepting(SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.state.State)
     */
    public boolean isAccepting(State state) {
        int index = ((DenotedState) state).index;

        if (index == -1) {
            for (int i = 0; i < nodes.length; ++i)
                if (nodes[i].isAccepting(nodes[i].getInitial()))
                    return true;

            return false;
        } else {
            return nodes[index].isAccepting(((DenotedState) state).state);
        }

    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getTransitions(SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.state.State)
     */
    public TransitionPairs getTransitions(State state) throws InvalidParameterException, CheckingException {
        DenotedState dstate = (DenotedState) state;
        int stindex = dstate.index;
        int nodecnt = nodes.length;
        int totallen = 0;

        if (stindex == -1) {
            TransitionPair[][] trans = new TransitionPair[nodecnt][];

            for (int i = 0; i < nodecnt; ++i) {
                trans[i] = nodes[i].getTransitions(nodes[i].getInitial()).transitions;
                totallen += trans[i].length;
            }

            TransitionPair[] result = new TransitionPair[totallen];

            for (int i = 0, idx = 0; idx < nodecnt; ++idx) {
                for (int j = 0; j < trans[idx].length; ++j, ++i) {

                    State newstate = trans[idx][j].state;
                    result[i] = new TransitionPair(trans[idx][j].eventIndex, new DenotedState(idx, newstate, logchildrencnt));
                }
            }

            return new TransitionPairs(result);

        } else {

            TransitionPair[] nodetrans = nodes[stindex].getTransitions(dstate.state).transitions;
            int nodelen = nodetrans.length;
            TransitionPair[] trans = new TransitionPair[nodetrans.length];

            for (int i = 0; i < nodetrans.length; ++i)
                trans[i] = new TransitionPair(nodetrans[i].eventIndex, new DenotedState(stindex, nodetrans[i].state, logchildrencnt));

            return new TransitionPairs(trans);
        }
    }

    /**
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getWeight()
     */
    public long getWeight() {
        if (weight != -1)
            return weight;
        else {
            long weight = 0;
            for (int i = 0; i < nodes.length; ++i)
                weight += nodes[i].getWeight();

            return weight;
        }
    }

    /**
     * @return the symbolic name of the treenode denoting its type
     */
    public String[] getTypeName() {
        String[] result = { "Alternative", "+" };
        return result;
    }

    /**
     * Overrides the default implementation, because of the non-determinism that
     * must be handled here. This is one of the only two places where
     * non-determinism can inherently appear.
     * 
     * @return number of bits needed to keep the determinized signature of this
     *         subtree
     */
    public int getLeafCount() {
        int cnt = 0;
        for (int i = 0; i < nodes.length; ++i)
            cnt += nodes[i].getLeafCount();

        int max = 0;
        for (int i = 0; i < 32; ++i) {
            if ((nodes.length >>> i) == 0) {
                max = i;
                break;
            }
        }

        this.logchildrencnt = max;

        // adding the space necessary for identifying a child
        // and assume that all of the children transition may start
        // with the same transition
        cnt += (nodes.length * max);

        return cnt;
    }

    /**
     * Sets the orparallel operator bit needed for protocol inverting.
     */
    public void setorparallel() {
        isorparallel = true;
    }

    /**
     * @return the orparallel bit
     */
    public boolean getorparallel() {
        return isorparallel;
    }

    /**
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getAnotatedProtocol(SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.state.State)
     */
    public AnotatedProtocol getAnotatedProtocol(State state) {
        DenotedState dstate = (DenotedState) state;
        int nodecnt = nodes.length;
        String result = new String();
        ArrayList indicesresult = new ArrayList();
        
        result = "(";
        for (int i = 0; i < nodecnt; ++i) {
            AnotatedProtocol subresult = nodes[i].getAnotatedProtocol(((dstate != null) && (dstate.index == i)) ? dstate.state : null);
            
            for (int j = 0; j < subresult.indices.size(); ++j)
                indicesresult.add(new Integer(((Integer)subresult.indices.get(j)).intValue() + result.length()));
            
            result += subresult.protocol;
            
            if (i < nodecnt - 1)
                result += ")+(";
        }
        result += ")";
        
        if ((dstate != null) && (dstate.index == -1))
            indicesresult.add(new Integer(0));
        
        return new AnotatedProtocol(result, indicesresult);
    }


    
    /**
     * This method is obsolete and used only for debuging, the protocol is
     * stored during the parse process.
     * 
     * @deprecated
     * @return the protocol of this tree.
     */
    static private String getProtocol(TreeNode[] nodes) {
        String result = new String();
        result = "(" + nodes[0].protocol + ")";

        for (int i = 1; i < nodes.length; ++i) {
            result += " + (" + nodes[i].protocol + ")";
        }

        return result;

    }

    //--------------------------------------------------------------------------
    /**
     * Initial state of the automaton
     */
    private DenotedState initState;

    /**
     * logarithm of child count
     */
    private int logchildrencnt;

    /**
     * true if this is an orparallel node
     */
    private boolean isorparallel;

}