/* 
 * $Id: RestrictionNode.java,v 1.4 2005/07/08 12:04:12 kofron Exp $ 
 *
 * Copyright 2004
 *    Distributed Systems Research Group
 *    Department of Software Engineering
 *    Faculty of Mathematics and Physics
 *    Charles University, Prague
 *
 * Copyright 2005
 *    Formal Methods In Software Engineering Group
 *    Institute of Computer Science
 *    Academy of Sciences of the Czech Republic
 * 
 * This code was developed by Jan Kofron <kofron@nenya.ms.mff.cuni.cz>
 */

package SOFA.SOFAnode.Util.DFSRChecker.node;

import java.util.TreeSet;
import java.util.Stack;
import java.util.ArrayList;
import java.util.Iterator;

import SOFA.SOFAnode.Util.DFSRChecker.DFSR.CheckingException;
import SOFA.SOFAnode.Util.DFSRChecker.state.State;
import SOFA.SOFAnode.Util.DFSRChecker.state.TransitionPair;
import SOFA.SOFAnode.Util.DFSRChecker.state.TransitionPairs;
import SOFA.SOFAnode.Util.DFSRChecker.utils.AnotatedProtocol;

/**
 * This class represents the Restriction operator
 */
public class RestrictionNode extends TreeNode {

    /** Creates a new instance of RestrictionNode */
    public RestrictionNode(TreeNode node, TreeSet restriction,
            ActionRepository repository) {
        super("(" + node.protocol + ")\\\\{" + getRestr(restriction, repository) + "}");
        this.restriction = restriction;

        this.nodes = new TreeNode[1];
        this.nodes[0] = node;

    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getInitial()
     */
    public SOFA.SOFAnode.Util.DFSRChecker.state.State getInitial() {
        return nodes[0].getInitial();
    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#isAccepting(SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.state.State)
     */
    public boolean isAccepting(
            SOFA.SOFAnode.Util.DFSRChecker.state.State state) {
        if (nodes[0].isAccepting(state))
            return true;
        else {
            try {
                return getAccepting(state);
            } catch (InvalidParameterException e) {
                System.out.println("!!!! Exception while processing RestrictionNode::getAccepting() !!!!\n");
                System.exit(1);
                return false;
            }
        }
    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getTransitions(SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.state.State)
     */
    public TransitionPairs getTransitions(State state)
            throws InvalidParameterException, CheckingException {

        Stack stack = new Stack();
        TreeSet processed = new TreeSet();
        ArrayList result = new ArrayList();

        stack.push(state);

        while (stack.size() > 0) {
            State actual = (State) stack.pop();

            if (!processed.contains(actual)) {
                TransitionPair[] trans = nodes[0].getTransitions(actual).transitions;

                for (int i = 0; i < trans.length; ++i) {
                    Integer evindex = new Integer(trans[i].eventIndex);
                    boolean restricted = restriction.contains(new Integer(ActionRepository.getPure(evindex.intValue())));

                    if (restricted)
                        result.add(trans[i]);
                    else
                        stack.push(trans[i].state);
                }

                processed.add(actual);
            }
        }

        TransitionPair[] result1 = new TransitionPair[result.size()];
        Iterator it = result.iterator();
        int i = 0;
        while (it.hasNext())
            result1[i++] = (TransitionPair) it.next();

        return new TransitionPairs(result1);

    }

    /**
     * Finds an accepting state among the states to be collapsed.
     * 
     * @return true if there is an accepting state amon the collapsed states,
     *         false otherwise
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#isAccepting(SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.state.State)
     */
    private boolean getAccepting(State state) throws InvalidParameterException {

        Stack stack = new Stack();
        TreeSet processed = new TreeSet();

        stack.push(state);

        while (stack.size() > 0) {
            State actual = (State) stack.pop();
            TransitionPair[] trans = null;
            try {
                trans = nodes[0].getTransitions(actual).transitions;
            }
            // only consent can throw this exception
            catch (CheckingException e) {
                throw new RuntimeException("Internal error error");
            }

            for (int i = 0; i < trans.length; ++i) {
                if (!restriction.contains(new Integer(ActionRepository.getPure(trans[i].eventIndex)))) {
                    if (nodes[0].isAccepting(trans[i].state))
                        return true;
                    else if (!processed.contains(trans[i].state))
                        stack.push(trans[i].state);
                }
            }
            processed.add(actual);
        }

        return false;
    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getWeight()
     */
    public long getWeight() {
        if (weight != -1)
            return weight;
        else {
            return weight = nodes[0].getWeight();
        }
    }

    /**
     * 
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#forwardCut(java.util.TreeSet)
     */
    public TreeNode forwardCut(TreeSet livingevents) {
        TreeSet exlivev = new TreeSet();

        Iterator it = livingevents.iterator();

        while (it.hasNext()) {
            Integer event = (Integer) it.next();
            if (restriction.contains(new Integer(ActionRepository.getPure(event.intValue()))))
                exlivev.add(event);
        }

        nodes[0] = nodes[0].forwardCut(exlivev);

        if (nodes[0] == null)
            return null;

        else
            return this;
    }

    /**
     * @return the symbolic name of the treenode denoting its type
     */
    public String[] getTypeName() {
        String[] result = { "Restriction", "\\\\" };
        return result;
    }

    /**
     * @see SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.node.TreeNode#getAnotatedProtocol(SOFA.SOFAnode.Util.objectweb.fractal.behprotocols.checker.state.State)
     */
    public AnotatedProtocol getAnotatedProtocol(State state) {
        return nodes[0].getAnotatedProtocol(state);
    }

    /**
     * Creates the human-readable list of operations
     * 
     * @param restriction
     *            the set of events indices
     * @param repository
     *            Action repository to obtain the events labels
     * @return the list of operation NOT to be restricted
     */
    private static String getRestr(TreeSet restriction,
            ActionRepository repository) {
        String result = new String();
        Iterator it = restriction.iterator();

        result = repository.getItemString(((Integer) it.next()).intValue());

        while (it.hasNext())
            result += ", " + repository.getItemString(((Integer) it.next()).intValue());

        return result;
    }

    //--------------------------------------------------------------------------

    /**
     * The set of transitions (eventIndices) that should be omitted from the
     * transition graph
     */
    final private TreeSet restriction;

}