/* $Id: checker.java,v 1.2 2004/05/20 14:23:53 bures Exp $ */

package SOFA.SOFAnode.Util;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;


/**
 * An implementation class for protocol inclusion checker.
 *
 * @author  Stanislav Visnovsky
 * @version 2.0.0
 * @see     Builder
 * @see     Machine
*/
public class checker implements BehaviorProtocolChecker {

/**
 * String containing the first protocol.
*/
    private static String protocol1 = null;

/**
 * String containing the second protocol.
*/
    private static String protocol2 = null;

/**
 * A reference to a set, which contains the action tokens of the external set.
*/
    private static String external[];

/**
 * A reference to a set, which contains the action tokens of all interface events for frame to architecture.
*/
    private static String all[];

/**
 * Whether to run the conformance test. Used by the {@link checker#parseArguments} to inform
 * {@link checker#main} in case of --version and --help arguments.
*/
    private static boolean dontrun = false;
    private static int test = -1;

    public checker() {
	EdgeFactory.initActionTokens();
    }
/**
 * The main entry point of the standalone invocation of checker class. The arguments passed
 * from the command line are filenames which contain protocols. Optionally an user can
 * specify the external set. The arguments are parsed by the method parseArguments.
 * Then, the conformance is tested via int2frame and frame2arch methods.
 * The result is printed on the standard output.
 *
 * @author	Stanislav Visnovsky
 * @version     2.0.0
 *
 * @param args	parameters passed from the command line.
 * @throws	SyntaxErrorException 	in case of syntactic error of the specified protocols or sets
 * @throws	IOException		if there are problems with reading the specified files
 * @see		checker#parseArguments
 * @see		checker#int2frame
 * @see		checker#frame2arch
 *
*/
    public static void main(String[] args) throws SyntaxErrorException, IOException {

    Debug.debug = Integer.parseInt( System.getProperty( "ProtocolDebug","0"));

    String result = null;
    
      try {
        checker check = new checker();

        parseArguments( args );
        if( dontrun ) return;

	long time = System.currentTimeMillis();
        switch( test ) {
            case 0: result = check.int2frame(protocol1,true,external,protocol2); break;
            case 1: result = check.int2frame(protocol1,false,external,protocol2); break;
            case 2: result = check.frame2arch(protocol1,external,protocol2,all); break;
            case 3: result = check.int2int(protocol1,external,protocol2); break;
            default: System.out.println( "Unknown conformance test"); return;
        }

        System.out.print("1st conforms to 2nd:");
    	if( result == null ) System.err.println("yes");
    	else System.err.println("no - "+result);

	System.out.println("Statistics:");
	System.out.println("Maximal number of states:"+machineImpl.maxstates );
	Debug.println(1,"Time:"+((System.currentTimeMillis() - time)/1000) );
      } catch (java.lang.Exception e) { System.out.println("caught "+e); e.printStackTrace(); };

    }

/**
 * Parses the arguments from the command line. For version and help options prints the corresponding
 * information. If there is debug option present, it sets the Debug.debug flag to 1 to allow
 * the printing of the debugging information to standard output. If the restricting sets are
 * specified, they are parsed and the instances of the TreeSet class are created. In case
 * of syntax error, the SyntaxErrorException is raised.
 *
 * @author	Stanislav Visnovsky
 * @version     1.0.0
 *
 * @throws	SyntaxErrorException in case of syntactic error of the specified protocols or sets
 * @throws	IOException	if there are problems with reading the specified files
 * @see		checker#main
 * @see		checker#printHelp
 * @see		checker#protocol1
 * @see		checker#protocol2
 * @see		checker#external
 * @see		checker#all
 * @see		Debug
 *
*/
    private static void parseArguments( String[] args ) throws SyntaxErrorException, IOException {
	int i;

    String s1 = null;
    String s2 = null;

	for( i = 0; i< args.length; i++ ) {
	    if( args[i].compareTo( "--debug" ) == 0 ) Debug.debug = 1;
	    else if( args[i].compareTo( "--version") == 0 ) {
		System.out.println("Behavior protocol verifier v2.1");
		dontrun = true;
		return;
	    } else if( args[i].compareTo( "--help" ) == 0 ) {
		printHelp();
		dontrun = true;
		return;
	    } else if( args[i].compareTo( "-t" ) == 0 ) {
            i++;
            if( i == args.length ) throw new SyntaxErrorException( "conformance type expected");
            try {
                test = Integer.parseInt(args[i]);
            } catch( NumberFormatException e ) {
                throw new SyntaxErrorException( "conformance type has to be 0-3");
            }
            if( test<0 || test>3 ) throw new SyntaxErrorException( "conformance type has to be 0-3");
	    } else if (args[i].compareTo( "--file") == 0 ) {
	      if ( i == args.length ) throw new SyntaxErrorException( "protocol filename expected");
	      i++;
	      BufferedReader br = new java.io.BufferedReader( new java.io.FileReader(args[i]));
	      StringBuffer sb = new StringBuffer();
	      String st;
	      while ((st=br.readLine())!=null) { sb.append(st); };
	      String protdata = sb.toString();
	      /* new java.io.BufferedReader( new java.io.FileReader(args[i])).readLine(); */
	      if ( protocol1 == null) protocol1 = protdata;
	      else if (protocol2 == null) protocol2 = protdata;
	      else throw new SyntaxErrorException( "too many --file arguments");
	    } else
	      if( protocol1 == null ) protocol1 = args[i];
		 else if( protocol2 == null ) protocol2 = args[i];
		      else if( s1 == null ) s1 = args[i];
				else if( s2 == null ) s2 = args[i];
                    else { printHelp(); throw new SyntaxErrorException("Too much arguments") ; }
	}

	if( protocol1 == null || protocol2 == null ) {
	    printHelp();
	    throw new SyntaxErrorException("Not enough arguments");
	}

    Collection extern = new LinkedList();

    if( s1 != null ) {

	    Reader source = new FileReader( s1 );
	    while( source.ready() ) {
		String event = "";
		char ch;
		do {
		    ch = (char)source.read();
		    if( ch != ',') event = event + ch;
		} while( ch != ',' && source.ready() );
		if( event.length() > 0 ) { 
		    extern.add( event );
		    ActionToken t = new ActionToken( event );
		    Debug.println(3, "Looking up "+t.name );
		    EdgeFactory.getActionTokenIdx( t );
		}
	    }
	}

    external = new String[ extern.size() ];
    int id = 0;
    for( Iterator it = extern.iterator(); it.hasNext(); )
	external[id++] = (String)it.next();

    Collection al = new LinkedList();

    if( s2 != null ) {

	    Reader source = new FileReader( s2 );
	    while( source.ready() ) {
		String event = "";
		char ch;
		do {
		    ch = (char)source.read();
		    if( ch != ',') event = event + ch;
		} while( ch != ',' && source.ready() );
		if( event.length() > 0 ) { 
		    al.add( event );
		    ActionToken t = new ActionToken(event);
		    EdgeFactory.getActionTokenIdx( t );
		}
	    }
	}

    all = new String[ al.size() ];
    id = 0;
    for( Iterator it = al.iterator(); it.hasNext(); )
	all[id++] = (String)it.next();

    if( Debug.debug == 0 ) Debug.debug = Integer.parseInt( System.getProperty( "ProtocolDebug","0"));

    Debug.println("Command line arguments:");
    Debug.println("Protocol 1: "+protocol1 );
    Debug.println("Protocol 2: "+protocol2 );
    Debug.println("Conformance type: "+test );
    Debug.println("Debug: " + Debug.debug );
    Debug.print("External: " );
	for( id = 0; id < external.length; id++)
	    Debug.print( external[id]+", ");
    Debug.println();
//    Debug.println("All (for type 2): "+all.toList() );
    }

/**
 * Prints the help for invoking the checker by its main method.
 *
 * @author	Stanislav Visnovsky
 * @version     1.0.0
 *
 * @see		checker#main
 *
*/
    private static void printHelp() {
	System.out.println("Behavior protocol verifier v2.1");
	System.out.println("usage: java SOFA.SOFAnode.Util.checker -t <type>[--debug] protocol1 protocol2 [external]");
	System.out.println("       java SOFA.SOFAnode.Util.checker --version");
	System.out.println("       java SOFA.SOFAnode.Util.checker --help");
	System.out.println();
	System.out.println("where <type> is:");
	System.out.println("0   for provides interface protocol vs frame protocol");
	System.out.println("1   for requires interface protocol vs frame protocol");
	System.out.println("2   for frame protocol vs architecture protocol");
	System.out.println("3   for interface protocol vs interface protocol");
    }

/**
 * Checks, whether the protocols passed as arguments generate language that are in 
 * inclusion, i.e., L(PA) is a subset of L(PB).
 * Before the inclusion homomorphism lookup, there must be the reduction performed.
 *
 * @author	Stanislav Visnovsky
 * @version     2.5.0
 *
 * @param mA	the first protocol
 * @param mB	the second protocol
 * @return	result of the inclusion test, null for success, counterexample otherwise
 * @see		checker#main
 * @see		Machine#Reduce
 * @see		Machine#Determinize
 * @see		Machine#Reachable
 * @see		Machine#isIncludedIn
 *
*/
    private static String inclusion( Protocol A, Protocol B ) throws SyntaxErrorException
    {

	Builder builder = new Builder();
	
        Debug.println(1,"Start of inclusion testing");
        
	A = distributeRepetition( A );
	B = distributeRepetition( B );
	
	Debug.println(1,"Heuristics done" );
	Debug.println(2,"Left protocol:");
	((Printer)A).Print( 2 );
	Debug.println(2,"Right protocol:");
	((Printer)B).Print( 2 );
	
	long time = System.currentTimeMillis();

	Protocol PmA = A.Copy() ;

	Machine mA = PmA.createMachine();
	Debug.println(1,"inclusion left (A) machine:");
	((Printer)mA).Print(2);

	System.gc();
	
	Machine mBc1 = B.createMachine();
	Debug.println(1,"inclusion right (B) machine:");
	((Printer)mBc1).Print(2);

	System.gc();
	
        Debug.println(1, "Time for building: "+ (System.currentTimeMillis()-time));

        // lookup inclusion
	time = System.currentTimeMillis();
        Debug.println(1,"Inclusion lookup:");
        String result = mA.isIncludedIn( mBc1 );
        Debug.println(1, "Time for inclusion lookup: "+ ((System.currentTimeMillis()-time)/1000)+"s");

	return result;
    }

/**
 * Checks, whether the protocol1 can be updated by protocol 2
 *
 * @author	Stanislav Visnovsky
 * @version     1.0.0
 *
 * @param protocol1	the first protocol
 * @param protocol2	the second protocol
 * @param external	a set of event tokens to be evaluated
 * @param provision	a set of event tokens to be considered as provision alphabet
 * @return		result of the inclusion test, null for success, counterexample otherwise
*/
  public String canAlwaysUpdate( Protocol protocol1, Protocol protocol2, ActionTokenArray external, ActionTokenArray provision)
    throws SyntaxErrorException
  {
    Debug.println(1,"Start of canAlwaysUpdate" );

    // condition 1

    Protocol leftproto = protocol2.Copy();

    leftproto.Restrict( new ActionTokenArray( provision ) );
    leftproto = new ProtocolRestriction( leftproto, provision );
    
    Debug.println(2, "Left of condition 1 restricted to "+provision.toList()+":" );
    leftproto.Print( 2 );
    
    Protocol rightproto = new ProtocolUpdateStripped( protocol2.Copy() ) ;
    rightproto = new ProtocolSequence( rightproto, protocol1.Copy() );

    rightproto.Restrict( new ActionTokenArray( provision ) );
    rightproto = new ProtocolRestriction( rightproto, provision );
    
    Debug.println(2, "Right of condition 1 restricted:" );
    rightproto.Print( 2 );
    
    String result = inclusion( leftproto,rightproto );
    if( result != null) return "Condition 1 violated:"+result;

    //condition 2

    Protocol pom = new ProtocolUpdateStripped( protocol2.Copy() ) ;
    pom = new ProtocolSequence( pom, protocol1.Copy() );
    pom.Restrict( new ActionTokenArray( external ) );
    pom = new ProtocolRestriction( pom, external );

    leftproto = new ProtocolAdjustment( leftproto , pom, new ActionTokenArray(provision) ) ;

    rightproto = protocol2;
    rightproto.Restrict( new ActionTokenArray( external ) );
    rightproto = new ProtocolRestriction( rightproto, external );

    result = 
inclusion( leftproto, rightproto );    return result == null ? null : "Condition 2 violated:"+result;

  }

  public String iscompliant( Protocol protocol1, Protocol protocol2, ActionTokenArray external)
    throws SyntaxErrorException
  {

    Debug.println(1,"Start of iscomplaint" );

    // condition 1

    Protocol leftproto = protocol2.Copy();

    leftproto.Restrict( new ActionTokenArray( external ) );
    leftproto = new ProtocolRestriction( leftproto, external );
    
    Debug.println(2, "Left of condition 1 restricted to "+external.toList()+":" );
    leftproto.Print( 2 );
    
    Protocol rightproto = protocol1.Copy();
    rightproto.Restrict( new ActionTokenArray( external ) );
    rightproto = new ProtocolRestriction( rightproto, external );
    
    Debug.println(2, "Right of condition 1 restricted:" );
    rightproto.Print( 2 );
    
    String result = inclusion( leftproto,rightproto );
    if( result != null) return "Condition 1 violated:"+result;

    //condition 2

    leftproto = new ProtocolAdjustment( leftproto , protocol1, new ActionTokenArray(external) ) ;
    rightproto = protocol2;

    result = 
inclusion( leftproto, rightproto );    return result == null ? null : "Condition 2 violated:"+result;
  }

/**
 * Checks the conformance relation between interface protocol and frame protocol.
 * This is done by preparing correct arguments for the iscompliant method, which
 * is used to determine the result (an error message or null for success).
 * This method works for both provides and requires protocols.
 * No renaming takes place. Implemented by invoking the method with the same name
 * but with an ActionTokenArray as a set of external events.
 *
 * @author	Stanislav Visnovsky
 * @version     1.0.0
 *
 * @return          null for successufull test or an error message.
 * @param ifaceProt interface protocol
 * @param frameProt frame protocol
 * @param ifaceOperations   an array of strings for each method on the interface
 * @param provide   true for provides interface, false for requires interface

 * @see		checker#iscompliant
 * @see		ActionTokenArray
 * @see		checker#frame2arch
 *
*/
  public String int2frame(String ifaceProt, boolean provide, String[] ifaceOperations, String frameProt)
    throws SyntaxErrorException
  {
    EdgeFactory.initActionTokens();
    Debug.debug = Integer.parseInt(System.getProperty( "ProtocolDebug","0" ));

    for( int i = 0 ; i<ifaceOperations.length; i++ )
	Builder.fillOneName( ifaceOperations[i],  false );

    ActionTokenArray ifaceOps = new ActionTokenArray( ifaceOperations );

	String result;

	Builder builder = new Builder();

	builder.fillIdentifiers( ifaceProt );
	builder.fillIdentifiers( frameProt );

	Protocol iface = builder.Build( ifaceProt );
	Protocol frame = builder.Build( frameProt );

	Debug.println(3,"Restrict to:" + ifaceOps.toList() );
	frame.Restrict( new ActionTokenArray( ifaceOps ) );

	Debug.println(2,"Int2Frame restricted frame:");
	((Printer)frame).Print(2);

        // we can ignore requirement for B(Pf), since interface cannot contain [UPDATE]
	if( provide ) {
            result = inclusion( iface, frame );  // this is reduced notion of protocol conformance for provides interface vs frame protocols
	    if( result == null ) return null;
            else return "Frame protocol does not conform to provides interface protocol: Condition 1 violated: "+result;
        } else {
            result = inclusion( frame, iface );
	    if( result == null ) return null; // this is reduced notion of protocol conformance for requires interface vs frame protocols
            else return "Frame protocol does not conform to requires interface protocol: Condition 2 vioated: "+result;
        }
  }

/**
 * Checks the conformance relation between frame protocol and architecture protocol.
 * This is done by preparing correct arguments for the iscompliant method, which
 * is used to determine the result (an error message or null for success). The architecture
 * protocol is restricted to only outer interfaces (both provides and requires).
 * No renaming takes place.
 *
 * @author	Stanislav Visnovsky
 * @version     1.0.0
 *
 * @return          null for successufull test or an error message.
 * @param archProt  architecture protocol
 * @param frameProt frame protocol
 * @param providesOperations    an array of strings for each method on each provides interface of the frame
 * @param ifaceOperations       an array of strings for each method on each interface of the frame

 * @see		checker#iscompliant
 * @see		ActionTokenArray
 * @see		checker#int2frame
 *
*/
  public String frame2arch(String frameProt, String[] providesOperations,
  String archProt, String[] ifaceOperations) throws SyntaxErrorException
  {
    EdgeFactory.initActionTokens();
    Debug.debug = Integer.parseInt(System.getProperty( "ProtocolDebug","0" ));

    for( int i = 0 ; i<providesOperations.length; i++ )
	Builder.fillOneName( providesOperations[i],  false );

    for( int i = 0 ; i<ifaceOperations.length; i++ )
	Builder.fillOneName( ifaceOperations[i],  false );

    Builder builder = new Builder();

    builder.fillIdentifiers( frameProt );
    builder.fillIdentifiers( archProt );

    ActionTokenArray provides = new ActionTokenArray( providesOperations );
    ActionTokenArray iface = new ActionTokenArray(ifaceOperations);

        Debug.println( 1, "Starting frame to architecture conformance" );

	Debug.println( 2, "Frame protocol: "+frameProt );
	Debug.println( 2, "Architecture protocol: "+archProt );
	Debug.println( 2, "External events: "+provides.toList() );
	Debug.println( 2, "All frame events: "+iface.toList() );

	builder.fillIdentifiers( frameProt );
	builder.fillIdentifiers( archProt );

	Protocol frame = builder.Build( frameProt );
	Protocol arch = builder.Build( archProt );
	
	arch = new ProtocolRestriction( arch,  iface );

	String result = iscompliant( arch, frame, provides );
        if( result == null ) return null;
        else return "Frame protocol does not conform to architecture protocol: "+result;
  }

/**
 * Checks the conformance relation between two interface protocols.
 * This is done by preparing correct arguments for the iscompliant method, which
 * is used to determine the result (an error message or null for success).
 * No renaming takes place. Implemented by invoking the method with the same name
 * but with an ActionTokenArray as a set of external events.
 *
 * @author	Stanislav Visnovsky
 * @version     1.0.0
 *
 * @return          null for successufull test or an error message.
 * @param intProt1 1st interface protocol
 * @param intProt2 2st interface protocol
 * @param providesOperations   an array of strings for each method on an interface

 * @see		ActionTokenArray
 * @see		checker#int2int
 *
*/

  public String int2int(String intProt1, String[] providesOperations, String intProt2)
    throws SyntaxErrorException
  {
    EdgeFactory.initActionTokens();
    Debug.debug = Integer.parseInt(System.getProperty( "ProtocolDebug","0" ));
    for( int i = 0 ; i<providesOperations.length; i++ )
	Builder.fillOneName( providesOperations[i],  false );

    ActionTokenArray provides = new ActionTokenArray( providesOperations );
    throw new SyntaxErrorException( "Conformance type not supported" );
  }

  public int StatisticsMaxStates() {
    return machineImpl.maxstates;
  }

  public void InitStatistics() {
    machineImpl.maxstates = 0;
  }

  private static Protocol distributeRepetition( Protocol prot ) {
    if( prot == null ) return null;

    if( prot.getElement() == "*" ) {
	Debug.println( 3, "Repetition distribution");
	Protocol l = prot.Left();
	if( l instanceof ProtocolOrParallel ) {
	    Protocol protl = new ProtocolRepetition( l.Left() );
	    Protocol protr = new ProtocolRepetition( l.Right() );
	    protl = distributeRepetition( protl );
	    protr = distributeRepetition( protr );
	    return new ProtocolOrParallel( protl, protr );
	} else return new ProtocolRepetition( distributeRepetition(l) );
    } else if( prot instanceof ProtocolToken ) {
	return new ProtocolToken( (ActionToken)prot.getElement() );
    } else { // if it is not repetition
	Object protl = distributeRepetition( prot.Left() );
	Object protr = distributeRepetition( prot.Right() );
	return prot.getSameType( protl, protr );
    }
  }
}
